using System;
using log4net;
using System.Data;
using System.Data.SqlClient;
using System.Data.SqlTypes;
using WebseitenMonitor.DomainObjects.Behaviour;
using System.Configuration;
using System.Collections;

namespace WebseitenMonitor.DataAccess 
{
	public class ChangeObservationsMediator 
	{
		private static readonly ILog log = LogManager.GetLogger(typeof(ChangeObservationsMediator));
		private static readonly string DbTableOwnerPrefix = ((string)ConfigurationSettings.AppSettings["DbTableOwnerPrefix"]).Trim();
		
		internal enum UpdateResult {Ok, InvalidParams, Error}
		internal enum WriteResult {Ok, InvalidParams, Error}
		internal enum ReadResult {Ok, InvalidParams, Error}
		
		/// <summary>
		/// This method writes a new ChangeObservation to the database.
		/// Initial default values: Date			... NULL
		///							ChangeState		... ChangeState.Init
		///							RootDocument	...	NULL
		/// </summary>
		/// <param name="name">Observation name. If null or length > 50 InvalidParams is returned.</param>
		/// <param name="url">Url to observe by the Observation. If null InvalidParams is returned.</param>
		/// <param name="categoryId">
		/// Id of the Category to which the Observation should belong.
		/// If less then 0 InvalidParams is returned.
		/// If not existent in database - Error is returned due to database-constraint violation while insert.
		/// </param>
		/// <param name="depth">ChangeObservation depth. If negative InvalidParams is returned.</param>
		/// <param name="errorMsg">Error Message if a value other than Ok is returned.</param>
		/// <param name="id">Automatically generated Database id of the new ChangeObservation.</param>
		/// <returns>Ok, InvalidParams or Error in case of problem while database access or database constraints.</returns>
		internal static ChangeObservationsMediator.WriteResult WriteNewChangeObservation(
			string name,
			Uri url,
			int categoryId,
			int depth,
			out int id,
			out string errorMsg
			) 
		{
			// Check Parameters
			if(name == null || name.Length > 50) 
			{
				errorMsg = "Es muss ein Name mit einer max. Lnge von 50 Zeichen angegeben werden.";
				id = -1;
				return WriteResult.InvalidParams;
			}
			if(url == null) 
			{
				errorMsg = "Es muss eine Url angegeben werden.";
				id = -1;
				return WriteResult.InvalidParams;
			}
			if(categoryId < 0) 
			{
				errorMsg = "Es muss ein gltiger Kategoriebezug angegeben werden.";
				id = -1;
				return WriteResult.InvalidParams;
			}
			if(depth < 0) 
			{
				errorMsg = "Es muss eine nichtnegative berwachungstiefe angegeben werden.";
				id = -1;
				return WriteResult.InvalidParams;
			}
			// assertion: parameters passed initial checking phase
			string conStr = ConfigurationSettings.AppSettings["DbConnectionString"];
			IDbConnection con = null;
			IDbTransaction trans = null;
			try 
			{
				con = new SqlConnection(conStr);
				con.Open();
				trans = con.BeginTransaction(); // Note: insertion in Observations and ChangeObservations in a transaction
				// First part: insert general properties into Observations table
				IDbCommand cmd = con.CreateCommand();
				cmd.CommandText = "INSERT INTO " + DbTableOwnerPrefix + "Observations (Kind,Name,Url,ChangeState,Fk_Categories) VALUES (@KIND,@NAME,@URL,@CHANGESTATE,@FKCATEGORIES);" +
					"SELECT SCOPE_IDENTITY() AS [SCOPE_IDENTITY]";
				cmd.Parameters.Add(new SqlParameter("@KIND", (int)ObservationKind.ChangeObservation));
				cmd.Parameters.Add(new SqlParameter("@NAME", name));
				cmd.Parameters.Add(new SqlParameter("@URL", url.ToString()));
				cmd.Parameters.Add(new SqlParameter("@CHANGESTATE", (int)ChangeStateEnum.Init));
				cmd.Parameters.Add(new SqlParameter("@FKCATEGORIES", categoryId));
				cmd.Transaction = trans;
				int insertedObservationId = Convert.ToInt32((decimal)cmd.ExecuteScalar());
				if(insertedObservationId < 0) 
				{
					trans.Rollback();
					errorMsg = "Es ist ein inkonsistenter Datenbanknderungsversuch aufgetreten.";
					id = -1;
					return WriteResult.Error;
				}
				// Second part: insert ChangeObservations specific properties
				cmd = con.CreateCommand();
				cmd.CommandText = "INSERT INTO " + DbTableOwnerPrefix + "ChangeObservations (Id,Depth) VALUES (@ID,@DEPTH)";
				cmd.Parameters.Add(new SqlParameter("@ID", insertedObservationId));
				cmd.Parameters.Add(new SqlParameter("@DEPTH", depth));
				cmd.Transaction = trans;
				cmd.ExecuteNonQuery();
				trans.Commit(); // commit transaction if both inserts were successfull
				errorMsg = null;
				id = insertedObservationId;
				return WriteResult.Ok;
			} 
			catch(Exception e)
			{
				if(trans != null)
					trans.Rollback();
				log.Error("Exception occured during writing new ChangeObservation to database.", e);
				errorMsg = "Es ist ein Problem beim Datenbankzugriff aufgetreten.";
				id = -1;
				return WriteResult.Error;
			} 
			finally 
			{
				try { if(con != null) con.Close(); } 
				catch(Exception ex) 
				{
					log.Warn("Unable to close connection - exception occured.", ex);
				}
			}
		} // end: WriteNewChangeObservation
		
		internal static ChangeObservationsMediator.UpdateResult UpdateChangedPropertiesChangeObservation(
			ChangeObservation chObs,
			out string errorMsg) 
		{
			// Check Parameters
			if(chObs == null) 
			{
				errorMsg = "Es muss ein passender Beobachtungskanal angegeben werden.";
				return UpdateResult.InvalidParams;
			}
			// The rootDocument of the ChangeObservation may only be null if in Init state after change of Properties
			if(chObs.RootDocument == null && !(chObs.ChangeState == ChangeStateEnum.Init)) 
			{
				errorMsg = "Der Beobachtungskanal ist inkonsistent.";
				return UpdateResult.InvalidParams;
			}
			if(chObs.Name == null || chObs.Name.Length > 50) 
			{
				errorMsg = "Der Beobachtungskanal muss einen Namen mit einer max. Lnge von 50 Zeichen haben.";
				return UpdateResult.InvalidParams;
			}
			if(chObs.Url == null) 
			{
				errorMsg = "Der Beobachtungskanal muss eine Url enthalten.";
				return UpdateResult.InvalidParams;
			}
			if(chObs.Depth < 0) 
			{
				errorMsg = "Der Beobachtungskanal muss eine nichtnegative berwachungstiefe haben.";
				return UpdateResult.InvalidParams;
			}
			// assertion: parameters passed initial checking phase
			string conStr = ConfigurationSettings.AppSettings["DbConnectionString"];
			IDbConnection con = null;
			IDbTransaction trans = null;
			try 
			{
				con = new SqlConnection(conStr);
				con.Open();
				trans = con.BeginTransaction();
				// Step 1: update general properties in Observations table
				IDbCommand cmd = con.CreateCommand();
				cmd.CommandText = "UPDATE " + DbTableOwnerPrefix + "Observations SET Kind = @KIND, Name = @NAME, Url = @URL, Date = @DATE , CheckCounterTimeStamp = @TIMESTAMP , ChangeState = @STATE WHERE Id = @ID";
				cmd.Parameters.Add(new SqlParameter("@KIND", (int)ObservationKind.ChangeObservation));
				cmd.Parameters.Add(new SqlParameter("@NAME", chObs.Name));
				cmd.Parameters.Add(new SqlParameter("@URL", chObs.Url.ToString()));
				if(chObs.Date.Equals(Observation.InitDateTime))
					cmd.Parameters.Add(new SqlParameter("@DATE", DBNull.Value));
				else
					cmd.Parameters.Add(new SqlParameter("@DATE", chObs.Date));
				if(chObs.CheckCounterTimeStamp == -1)
					cmd.Parameters.Add(new SqlParameter("@TIMESTAMP", DBNull.Value));
				else
					cmd.Parameters.Add(new SqlParameter("@TIMESTAMP", chObs.CheckCounterTimeStamp));
				cmd.Parameters.Add(new SqlParameter("@STATE", (int)chObs.ChangeState));
				cmd.Parameters.Add(new SqlParameter("@ID", chObs.Id));
				cmd.Transaction = trans;
				int updateCount = cmd.ExecuteNonQuery();
				if(updateCount != 1) 
				{
					log.Error("Update of Observations table for properties changed ChangeObservation didn't affect exactly one row [#affected: " + updateCount + "].");
					trans.Rollback();
					errorMsg = "Es ist eine Datenbankinkonsistenz aufgetreten.";
					return UpdateResult.Error;
				}
				// Step 2: if RootDocument is null delete all possible old ObservedDocuments
				if(chObs.RootDocument == null) 
				{
					cmd = con.CreateCommand();
					cmd.CommandText = "DELETE FROM " + DbTableOwnerPrefix + "ObservedDocuments WHERE Fk_ChangeObservation = @ID";
					cmd.Parameters.Add(new SqlParameter("@ID", chObs.Id));
					cmd.Transaction = trans;
					int rowCount = cmd.ExecuteNonQuery();
				}
				// Step 3: update ChangeObservation specific properties in ChangeObservations table
				cmd = con.CreateCommand();
				if(chObs.RootDocument == null) 
				{
					cmd.CommandText = "UPDATE " + DbTableOwnerPrefix + "ChangeObservations SET RootDocument = @ROOTDOC , Depth = @Depth WHERE Id = @ID";
					cmd.Parameters.Add(new SqlParameter("@ROOTDOC", DBNull.Value));
					cmd.Parameters.Add(new SqlParameter("@DEPTH", chObs.Depth));
				} 
				else 
				{
					cmd.CommandText = "UPDATE " + DbTableOwnerPrefix + "ChangeObservations SET Depth = @Depth WHERE Id = @ID";
					cmd.Parameters.Add(new SqlParameter("@DEPTH", chObs.Depth));
				}
				cmd.Parameters.Add(new SqlParameter("@ID", chObs.Id));
				cmd.Transaction = trans;
				updateCount = cmd.ExecuteNonQuery();
				if(updateCount != 1) 
				{
					log.Error("Update of ChangeObservations table didn't affect exactly one row [#affected: " + updateCount + "].");
					trans.Rollback();
					errorMsg = "Es ist eine Datenbankinkonsistenz aufgetreten.";
					return UpdateResult.Error;
				}
				trans.Commit();
				errorMsg = null;
				return UpdateResult.Ok;
			} 
			catch(Exception e)
			{
				if(trans != null)
					trans.Rollback();
				log.Error("Exception occured during updating changed properties ChangeObservation to database.", e);
				errorMsg = "Es ist ein Problem beim Datenbankzugriff aufgetreten.";
				return UpdateResult.Error;
			} 
			finally 
			{
				try { if(con != null) con.Close(); } 
				catch(Exception ex) 
				{
					log.Warn("Unable to close connection - exception occured.", ex);
				}
			}
		} // end: UpdateChangedPropertiesChangeObservation
		
		internal static ChangeObservationsMediator.UpdateResult UpdateCheckedChangeObservation(
			ChangeObservation chObs,
			out string errorMsg) 
		{
			if(chObs == null) 
			{
				errorMsg = "Es muss ein passender Beobachtungskanal angegeben werden.";
				return UpdateResult.InvalidParams;
			}
			// The rootDocument of the ChangeObservation may only be null if in Init or Error state - otherwise inconsistent.
			if(chObs.RootDocument == null && !(chObs.ChangeState == ChangeStateEnum.Init) && !(chObs.ChangeState == ChangeStateEnum.Error)) 
			{
				errorMsg = "Der Beobachtungskanal ist inkonsistent.";
				return UpdateResult.InvalidParams;
			}
			if(chObs.Depth < 0) 
			{
				errorMsg = "Der Beobachtungskanal ist inkonsistent";
				return UpdateResult.InvalidParams;
			}
			// assertion: parameters passed initial checking phase
			string conStr = ConfigurationSettings.AppSettings["DbConnectionString"];
			IDbConnection con = null;
			IDbTransaction trans = null;
			try 
			{
				con = new SqlConnection(conStr);
				con.Open();
				trans = con.BeginTransaction();
				// Step 1: update general properties in Observations table
				IDbCommand cmd = con.CreateCommand();
				cmd.CommandText = "UPDATE " + DbTableOwnerPrefix + "Observations SET Date = @DATE , CheckCounterTimeStamp = @TIMESTAMP , ChangeState = @STATE WHERE Id = @ID";
				cmd.Parameters.Add(new SqlParameter("@DATE", chObs.Date));
				cmd.Parameters.Add(new SqlParameter("@TIMESTAMP", chObs.CheckCounterTimeStamp));
				cmd.Parameters.Add(new SqlParameter("@STATE", (int)chObs.ChangeState));
				cmd.Parameters.Add(new SqlParameter("@ID", chObs.Id));
				cmd.Transaction = trans;
				int updateCount = cmd.ExecuteNonQuery();
				if(updateCount != 1) 
				{
					log.Error("Update of Observations table for newly checked ChangeObservation didn't affect exactly one row [#affected: " + updateCount + "].");
					trans.Rollback();
					errorMsg = "Es ist eine Datenbankinkonsistenz aufgetreten.";
					return UpdateResult.Error;
				}
				// Step 2: delete possible old ObservedDocuments for this ChangeObservation
				cmd = con.CreateCommand();
				cmd.CommandText = "DELETE FROM " + DbTableOwnerPrefix + "ObservedDocuments WHERE Fk_ChangeObservation = @ID";
				cmd.Parameters.Add(new SqlParameter("@ID", chObs.Id));
				cmd.Transaction = trans;
				int rowCount = cmd.ExecuteNonQuery();
				// Step 2.1: if no rootDocument (state = Init) update ChangeObservation specific properties and ready
				if(chObs.RootDocument == null) 
				{
					cmd = con.CreateCommand();
					cmd.CommandText = "UPDATE " + DbTableOwnerPrefix + "ChangeObservations SET RootDocument = @ROOTDOC , Depth = @Depth WHERE Id = @ID";
					cmd.Parameters.Add(new SqlParameter("@ROOTDOC", DBNull.Value));
					cmd.Parameters.Add(new SqlParameter("@DEPTH", chObs.Depth));
					cmd.Parameters.Add(new SqlParameter("@ID", chObs.Id));
					cmd.Transaction = trans;
					updateCount = cmd.ExecuteNonQuery();
					if(updateCount != 1) 
					{
						log.Error("Update of ChangeObservation table didn't affect exactly one row [#affected: " + updateCount + "].");
						trans.Rollback();
						errorMsg = "Es ist eine Datenbankinkonsistenz aufgetreten.";
						return UpdateResult.Error;
					}
					trans.Commit();
					errorMsg = null;
					return UpdateResult.Ok;
				}
				// Step 3: insert RootDocument of ChangeObservation as ObservedDocument with no father into ObservedDocuments table
				int rootDocId = writeSingleObservedDocument(chObs.RootDocument, chObs.Id, -1, con, trans);
				log.Debug("rootDocId: " + rootDocId + " , chObs.Id: " + chObs.Id);
				if(rootDocId < 0) // Note: -1 is error-return-value of writeSingleObservedDocument-method
				{
					log.Error("Insert of rootDocument of ChangeObservation into ObservedDocuments table failed.");
					trans.Rollback();
					errorMsg = "Es ist ein Fehler beim Datenbankzugriff aufgetreten.";
					return UpdateResult.Error;
				}
				chObs.RootDocument.id = rootDocId;
				// Step 4: update (set) RootDocument and Depth of ChangeObservation in ChangeObservations table
				cmd = con.CreateCommand();
				cmd.CommandText = "UPDATE " + DbTableOwnerPrefix + "ChangeObservations SET RootDocument = @ROOTDOC , Depth = @DEPTH WHERE Id = @ID";
				cmd.Parameters.Add(new SqlParameter("@ROOTDOC", rootDocId));
				cmd.Parameters.Add(new SqlParameter("@DEPTH", chObs.Depth));
				cmd.Parameters.Add(new SqlParameter("@ID", chObs.Id));
				cmd.Transaction = trans;
				updateCount = cmd.ExecuteNonQuery();
				if(updateCount != 1) 
				{
					log.Error("Update of RootDocument-reference in ChangeObservations table didn't affect exactly one row [#affected: " + updateCount + "].");
					trans.Rollback();
					errorMsg = "Es ist eine Datenbankinkonsistenz aufgetreten.";
					return UpdateResult.Error;
				}
				// Step 4: Insert ObservedDocuments Tree into ObservedDocuments tree
				bool docTreeWriteSuccess = writeObservedDocumentsTreeRootExclusive(chObs.RootDocument, chObs.Id, con, trans);
				if(!docTreeWriteSuccess) 
				{
					log.Error("Insert of ObservedDocuments-tree into ObservedDocuments table failed.");
					trans.Rollback();
					errorMsg = "Es ist ein Fehler beim Datenbankzugriff aufgetreten.";
					return UpdateResult.Error;
				}
				trans.Commit();
				errorMsg = null;
				return UpdateResult.Ok;
			} 
			catch(Exception e)
			{
				if(trans != null)
					trans.Rollback();
				log.Error("Exception occured writing checked ChangeObservation to database.", e);
				errorMsg = "Es ist ein Problem beim Datenbankzugriff aufgetreten.";
				return UpdateResult.Error;
			} 
			finally 
			{
				try { if(con != null) con.Close(); } 
				catch(Exception ex) 
				{
					log.Warn("Unable to close connection - exception occured.", ex);
				}
			}
		} // end: method UpdateChangeObservation
		
		// This method writes a tree of ObservedDocuments into the ObservedDocuments table of the database.
		// Notes: -) rootDoc itself isn't written to the database.
		//        -) This method uses the writeSingleObservedDocument method to write a single ObservedDocument
		//           to the database.
		//        -) This method should be called within a try/catch block for database-access-exceptions, because
		//           it internally doesn't catch any exceptions.
		//        -) returns true if all ObservedDocuments in tree were written without any problem, otherwise false.
		private static bool writeObservedDocumentsTreeRootExclusive(
			ChangeObservation.ObservedDocument rootDoc,
			int observationId,
			IDbConnection con,
			IDbTransaction trans) 
		{
			ArrayList curOpenDocumentList = new ArrayList();
			ArrayList tmpOpenDocumentList;
			// initialize breadth-first-traversal
			curOpenDocumentList.Add(rootDoc);
			// breadth-first-traverse ObservedDocumentTree for writing to database
			while(curOpenDocumentList.Count > 0) 
			{
				tmpOpenDocumentList = new ArrayList();
				foreach(ChangeObservation.ObservedDocument curDoc in curOpenDocumentList) 
				{
					if(curDoc.childDocuments != null) // null if observation-depth reached
					{
						foreach(ChangeObservation.ObservedDocument tmpDoc in curDoc.childDocuments) 
						{
							int tmpDocId = writeSingleObservedDocument(tmpDoc, observationId, curDoc.id, con, trans);
							if(tmpDocId < 0) // Note: -1 is error-return-value writeSingleObservedDocument
								return false;
							tmpDoc.id = tmpDocId;
							// Insert ObservedDocument into tmp-list which gets cur-list for next loop
							// Note: Delegate-kinded ObservedDocuments need not get regarded, because they
							//       have no further child-ObservedDocuments
							if(tmpDoc.kind != ChangeObservation.ObservedDocumentKind.Delegate)
								tmpOpenDocumentList.Add(tmpDoc);
						}
					}
				}
				// set cur-list for next loop as tmp-list of last loop, which contains obtained childs of last loop.
				curOpenDocumentList = tmpOpenDocumentList; 
			}
			return true;
		} // end: method writeObservedDocumentsTreeRootExclusive

		// returns automatic created Id of inserted ObservedDocuments row or -1 in case of an error
		
		// This method writes a single ObservedDocument into the ObservedDocuments table of the database and
		// returns the automatic created Id of the inserted ObservedDocuments row or -1 ic case of an error
		// occured while accessing the database.
		private static int writeSingleObservedDocument(
			ChangeObservation.ObservedDocument doc,
			int observationId,
			int parentObservedDocumentId,
			IDbConnection con,
			IDbTransaction trans) 
		{
			log.Debug("writeSingleObservedDocument: observationId: " + observationId + " parentObservedDocumentId: " + parentObservedDocumentId);
			IDbCommand cmd = con.CreateCommand();
			cmd.CommandText = "INSERT INTO " + DbTableOwnerPrefix + "ObservedDocuments (Kind,State,Title,Url,HashValue,Fk_DelegateFor,Fk_ParentDocument,Fk_ChangeObservation) " +
				"VALUES (@KIND,@STATE,@TITLE,@URL,CONVERT(VARBINARY(100),@HASHVALUE),@FKDELEGATEFOR,@FKPARENTDOCUMENT,@FKCHANGEOBSERVATION); " +
				"SELECT SCOPE_IDENTITY() AS [SCOPE_IDENTITY]";
			cmd.Parameters.Add(new SqlParameter("@KIND", (int)doc.kind));
			cmd.Parameters.Add(new SqlParameter("@STATE", (int)doc.state));
			if(doc.title == null)
				cmd.Parameters.Add(new SqlParameter("@TITLE", DBNull.Value));
			else
				cmd.Parameters.Add(new SqlParameter("@TITLE", doc.title));
			cmd.Parameters.Add(new SqlParameter("@URL", doc.url.ToString()));
			if(doc.hashValue == null)
				cmd.Parameters.Add(new SqlParameter("@HASHVALUE", DBNull.Value));
			else
				cmd.Parameters.Add(new SqlParameter("@HASHVALUE", doc.hashValue));
			if(doc.delegateFor == null)
				cmd.Parameters.Add(new SqlParameter("@FKDELEGATEFOR", DBNull.Value));
			else
				cmd.Parameters.Add(new SqlParameter("@FKDELEGATEFOR", doc.delegateFor.id));
			if(parentObservedDocumentId < 0)
				cmd.Parameters.Add(new SqlParameter("@FKPARENTDOCUMENT", DBNull.Value));
			else
				cmd.Parameters.Add(new SqlParameter("@FKPARENTDOCUMENT", parentObservedDocumentId));
			cmd.Parameters.Add(new SqlParameter("@FKCHANGEOBSERVATION", observationId));
			cmd.Transaction = trans;
			int insertedDocId = Convert.ToInt32((decimal)cmd.ExecuteScalar());
			if(insertedDocId < 0) 
				return -1;
			return insertedDocId;
		} // end: method writeSingleObservedDocument
		
		/// <summary>
		/// This method loads a ChangeObservation object from the Database passes it back by the
		/// out parameter changeObservation.
		/// The ChangeObservation to load must be specified by it's Id, which corresponds to the
		/// Id-field in the database.
		/// </summary>
		/// <param name="observationId"></param>
		/// <param name="changeObservation"></param>
		/// <param name="errorMsg"></param>
		/// <returns></returns>
		internal static ChangeObservationsMediator.ReadResult LoadChangeObservation(
			int observationId,
			out ChangeObservation changeObservation,
			out string errorMsg) 
		{
			changeObservation = null;
			errorMsg = null;
			if(observationId < 0) 
			{
				errorMsg = "Es wurde eine ungltige Identifikationsnummer bergeben.";
				return ReadResult.InvalidParams;
			}
			// attributes of the ChangeObservation to load:
			int chObsId = -1;
			string chObsName = null;
			Uri chObsUrl = null;
			DateTime chObsDate = Observation.InitDateTime;
			int chObsCheckCounterTimeStamp = -1;
			ChangeStateEnum chObsChangeState = ChangeStateEnum.Error;
			int chObsDepth = -1;
			ChangeObservation.ObservedDocument chObsRootDocument = null;

			// assertion: parameters passed initial checking phase
			string conStr = ConfigurationSettings.AppSettings["DbConnectionString"];
			IDbConnection con = null;
			try 
			{
				con = new SqlConnection(conStr);
				con.Open();
				// Step 1: load general properties from Observations-table
				IDbCommand cmd = con.CreateCommand();
				cmd.CommandText = "SELECT * FROM " + DbTableOwnerPrefix + "Observations WHERE Id = @ID";
				cmd.Parameters.Add(new SqlParameter("@ID", observationId));
				IDataReader reader = cmd.ExecuteReader();
				int rowCount = 0;
				while(reader.Read()) 
				{
					if(rowCount > 0) 
					{
						log.Error("Id of requested Observation seems to be not unique, multiple rows found.");
						errorMsg = "Es ist eine Datenbankinkonsistenz aufgetreten.";
						return ReadResult.Error;
					}
					chObsId = (int)reader["Id"];
					if((ObservationKind)reader["Kind"] != ObservationKind.ChangeObservation) 
					{
						log.Error("Wrong Observation-Kind for load of ChangeObservation found in database [Id: " + chObsId + "]");
						errorMsg = "bergebene Identifikationsnummer kennzeichnet eine andere Art von Beobachtungskanal.";
						return ReadResult.InvalidParams;
					}
					if(reader["Date"] is DBNull)
						chObsDate = Observation.InitDateTime;
					else
						chObsDate = (DateTime)reader["Date"];
					if(reader["CheckCounterTimeStamp"] is DBNull)
						chObsCheckCounterTimeStamp = -1;
					else
						chObsCheckCounterTimeStamp = (int)reader["CheckCounterTimeStamp"];
					chObsName = (string)reader["Name"];
					chObsUrl = new Uri((string)reader["Url"]);
					chObsChangeState = (ChangeStateEnum)reader["ChangeState"];
					rowCount++;
				}
				reader.Close();
				if(rowCount == 0) 
				{
					log.Error("No corresponding row, to load requested ChangeObservation, found in Observations table [Id: " + observationId + "].");
					errorMsg = "Es konnte kein entsprechender Beobachtungskanal gefunden werden.";
					return ReadResult.InvalidParams;
				}
				// Step 2: load ChangeObservation specific properties from ChangeObservations-table
				cmd = con.CreateCommand();
				cmd.CommandText = "SELECT * FROM " + DbTableOwnerPrefix + "ChangeObservations WHERE Id = @ID";
				cmd.Parameters.Add(new SqlParameter("@ID", chObsId));
				reader = cmd.ExecuteReader();
				rowCount = 0;
				int rootDocId = -1;
				while(reader.Read()) 
				{
					if(rowCount > 0) 
					{
						log.Error("ChangeObservations-table entries not unique by Id, multiple occurence detected [Id: " + chObsId + "]");
						errorMsg = "Es ist eine Datenbankinkonsistenz aufgetreten.";
						return ReadResult.Error;
					}
					int tmpRootDocId;
					if(reader["RootDocument"] is DBNull)
						tmpRootDocId = -1;
					else
						tmpRootDocId = (int)reader["RootDocument"];
					chObsDepth = (int)reader["Depth"];
					rootDocId = tmpRootDocId;
					rowCount++;
				}
				reader.Close();
				if(rowCount == 0) 
				{
					log.Error("No corresponding ChangeObservations-table entry found for already found Observations-table entry [Id: " + chObsId + "]");
					errorMsg = "Es ist eine Datenbankinkonsistenz aufgetreten.";
					return ReadResult.Error;
				}
				// Check if Observation is in init-state - in this case no ObservedDocuments need to get loaded
				if(chObsChangeState == ChangeStateEnum.Init) 
				{
					try 
					{
						changeObservation = new ChangeObservation(chObsId, chObsName, chObsUrl, chObsDate, chObsCheckCounterTimeStamp, chObsChangeState, chObsDepth, null);
					} 
					catch (Exception e) 
					{
						log.Error("Inconsistent ChangeObservation found in database during loading it [Id: " + chObsId + "]", e);
						errorMsg = "Es ist eine Datenbankinkonsistenz aufgetreten.";
						return ReadResult.Error;
					}
					errorMsg = null;
					return ReadResult.Ok;
				}
				if(rootDocId == -1) 
				{
					// Note: If rootDocId is -1 (null) only Error or Init (treated above) are allowed as state.
					//       In this case we are ready with loading (no ObservedDocuments), otherwise an inconsistence
					//       will throw an Exception in the ChangeObservation constructor.
					try 
					{
						changeObservation = new ChangeObservation(chObsId, chObsName, chObsUrl, chObsDate, chObsCheckCounterTimeStamp, chObsChangeState, chObsDepth, null);
					} 
					catch (Exception e) 
					{
						log.Error("Inconsistent ChangeObservation found in database during loading it [Id: " + chObsId + "]", e);
						errorMsg = "Es ist eine Datenbankinkonsistenz aufgetreten.";
						return ReadResult.Error;
					}
					errorMsg = null;
					return ReadResult.Ok;
				}
				// Step 3: load root-ObservedDocument from ObservedDocument-table (if not in Init-state or no rootDocument found in db)
				cmd = con.CreateCommand();
				cmd.CommandText = "SELECT * FROM " + DbTableOwnerPrefix + "ObservedDocuments WHERE Id = @ID";
				cmd.Parameters.Add(new SqlParameter("@ID", rootDocId));
				reader = cmd.ExecuteReader();
				rowCount = 0;
				ChangeObservation.ObservedDocument rootDoc = new ChangeObservation.ObservedDocument();
				while(reader.Read()) 
				{
					if(rowCount > 0) 
					{
						log.Error("ChangeObservations-table entries not unique by Id, multiple occurence detected [Id: " + rootDocId + "]");
						errorMsg = "Es ist eine Datenbankinkonsistenz aufgetreten.";
						return ReadResult.Error;
					}
					rootDoc.id = rootDocId;
					rootDoc.kind = (ChangeObservation.ObservedDocumentKind)reader["Kind"];
					if(rootDoc.kind == ChangeObservation.ObservedDocumentKind.Delegate) 
					{
						log.Error("Root ObservedDocument found as Delegate - database inconsistence.");
						errorMsg = "Es ist eine Datenbankinkonsistenz aufgetreten.";
						return ReadResult.Error;
					}
					rootDoc.state = (ChangeObservation.ObservedDocumentState)reader["State"];
					rootDoc.title = (string)reader["Title"];
					rootDoc.url = new Uri((string)reader["Url"]);
					rootDoc.hashValue = (byte[])reader["HashValue"];
					rowCount++;
				}
				reader.Close();
				if(rowCount == 0) 
				{
					log.Error("No corresponding ObservedDocument-table entry found for root ObservedDocument with Id: " + rootDocId );
					errorMsg = "Es ist eine Datenbankinkonsistenz aufgetreten.";
					return ReadResult.Error;
				}
				chObsRootDocument = rootDoc;
				// Step 4: load ObservedDocuments-tree from ObservedDocuments-table
				// alreadyLoadedDocuments holds all loaded Documents with Id as Key, this Hashtable
				// is used to create the object-reference representation of the database id-reference.
				Hashtable alreadyLoadedDocuments = new Hashtable();
				alreadyLoadedDocuments.Add(rootDocId, chObsRootDocument);
				ArrayList curParentIdList = new ArrayList();
				curParentIdList.Add(rootDocId);
				ArrayList tmpParentIdList = new ArrayList();
				for(int i = 0; i < chObsDepth; i++) 
				{
					foreach(int tmpParentId in curParentIdList) 
					{
						ChangeObservation.ObservedDocument tmpParentDoc = (ChangeObservation.ObservedDocument)alreadyLoadedDocuments[tmpParentId];
						tmpParentIdList = new ArrayList();
						cmd = con.CreateCommand();
						cmd.CommandText = "SELECT * FROM " + DbTableOwnerPrefix + "ObservedDocuments WHERE Fk_ParentDocument = @ID";
						cmd.Parameters.Add(new SqlParameter("@ID", tmpParentDoc.id));
						reader = cmd.ExecuteReader();
						bool firstFound = false;
						while(reader.Read()) 
						{
							if(!firstFound) 
							{
								tmpParentDoc.childDocuments = new ArrayList();
								firstFound = true;
							}
							ChangeObservation.ObservedDocument tmpChildDoc = new ChangeObservation.ObservedDocument();
							tmpChildDoc.id = (int)reader["Id"];
							tmpChildDoc.kind = (ChangeObservation.ObservedDocumentKind)reader["Kind"];
							tmpChildDoc.state = (ChangeObservation.ObservedDocumentState)reader["State"];
							if(reader["Title"] is DBNull)
								tmpChildDoc.title = null;
							else
								tmpChildDoc.title = (string)reader["Title"];
							tmpChildDoc.url = new Uri((string)reader["Url"]);
							if(reader["HashValue"] is DBNull)
								tmpChildDoc.hashValue = null;
							else
								tmpChildDoc.hashValue = (byte[])reader["HashValue"];
							if(tmpChildDoc.kind == ChangeObservation.ObservedDocumentKind.Delegate) 
							{

								int delegateId = -1;
								if(reader["Fk_DelegateFor"] is DBNull || (int)reader["Fk_DelegateFor"] < 0) 
								{
									log.Error("Delegate-Kinded ObservedDocument loaded with unset Delegate-Id of " + delegateId + " [Id: " + tmpChildDoc.id + "]");
									errorMsg = "Es ist eine Datenbankinkonsistenz aufgetreten.";
									return ReadResult.Error;
								}
								else
									delegateId = (int)reader["Fk_DelegateFor"];

								tmpChildDoc.delegateFor = (ChangeObservation.ObservedDocument)alreadyLoadedDocuments[delegateId];
								if(tmpChildDoc.delegateFor == null) 
								{
									log.Error("Delegate-Kinded ObservedDocument delegating to a not already loaded ObservedDocument loaded [Id: " + tmpChildDoc.id + "]");
									errorMsg = "Es ist ein Verarbeitungsfehler aufgetreten.";
									return ReadResult.Error;
								}
								tmpParentDoc.childDocuments.Add(tmpChildDoc);
							} 
							else 
							{
								tmpParentDoc.childDocuments.Add(tmpChildDoc);
								tmpParentIdList.Add(tmpChildDoc.id);
								alreadyLoadedDocuments.Add(tmpChildDoc.id, tmpChildDoc);
							}
						} // end: loop while(reader.Read())
						reader.Close();
					} // end: loop foreach(curParentIdList)
					curParentIdList = tmpParentIdList;
				} // end: loop for([0,chObs.Depth[)
				try 
				{
					changeObservation = new ChangeObservation(chObsId, chObsName, chObsUrl, chObsDate, chObsCheckCounterTimeStamp, chObsChangeState, chObsDepth, chObsRootDocument);
				} 
				catch(Exception e) 
				{
					log.Error("Inconsistent ChangeObservation found in database during loading it [Id: " + chObsId + "]", e);
					errorMsg = "Es ist eine Datenbankinkonsistenz aufgetreten.";
					return ReadResult.Error;
				}
				errorMsg = null;
				return ReadResult.Ok;
			} 
			catch(Exception e)
			{
				log.Error("Exception occured during load of ChangeObservation to database.", e);
				errorMsg = "Es ist ein Problem beim Datenbankzugriff aufgetreten.";
				return ReadResult.Error;
			} 
			finally 
			{
				try { if(con != null) con.Close(); } 
				catch(Exception ex) 
				{
					log.Warn("Unable to close connection - exception occured.", ex);
				}
			}
		} // end: method LoadChangeObservation

	} // end: class ChangeObservationsMediator
}
