using System;
using System.Data;
using System.Data.SqlClient;
using WebseitenMonitor.DomainObjects.Behaviour;
using System.Collections;
using System.Configuration;
using log4net;

namespace WebseitenMonitor.DataAccess 
{
	public class ObservationsMediator
	{
		private static readonly ILog log = LogManager.GetLogger(typeof(ObservationsMediator));
		private static readonly string DbTableOwnerPrefix = ((string)ConfigurationSettings.AppSettings["DbTableOwnerPrefix"]).Trim();

		public class CategoryObservationItem 
		{

			private int id;
			private ObservationKind kind;
			private string name;
			private Uri url;
			private DateTime date;
			private ChangeStateEnum changeState;
			
			public int Id { get { return this.id; } }
			public ObservationKind Kind { get { return this.kind; } }
			public string Name { get { return this.name; } }
			public Uri Url { get { return this.url; } }
			public DateTime Date { get { return this.date; } set { this.date = value; } }
			public ChangeStateEnum ChangeState { get { return this.changeState; } set { this.changeState = value; } }

			public CategoryObservationItem(
				int id,
				ObservationKind kind,
				string name,
				Uri url,
				DateTime date,
				ChangeStateEnum changeState) 
			{
				this.id = id;
				this.kind = kind;
				this.name = name;
				this.url = url;
				this.date = date;
				this.changeState = changeState;
			}
		} // end: class CategoryObservationItem

		internal enum ReadResult {Ok, InvalidParams, Error}
		internal enum DeleteResult {Ok, InvalidParams, Error}

		internal static ObservationsMediator.DeleteResult DeleteObservation(int id, out string errorMsg) 
		{
			if(id < 0) 
			{
				errorMsg = "Es wurde eine ungltige Identifikationsnummer bergeben.";
				return DeleteResult.InvalidParams;
			}
			// assertion: parameters passed initial checking phase
			string conStr = ConfigurationSettings.AppSettings["DbConnectionString"];
			IDbConnection con = null;
			IDbTransaction trans = null;
			try 
			{
				con = new SqlConnection(conStr);
				con.Open();
				trans = con.BeginTransaction();
				// Delete Observation from Obserations-table - everything else is automatically deleted by
				// cascading delete by the dbms
				IDbCommand cmd = con.CreateCommand();
				cmd.CommandText = "DELETE FROM " + DbTableOwnerPrefix + "Observations WHERE Id = @ID";
				cmd.Parameters.Add(new SqlParameter("@ID", id));
				cmd.Transaction = trans;
				int affectedRows = cmd.ExecuteNonQuery();
				if(affectedRows == 0) 
				{
					log.Error("Delete of Observation didn't affect any Row [Id: " + id + "]");
					trans.Rollback();
					errorMsg = "Es wurde eine nicht existente Identifikationsnummer bergeben.";
					return DeleteResult.InvalidParams;
				} 
				else if(affectedRows != 1) 
				{
					log.Error("Delete of Observation didn't affect exactly one row [Id: " + id + ", #affected: " + affectedRows + "]");
					trans.Rollback();
					errorMsg = "Es ist eine Datenbankinkonsistenz aufgetreten.";
					return DeleteResult.Error;
				}
				trans.Commit();
				errorMsg = null;
				return DeleteResult.Ok;
			}
			catch(Exception e)
			{
				if(trans != null)
					trans.Rollback();
				log.Error("Exception occured writing checked ChangeObservation to database.", e);
				errorMsg = "Es ist ein Problem beim Datenbankzugriff aufgetreten.";
				return DeleteResult.Error;
			} 
			finally 
			{
				try { if(con != null) con.Close(); } 
				catch(Exception ex) 
				{
					log.Warn("Unable to close connection - exception occured.", ex);
				}
			}
		}
		
		/// <summary>
		/// This method reads all Observations corresponding to a Category, which is identified by it's Id, from the Database.
		/// If an invalid Id gets passed the method returns ReadResult.InvalidParams.
		/// If an error occurs during the read from the Database the method returns ReadResult.Error. 
		/// </summary>
		/// <param name="categoryId">Id of Category which Observations should get read.</param>
		/// <param name="observationItemList">Resulting list of CategoryObservationItem-objects each descriping an observation.</param>
		/// <param name="errorMsg">Error message in case of not ReadResult.Ok returned.</param>
		/// <returns>Resulting ReadResult - Ok, InvalidParams or Error.</returns>
		internal static ObservationsMediator.ReadResult ReadCategoryObservationItemListForCategory(
			int categoryId,
			out ArrayList observationItemList,
			out string errorMsg) 
		{
			if(categoryId < 0) 
			{
				observationItemList = null;
				errorMsg = "Ungltige Kategorie-Id bergeben.";
				return ReadResult.InvalidParams;
			}
			ArrayList tmpList = new ArrayList();

			string conStr = ConfigurationSettings.AppSettings["DbConnectionString"];
			IDbConnection con = null;
			try 
			{
				con = new SqlConnection(conStr);
				con.Open();
				IDbCommand cmd = con.CreateCommand();
				cmd.CommandText = "SELECT * FROM " + DbTableOwnerPrefix + "Observations WHERE Fk_Categories = '" + categoryId + "'";
				//cmd.Parameters.Add(new SqlParameter("@CatId",SqlDbType.Int));
				//cmd.Parameters["@CatId"] = CategoryId;
				IDataReader reader = cmd.ExecuteReader();
				while(reader.Read()) 
				{
					// Note DateTime is a struct, so a possible null in the database isn't castable.
					DateTime tmpDate = Observation.InitDateTime;
					if(!(reader["Date"] is DBNull))
						tmpDate = (DateTime)reader["Date"];
					tmpList.Add(new CategoryObservationItem(
						(int)reader["Id"],
						(ObservationKind)reader["Kind"],
						(string)reader["Name"],
						new Uri((string)reader["Url"]),
						tmpDate,
						(ChangeStateEnum)((int)reader["ChangeState"])));
				}
			} 
			catch(Exception e)
			{
				log.Error("Exception occured during read of Observations in Category with Id=" + categoryId, e);
				observationItemList = null;
				errorMsg = "Problem beim Datenbankzugriff aufgetreten.";
				return ReadResult.Error;
			} 
			finally 
			{
				try { if(con != null) con.Close(); } 
				catch(Exception ex) 
				{
					log.Warn("Unable to close connection - exception occured.", ex);
				}
			}
			observationItemList = tmpList;
			errorMsg = null;
			return ReadResult.Ok;
			
		} // end: method ReadCategoryObservationItemListForCategory
	}
}