using System;
using log4net;
using System.Data;
using System.Data.SqlClient;
using System.Data.SqlTypes;
using WebseitenMonitor.DomainObjects.Behaviour;
using System.Configuration;
using System.Collections;

namespace WebseitenMonitor.DataAccess 
{
	public class UserDefinedMessageObservationMediator 
	{
		private static readonly ILog log = LogManager.GetLogger(typeof(UserDefinedMessageObservationMediator));
		private static readonly string DbTableOwnerPrefix = ((string)ConfigurationSettings.AppSettings["DbTableOwnerPrefix"]).Trim();
		
		internal enum UpdateResult {Ok, InvalidParams, Error}
		internal enum WriteResult {Ok, InvalidParams, Error}
		internal enum ReadResult {Ok, InvalidParams, Error}
		
		/// <summary>
		/// This method writes a new initial UserDefinedMessageObservation to the database.
		/// Initial default values: Date			... NULL
		///							ChangeState		... ChangeState.Init
		///							ResultState		... UserDefinedMessageObservation.ResultState.Undefined
		/// </summary>
		/// <param name="name">Observation name. If null or length > 50 InvalidParams is returned.</param>
		/// <param name="url">Url to observe by the Observation. If null InvalidParams is returned.</param>
		/// <param name="categoryId">
		/// Id of the Category to which the Observation should belong.
		/// If less then 0 InvalidParams is returned.
		/// If not existent in database - Error is returned due to database-constraint violation while insert.
		/// </param>
		/// <param name="searchPatternEnabled">if true searchPattern and messageTemplate must not be null.</param>
		/// <param name="searchPattern"></param>
		/// <param name="messageTemplate"></param>
		/// <param name="selectedElementsPositionDescriptor"></param>
		/// <param name="id">Automatically generated Database id of the new ChangeObservation.</param>
		/// <param name="errorMsg">Error Message if a value other than Ok is returned.</param>
		/// <returns>Ok, InvalidParams or Error in case of problem while database access or database constraints.</returns>
		internal static UserDefinedMessageObservationMediator.WriteResult WriteInitUserDefinedMessageObservation(
			string name,
			Uri url,
			int categoryId,
			bool searchPatternEnabled,
			string searchPattern,
			string messageTemplate,
			string selectedElementsPositionDescriptor,
			out int id,
			out string errorMsg) 
		{
			// Check Parameters
			if(name == null || name.Length > 50)
			{
				errorMsg = "Es muss ein Name mit einer max. Lnge von 50 Zeichen angegeben werden.";
				id = -1;
				return WriteResult.InvalidParams;
			}
			if(url == null) 
			{
				errorMsg = "Es muss eine Url angegeben werden.";
				id = -1;
				return WriteResult.InvalidParams;
			}
			if(categoryId < 0) 
			{
				errorMsg = "Es muss ein gltiger Kategoriebezug angegeben werden.";
				id = -1;
				return WriteResult.InvalidParams;
			}
			if(searchPatternEnabled) 
			{
				if(searchPattern == null || messageTemplate == null) 
				{
					errorMsg = "Es muss ein konsistentes SearchPattern und MessageTemplate angegeben werden.";
					id = -1;
					return WriteResult.InvalidParams;
				}
			}
			// assertion: parameters passed initial checking phase
			string conStr = ConfigurationSettings.AppSettings["DbConnectionString"];
			IDbConnection con = null;
			IDbTransaction trans = null;
			try
			{
				con = new SqlConnection(conStr);
				con.Open();
				trans = con.BeginTransaction(); // Note: insertion in Observations and UserDefinedMessageObservations in a transaction
				// First part: insert general properties into Observations table
				IDbCommand cmd = con.CreateCommand();                                                             
				cmd.CommandText = "INSERT INTO " + DbTableOwnerPrefix + "Observations (Kind,Name,Url,ChangeState,Fk_Categories) VALUES (@KIND,@NAME,@URL,@CHANGESTATE,@FK_CATEGORIES);" +
					"SELECT SCOPE_IDENTITY() AS [SCOPE_IDENTITY]";
				cmd.Parameters.Add(new SqlParameter("@KIND", (int)ObservationKind.UserDefinedMessageObservation));
				cmd.Parameters.Add(new SqlParameter("@NAME", name));
				cmd.Parameters.Add(new SqlParameter("@URL", url.ToString()));
				cmd.Parameters.Add(new SqlParameter("@CHANGESTATE",(int)ChangeStateEnum.Init));
				cmd.Parameters.Add(new SqlParameter("@FK_CATEGORIES",categoryId));
				cmd.Transaction = trans;
				int insertedObservationId = Convert.ToInt32((decimal)cmd.ExecuteScalar());
				if(insertedObservationId < 0) 
				{
					trans.Rollback();
					errorMsg = "Es ist ein inkonsistenter Datenbanknderungsversuch aufgetreten.";
					id = -1;
					return WriteResult.Error;
				}
				// Second part: insert UserDefinedMessageObservations specific properties
				cmd = con.CreateCommand();
				cmd.CommandText = "INSERT INTO " + DbTableOwnerPrefix + "UserDefinedMessageObservations (Id,SearchPatternEnabled,SearchPattern,MessageTemplate,SelectedElementsPositionDescriptor,ResultState) VALUES(@ID,@SEARCHPATTERNENABLED,@SEARCHPATTERN,@MESSAGETEMPLATE,@SELECTEDELEMENTSPOSITIONDESCRIPTOR,@RESULTSTATE);";
				cmd.Parameters.Add(new SqlParameter("@ID",insertedObservationId));
				cmd.Parameters.Add(new SqlParameter("@SEARCHPATTERNENABLED", searchPatternEnabled));
				if(searchPattern == null)
					cmd.Parameters.Add(new SqlParameter("@SEARCHPATTERN", DBNull.Value));
				else
					cmd.Parameters.Add(new SqlParameter("@SEARCHPATTERN", searchPattern));
				if(messageTemplate == null)
					cmd.Parameters.Add(new SqlParameter("@MESSAGETEMPLATE", DBNull.Value));
				else
					cmd.Parameters.Add(new SqlParameter("@MESSAGETEMPLATE", messageTemplate));
				if(selectedElementsPositionDescriptor == null)
					cmd.Parameters.Add(new SqlParameter("@SELECTEDELEMENTSPOSITIONDESCRIPTOR", DBNull.Value));
				else
					cmd.Parameters.Add(new SqlParameter("@SELECTEDELEMENTSPOSITIONDESCRIPTOR", selectedElementsPositionDescriptor));
				cmd.Parameters.Add(new SqlParameter("@RESULTSTATE", (int)UserDefinedMessageObservation.ResultStateEnum.Undefined));
				cmd.Transaction = trans;
				cmd.ExecuteNonQuery();
				trans.Commit();
				id = insertedObservationId;
				errorMsg = null;
				return WriteResult.Ok;
			} 
			catch(Exception e) 
			{
				if(trans != null)
					trans.Rollback();
				log.Error("Exception occured during writing new UserDefinedMessageObservation to database", e);
				errorMsg = "Es ist ein Problem beim Datenbankzugriff aufgetreten.";
				id = -1;
				return WriteResult.Error;
			}
			finally 
			{
				try { if(con != null) con.Close(); } 
				catch(Exception ex) 
				{
					log.Warn("Unable to close connection - exception occured.", ex);
				}
			}
		} // end: WriteNewUserDefinedMessageObservation
		
		/// <summary>
		/// This method writes a new UserDefinedMessageObservation to the database.
		/// It's intention is to write a UserDefinedMessageObservation object as returned by
		/// UserDefinedMessageObservation.CreateUserDefinedMessageObservation(), but it can also
		/// be used for an arbitrary but consistent UserDefinedMessageObservation object.
		/// It is important to note that, since the method write a new one to database, the id of the passed
		/// UserDefinedMessageObservation object must be -1 which indicates that it has no database-id. The
		/// database id will be generated by the insertion in the database and the id-property of the
		/// UserDefinedMessageObservation object will get changed accordingly.
		/// </summary>
		/// <param name="categoryId"></param>
		/// <param name="udm"></param>
		/// <param name="errorMsg"></param>
		/// <returns></returns>
		internal static UserDefinedMessageObservationMediator.WriteResult WriteNewUserDefinedMessageObservation(
			int categoryId,
			ref UserDefinedMessageObservation udm,
			out string errorMsg) 
		{
			log.Debug("WriteNewUserDefinedMessageObservation called!");
			// Check Parameters
			if(udm == null) 
			{
				errorMsg = "Es muss ein entsprechendes Beobachtungsobjekt bergeben werden.";
				return WriteResult.InvalidParams;
			}
			if(udm.Id != -1) 
			{
				errorMsg = "Das bergebene Beobachtungsobjekt darf noch keine gltige id haben.";
				return WriteResult.InvalidParams;
			}
			if(udm.Name == null || udm.Name.Length > 50)
			{
				errorMsg = "Das Beobachtungsobjekt muss einen Namen mit einer max. Lnge von 50 Zeichen haben.";
				return WriteResult.InvalidParams;
			}
			if(udm.Url == null) 
			{
				errorMsg = "Das Beobachtungsobjekt muss eine Url haben.";
				return WriteResult.InvalidParams;
			}
			if(categoryId < 0) 
			{
				errorMsg = "Es muss ein gltiger Kategoriebezug angegeben werden.";
				return WriteResult.InvalidParams;
			}
			if(udm.SearchPatternEnabled) 
			{
				if(udm.SearchPattern == null || udm.MessageTemplate == null) 
				{
					errorMsg = "Das Beobachtungsobjekt muss ein konsistentes SearchPattern und MessageTemplate haben.";
					return WriteResult.InvalidParams;
				}
			}
			// assertion: parameters passed initial checking phase
			string conStr = ConfigurationSettings.AppSettings["DbConnectionString"];
			IDbConnection con = null;
			IDbTransaction trans = null;
			try
			{
				con = new SqlConnection(conStr);
				con.Open();
				trans = con.BeginTransaction(); // Note: insertion in Observations and UserDefinedMessageObservations in a transaction
				// First part: insert general properties into Observations table
				IDbCommand cmd = con.CreateCommand();                                                             
				cmd.CommandText = "INSERT INTO " + DbTableOwnerPrefix + "Observations (Kind,Name,Url,Date,CheckCounterTimeStamp,ChangeState,Fk_Categories) VALUES (@KIND,@NAME,@URL,@DATE,@CHECKCOUNTER,@CHANGESTATE,@FK_CATEGORIES);" +
					"SELECT SCOPE_IDENTITY() AS [SCOPE_IDENTITY]";
				cmd.Parameters.Add(new SqlParameter("@KIND", (int)ObservationKind.UserDefinedMessageObservation));
				cmd.Parameters.Add(new SqlParameter("@NAME", udm.Name));
				cmd.Parameters.Add(new SqlParameter("@URL", udm.Url.ToString()));
				cmd.Parameters.Add(new SqlParameter("@DATE", udm.Date));
				cmd.Parameters.Add(new SqlParameter("@CHECKCOUNTER", udm.CheckCounterTimeStamp));
				cmd.Parameters.Add(new SqlParameter("@CHANGESTATE",(int)udm.ChangeState));
				cmd.Parameters.Add(new SqlParameter("@FK_CATEGORIES",categoryId));
				cmd.Transaction = trans;
				int insertedObservationId = Convert.ToInt32((decimal)cmd.ExecuteScalar());
				if(insertedObservationId < 0) 
				{
					trans.Rollback();
					errorMsg = "Es ist ein inkonsistenter Datenbanknderungsversuch aufgetreten.";
					return WriteResult.Error;
				}
				udm.SetId(insertedObservationId);
				// Second part: insert UserDefinedMessageObservations specific properties
				cmd = con.CreateCommand();
				cmd.CommandText = "INSERT INTO " + DbTableOwnerPrefix + "UserDefinedMessageObservations (Id,SearchPatternEnabled,SearchPattern,MessageTemplate,SelectedElementsPositionDescriptor,ResultState,ResultText) VALUES(@ID,@SEARCHPATTERNENABLED,@SEARCHPATTERN,@MESSAGETEMPLATE,@SELECTEDELEMENTSPOSITIONDESCRIPTOR,@RESULTSTATE,@RESULTTEXT);";
				cmd.Parameters.Add(new SqlParameter("@ID", udm.Id));
				cmd.Parameters.Add(new SqlParameter("@SEARCHPATTERNENABLED", udm.SearchPatternEnabled));
				if(udm.SearchPattern == null)
					cmd.Parameters.Add(new SqlParameter("@SEARCHPATTERN", DBNull.Value));
				else
					cmd.Parameters.Add(new SqlParameter("@SEARCHPATTERN", udm.SearchPattern));
				if(udm.MessageTemplate == null)
					cmd.Parameters.Add(new SqlParameter("@MESSAGETEMPLATE", DBNull.Value));
				else
					cmd.Parameters.Add(new SqlParameter("@MESSAGETEMPLATE", udm.MessageTemplate));
				if(udm.SelectedElementsPositionDescriptor == null)
					cmd.Parameters.Add(new SqlParameter("@SELECTEDELEMENTSPOSITIONDESCRIPTOR", DBNull.Value));
				else
					cmd.Parameters.Add(new SqlParameter("@SELECTEDELEMENTSPOSITIONDESCRIPTOR", udm.SelectedElementsPositionDescriptor));
				cmd.Parameters.Add(new SqlParameter("@RESULTSTATE", (int)udm.ResultState));
				if(udm.ResultText == null)
					cmd.Parameters.Add(new SqlParameter("@RESULTTEXT", DBNull.Value));
				else
					cmd.Parameters.Add(new SqlParameter("@RESULTTEXT", udm.ResultText));
				cmd.Transaction = trans;
				cmd.ExecuteNonQuery();
				trans.Commit();
				errorMsg = null;
				return WriteResult.Ok;
			} 
			catch(Exception e) 
			{
				if(trans != null)
					trans.Rollback();
				log.Error("Exception occured during writing new UserDefinedMessageObservation to database", e);
				errorMsg = "Es ist ein Problem beim Datenbankzugriff aufgetreten.";
				return WriteResult.Error;
			}
			finally 
			{
				try { if(con != null) con.Close(); } 
				catch(Exception ex) 
				{
					log.Warn("Unable to close connection - exception occured.", ex);
				}
			}
		} // end: WriteNewUserDefinedMessageObservation
		
		internal static UserDefinedMessageObservationMediator.ReadResult LoadUserDefinedMessageObservation(
			int observationId,
			out UserDefinedMessageObservation udmObservation,
			out string errorMsg) 
		{
			udmObservation = null;
			errorMsg = null;
			if(observationId < 0) 
			{
				errorMsg = "Es wurde eine ungltige Identifikationsnummer bergeben.";
				return ReadResult.InvalidParams;
			}
			// attributes of the UserDefinedMessageObservation to load:
			int udmObsId = -1;
			string udmObsName = null;
			Uri udmObsUrl = null;
			DateTime udmObsDate = Observation.InitDateTime;
			int udmObsCheckCounterTimeStamp = -1;
			ChangeStateEnum udmObsChangeState = ChangeStateEnum.Error;
			bool udmObsSearchPatternEnabled = false;
			string udmObsSearchPattern = null;
			string udmObsMessageTemplate = null;
			string udmObsSelectedElementsPositionDescriptor = null;
			UserDefinedMessageObservation.ResultStateEnum udmObsResultState = UserDefinedMessageObservation.ResultStateEnum.Undefined;
			string udmObsResultText = null;

			// assertion: parameters passed initial checking phase
			string conStr = ConfigurationSettings.AppSettings["DbConnectionString"];
			IDbConnection con = null;
			try 
			{
				con = new SqlConnection(conStr);
				con.Open();
				// Step 1: load general properties from Observations-table
				IDbCommand cmd = con.CreateCommand();
				cmd.CommandText = "SELECT * FROM " + DbTableOwnerPrefix + "Observations WHERE Id = @ID";
				cmd.Parameters.Add(new SqlParameter("@ID", observationId));
				IDataReader reader = cmd.ExecuteReader();
				int rowCount = 0;
				while(reader.Read()) 
				{
					if(rowCount > 0) 
					{
						log.Error("Id of requested Observation seems to be not unique, multiple rows found.");
						errorMsg = "Es ist eine Datenbankinkonsistenz aufgetreten.";
						return ReadResult.Error;
					}
					udmObsId = (int)reader["Id"];
					if((ObservationKind)reader["Kind"] != ObservationKind.UserDefinedMessageObservation) 
					{
						log.Error("Wrong Observation-Kind for load of ChangeObservation found in database [Id: " + udmObsId + "]");
						errorMsg = "bergebene Identifikationsnummer kennzeichnet eine andere Art von Beobachtungskanal.";
						return ReadResult.InvalidParams;
					}
					if(reader["Date"] is DBNull)
						udmObsDate = Observation.InitDateTime;
					else
						udmObsDate = (DateTime)reader["Date"];
					if(reader["CheckCounterTimeStamp"] is DBNull)
						udmObsCheckCounterTimeStamp = -1;
					else
						udmObsCheckCounterTimeStamp = (int)reader["CheckCounterTimeStamp"];
					udmObsName = (string)reader["Name"];
					udmObsUrl = new Uri((string)reader["Url"]);
					udmObsChangeState = (ChangeStateEnum)reader["ChangeState"];
					rowCount++;
				}
				reader.Close();
				if(rowCount == 0) 
				{
					log.Error("No corresponding row, to load requested UserDefinedMessageObservation, found in Observations table [Id: " + observationId + "].");
					errorMsg = "Es konnte kein entsprechender Beobachtungskanal gefunden werden.";
					return ReadResult.InvalidParams;
				}
				// Step 2: load UserDefinedMessageObservation specific properties from UserDefinedMessageObservations-table
				cmd = con.CreateCommand();
				cmd.CommandText = "SELECT * FROM " + DbTableOwnerPrefix + "UserDefinedMessageObservations WHERE Id = @ID";
				cmd.Parameters.Add(new SqlParameter("@ID", udmObsId));
				reader = cmd.ExecuteReader();
				rowCount = 0;
				while(reader.Read()) 
				{
					if(rowCount > 0) 
					{
						log.Error("UserDefinedMessageObservations-table entries not unique by Id, multiple occurence detected [Id: " + udmObsId + "]");
						errorMsg = "Es ist eine Datenbankinkonsistenz aufgetreten.";
						return ReadResult.Error;
					}
					udmObsSearchPatternEnabled = (bool)reader["SearchPatternEnabled"];
					if(reader["SearchPattern"] is DBNull)
						udmObsSearchPattern = null;
					else 
						udmObsSearchPattern = (string)reader["SearchPattern"];
					if(reader["MessageTemplate"] is DBNull)
						udmObsMessageTemplate = null;
					else 
						udmObsMessageTemplate = (string)reader["MessageTemplate"];
					if(reader["SelectedElementsPositionDescriptor"] is DBNull)
						udmObsSelectedElementsPositionDescriptor = null;
					else 
						udmObsSelectedElementsPositionDescriptor = (string)reader["SelectedElementsPositionDescriptor"];
					if(reader["ResultText"] is DBNull)
						udmObsResultText = null;
					else 
						udmObsResultText = (string)reader["ResultText"];
					udmObsResultState = (UserDefinedMessageObservation.ResultStateEnum)reader["ResultState"];
					rowCount++;
				}
				reader.Close();
				if(rowCount == 0) 
				{
					log.Error("No corresponding UserDefinedMessageObservations-table entry found for already found Observations-table entry [Id: " + udmObsId + "]");
					errorMsg = "Es ist eine Datenbankinkonsistenz aufgetreten.";
					return ReadResult.Error;
				}
				try 
				{
					udmObservation = new UserDefinedMessageObservation(
						udmObsId,
						udmObsName,
						udmObsUrl,
						udmObsDate,
						udmObsCheckCounterTimeStamp, 
						udmObsChangeState,
						udmObsSearchPatternEnabled,
						udmObsSearchPattern,
						udmObsMessageTemplate,
						udmObsSelectedElementsPositionDescriptor,
						udmObsResultText,
						udmObsResultState);
				} 
				catch(Exception e) 
				{
					log.Error("Inconsistent UserDefinedObservation found in database during loading it [Id: " + udmObsId + "]", e);
					errorMsg = "Es ist eine Datenbankinkonsistenz aufgetreten.";
					return ReadResult.Error;
				}
				errorMsg = null;
				return ReadResult.Ok;
			} 
			catch(Exception e)
			{
				log.Error("Exception occured during load of UserDefinedObservation to database.", e);
				errorMsg = "Es ist ein Problem beim Datenbankzugriff aufgetreten.";
				return ReadResult.Error;
			} 
			finally 
			{
				try { if(con != null) con.Close(); } 
				catch(Exception ex) 
				{
					log.Warn("Unable to close connection - exception occured.", ex);
				}
			}
		} // end: method LoadUserDefinedObservation

		internal static UserDefinedMessageObservationMediator.UpdateResult UpdateChangedPropertiesUserDefinedMessageObservation(
			UserDefinedMessageObservation udmObs,
			out string errorMsg) 
		{
			// Check Parameters
			if(udmObs == null) 
			{
				errorMsg = "Es muss ein passender Beobachtungskanal angegeben werden.";
				return UpdateResult.InvalidParams;
			}
			if(udmObs.Name == null || udmObs.Name.Length > 50) 
			{
				errorMsg = "Der Beobachtungskanal muss einen Namen mit einer max. Lnge von 50 Zeichen haben.";
				return UpdateResult.InvalidParams;
			}
			if(udmObs.Url == null) 
			{
				errorMsg = "Der Beobachtungskanal muss eine Url enthalten.";
				return UpdateResult.InvalidParams;
			}
			// assertion: parameters passed initial checking phase
			string conStr = ConfigurationSettings.AppSettings["DbConnectionString"];
			IDbConnection con = null;
			IDbTransaction trans = null;
			try 
			{
				con = new SqlConnection(conStr);
				con.Open();
				trans = con.BeginTransaction();
				// Step 1: update general properties in Observations table
				IDbCommand cmd = con.CreateCommand();
				cmd.CommandText = "UPDATE " + DbTableOwnerPrefix + "Observations SET Kind = @KIND, Name = @NAME, Url = @URL, Date = @DATE , CheckCounterTimeStamp = @TIMESTAMP , ChangeState = @STATE WHERE Id = @ID";
				cmd.Parameters.Add(new SqlParameter("@KIND", (int)ObservationKind.UserDefinedMessageObservation));
				cmd.Parameters.Add(new SqlParameter("@NAME", udmObs.Name));
				cmd.Parameters.Add(new SqlParameter("@URL", udmObs.Url.ToString()));
				if(udmObs.Date.Equals(Observation.InitDateTime))
					cmd.Parameters.Add(new SqlParameter("@DATE", DBNull.Value));
				else
					cmd.Parameters.Add(new SqlParameter("@DATE", udmObs.Date));
				if(udmObs.CheckCounterTimeStamp == -1)
					cmd.Parameters.Add(new SqlParameter("@TIMESTAMP", DBNull.Value));
				else
					cmd.Parameters.Add(new SqlParameter("@TIMESTAMP", udmObs.CheckCounterTimeStamp));
				cmd.Parameters.Add(new SqlParameter("@STATE", (int)udmObs.ChangeState));
				cmd.Parameters.Add(new SqlParameter("@ID", udmObs.Id));
				cmd.Transaction = trans;
				int updateCount = cmd.ExecuteNonQuery();
				if(updateCount != 1) 
				{
					log.Error("Update of Observations table for properties changed UserDefinedMessageObservation didn't affect exactly one row [#affected: " + updateCount + "].");
					trans.Rollback();
					errorMsg = "Es ist eine Datenbankinkonsistenz aufgetreten.";
					return UpdateResult.Error;
				}
				// Step 2: update UserDefinedMessageObservation specific properties in UserDefinedMessageObservations table
				cmd = con.CreateCommand();
				cmd.CommandText = "UPDATE " + DbTableOwnerPrefix + "UserDefinedMessageObservations SET SearchPatternEnabled = @SPE, SearchPattern = @SP, MessageTemplate = @MT, SelectedElementsPositionDescriptor = @SEPD, ResultState = @RS, ResultText = @RT WHERE Id = @ID";
				cmd.Parameters.Add(new SqlParameter("@SPE", udmObs.SearchPatternEnabled));
				if(udmObs.SearchPattern == null)
					cmd.Parameters.Add(new SqlParameter("@SP", DBNull.Value));
				else
					cmd.Parameters.Add(new SqlParameter("@SP", udmObs.SearchPattern));
				if(udmObs.MessageTemplate == null)
					cmd.Parameters.Add(new SqlParameter("@MT", DBNull.Value));
				else
					cmd.Parameters.Add(new SqlParameter("@MT", udmObs.MessageTemplate));
				if(udmObs.SelectedElementsPositionDescriptor == null)
					cmd.Parameters.Add(new SqlParameter("@SEPD", DBNull.Value));
				else
					cmd.Parameters.Add(new SqlParameter("@SEPD", udmObs.SelectedElementsPositionDescriptor));
				cmd.Parameters.Add(new SqlParameter("@RS", (int)udmObs.ResultState));
				if(udmObs.ResultText == null)
					cmd.Parameters.Add(new SqlParameter("@RT", DBNull.Value));
				else
					cmd.Parameters.Add(new SqlParameter("@RT", udmObs.ResultText));
				cmd.Parameters.Add(new SqlParameter("@ID", udmObs.Id));
				cmd.Transaction = trans;
				updateCount = cmd.ExecuteNonQuery();
				if(updateCount != 1) 
				{
					log.Error("Update of UserDefinedMessageObservations table didn't affect exactly one row [#affected: " + updateCount + "].");
					trans.Rollback();
					errorMsg = "Es ist eine Datenbankinkonsistenz aufgetreten.";
					return UpdateResult.Error;
				}
				trans.Commit();
				errorMsg = null;
				return UpdateResult.Ok;
			} 
			catch(Exception e)
			{
				if(trans != null)
					trans.Rollback();
				log.Error("Exception occured during updating changed properties UserDefinedMessageObservation to database.", e);
				errorMsg = "Es ist ein Problem beim Datenbankzugriff aufgetreten.";
				return UpdateResult.Error;
			} 
			finally 
			{
				try { if(con != null) con.Close(); } 
				catch(Exception ex) 
				{
					log.Warn("Unable to close connection - exception occured.", ex);
				}
			}
		} // end: method UpdateChangedPropertiesUserDefinedMessageObservation

		internal static UserDefinedMessageObservationMediator.UpdateResult UpdateCheckedUserDefinedMessageObservation(
			UserDefinedMessageObservation udmObs,
			out string errorMsg) 
		{
			if(udmObs == null) 
			{
				errorMsg = "Es muss ein passender Beobachtungskanal angegeben werden.";
				return UpdateResult.InvalidParams;
			}
			// assertion: parameters passed initial checking phase
			string conStr = ConfigurationSettings.AppSettings["DbConnectionString"];
			IDbConnection con = null;
			IDbTransaction trans = null;
			try 
			{
				con = new SqlConnection(conStr);
				con.Open();
				trans = con.BeginTransaction();
				// Step 1: update general properties in Observations table
				IDbCommand cmd = con.CreateCommand();
				cmd.CommandText = "UPDATE " + DbTableOwnerPrefix + "Observations SET Date = @DATE , CheckCounterTimeStamp = @TIMESTAMP , ChangeState = @STATE WHERE Id = @ID";
				cmd.Parameters.Add(new SqlParameter("@DATE", udmObs.Date));
				cmd.Parameters.Add(new SqlParameter("@TIMESTAMP", udmObs.CheckCounterTimeStamp));
				cmd.Parameters.Add(new SqlParameter("@STATE", (int)udmObs.ChangeState));
				cmd.Parameters.Add(new SqlParameter("@ID", udmObs.Id));
				cmd.Transaction = trans;
				int updateCount = cmd.ExecuteNonQuery();
				if(updateCount != 1) 
				{
					log.Error("Update of Observations table for newly checked UserDefinedMessageObservation didn't affect exactly one row [#affected: " + updateCount + "].");
					trans.Rollback();
					errorMsg = "Es ist eine Datenbankinkonsistenz aufgetreten.";
					return UpdateResult.Error;
				}
				// Step 2: update detail properties in UserDefinedMessageObservation table
				cmd = con.CreateCommand();
				cmd.CommandText = "UPDATE " + DbTableOwnerPrefix + "UserDefinedMessageObservations SET ResultState = @RESSTATE, ResultText = @RESTEXT WHERE Id = @ID";
				cmd.Parameters.Add(new SqlParameter("@RESSTATE", (int)udmObs.ResultState));
				if(udmObs.ResultText == null)
					cmd.Parameters.Add(new SqlParameter("@RESTEXT", DBNull.Value));
				else
					cmd.Parameters.Add(new SqlParameter("@RESTEXT", udmObs.ResultText));
				cmd.Parameters.Add(new SqlParameter("@ID", udmObs.Id));
				cmd.Transaction = trans;
				updateCount = cmd.ExecuteNonQuery();
				if(updateCount != 1) 
				{
					log.Error("Update of UserDefinedMessageObservations table for newly checked UserDefinedMessageObservation didn't affect exactly one row [#affected: " + updateCount + "].");
					trans.Rollback();
					errorMsg = "Es ist eine Datenbankinkonsistenz aufgetreten.";
					return UpdateResult.Error;
				}
				trans.Commit();
				errorMsg = null;
				return UpdateResult.Ok;
			}
			catch(Exception e)
			{
				if(trans != null)
					trans.Rollback();
				log.Error("Exception occured during update of newly checked UserDefinedMessageObservation in database.", e);
				errorMsg = "Es ist ein Problem beim Datenbankzugriff aufgetreten.";
				return UpdateResult.Error;
			} 
			finally 
			{
				try { if(con != null) con.Close(); } 
				catch(Exception ex) 
				{
					log.Warn("Unable to close connection - exception occured.", ex);
				}
			}
		} // end: method UpdateCheckedUserDefinedMessageObservation
	}
}