using System;
using log4net;
using System.Configuration;
using System.Data;
using System.Data.SqlClient;
using System.Collections;
using System.Web.Security;

namespace WebseitenMonitor.DataAccess 
{
	public class UsersMediator 
	{
		private static readonly ILog log = LogManager.GetLogger(typeof(UsersMediator));
		private static readonly string DbTableOwnerPrefix = ((string)ConfigurationSettings.AppSettings["DbTableOwnerPrefix"]).Trim();
		
		internal enum WriteResult {Ok, InvalidParams, Error}
		internal enum ReadResult {Ok, InvalidParams, Error}
		internal enum UpdateResult {Ok, InvalidParams, Error}
		internal enum DeleteResult {Ok, InvalidParams, Error}
		
		internal static UsersMediator.WriteResult CreateNewUser(string username, string passwd, string role, out string errorMsg) 
		{
			if(username == null || passwd == null || role == null) 
			{
				log.Error("Tried to create new user with invalid parameters : (" + username + "," + passwd + "," + role + ")");
				errorMsg = "Der Datenbankzugriff wurde ungltig parametrisiert.";
				return UsersMediator.WriteResult.InvalidParams;
			}
			string passwdHash = FormsAuthentication.HashPasswordForStoringInConfigFile(passwd,"MD5");
			if(passwdHash.Length > 50) 
			{
				log.Error("Tried to create new user with invalid parameters - hash of password is longer than 50 signs!");
				errorMsg = "Der Datenbankzugriff wurde ungltig parametrisiert.";
				return UsersMediator.WriteResult.InvalidParams;
			}
			string conStr = ConfigurationSettings.AppSettings["DbConnectionString"];
			IDbConnection con = null;
			try 
			{
				con = new SqlConnection(conStr);
				con.Open();
				IDbCommand cmd = con.CreateCommand();
				cmd.CommandText = "INSERT INTO " + DbTableOwnerPrefix + "Users (LoginName, PwdHash, Role) VALUES(@LoginName,@PwdHash,@Role);";
				cmd.Parameters.Add(new SqlParameter("@LoginName",username));
				cmd.Parameters.Add(new SqlParameter("@PwdHash", passwdHash));
				cmd.Parameters.Add(new SqlParameter("@Role", role));
				cmd.ExecuteNonQuery();
				errorMsg = null;
				return WriteResult.Ok;
			} 
			catch(Exception e) 
			{
				log.Error("Exception occured during writing new User to database", e);
				errorMsg = "Es ist ein Problem beim Datenbankzugriff aufgetreten.";
				return WriteResult.Error;
			}
			finally 
			{
				try { if(con != null) con.Close(); } 
				catch(Exception ex) 
				{
					log.Warn("Unable to close connection - exception occured.", ex);
				}
			}
		}

		internal static UsersMediator.UpdateResult UpdateUser(int userId, string username, string passwd, out string errorMsg) 
		{
			if(username == null || passwd == null) 
			{
				log.Error("Tried to update user with invalid parameters: (" + userId + "," + username + "," + passwd + ")");
				errorMsg = "Der Datenbankzugriff wurde ungltig parametrisiert.";
				return UsersMediator.UpdateResult.InvalidParams;
			}
			string passwdHash = FormsAuthentication.HashPasswordForStoringInConfigFile(passwd,"md5");
			if(passwdHash.Length > 50) 
			{
				log.Error("Tried to update user with invalid parameters - hash of password is longer than 50 signs!");
				errorMsg = "Der Datenbankzugriff wurde ungltig parametrisiert.";
				return UsersMediator.UpdateResult.InvalidParams;
			}
			string conStr = ConfigurationSettings.AppSettings["DbConnectionString"];
			IDbConnection con = null;
			try 
			{
				con = new SqlConnection(conStr);
				con.Open();
				IDbCommand cmd = con.CreateCommand();
				cmd.CommandText = "UPDATE " + DbTableOwnerPrefix + "Users SET LoginName=@LoginName, PwdHash=@PwdHash WHERE Id = @ID";
				cmd.Parameters.Add(new SqlParameter("@LoginName", username));
				cmd.Parameters.Add(new SqlParameter("@PwdHash", passwdHash));
				cmd.Parameters.Add(new SqlParameter("@ID", userId));
				int updateCount = cmd.ExecuteNonQuery();
				if(updateCount != 1) 
				{
					log.Error("Update of Users table for user update didn't affect exactly one row (#affected: " + updateCount + "].");
					errorMsg = "Es ist eine Datenbankinkonsistenz aufgetreten.";
					return UpdateResult.Error;
				}
				errorMsg = null;
				return UpdateResult.Ok;
			}
			catch(Exception e)
			{
				log.Error("Exception occured during update of user Users with Id=" + userId, e);
				errorMsg = "Problem beim Datenbankzugriff aufgetreten.";
				return UpdateResult.Error;
			} 
			finally 
			{
				try { if(con != null) con.Close(); } 
				catch(Exception ex) 
				{
					log.Warn("Unable to close connection - exception occured.", ex);
				}
			}
			
		}

		internal static UsersMediator.DeleteResult DeleteUser(int userId, out string errorMsg) 
		{
			string conStr = ConfigurationSettings.AppSettings["DbConnectionString"];
			IDbConnection con = null;
			try 
			{
				con = new SqlConnection(conStr);
				con.Open();
				// Delete user from Users-table - everything related is automatically deleted by
				// cascading delete by the dbms
				IDbCommand cmd = con.CreateCommand();
				cmd.CommandText = "DELETE FROM " + DbTableOwnerPrefix + "Users WHERE Id = @ID";
				cmd.Parameters.Add(new SqlParameter("@ID", userId));
				int affectedRows = cmd.ExecuteNonQuery();
				if(affectedRows == 0) 
				{
					log.Error("Delete of user in Users-table didn't affect any Row [Id: " + userId + "]");
					errorMsg = "Der Benutzer konnte nicht korrekt identifiziert werden.";
					return DeleteResult.InvalidParams;
				} 
				else if(affectedRows != 1) 
				{
					log.Error("Delete of user in Users-table didn't affect exactly one row [Id: " + userId + ", #affected: " + affectedRows + "]");
					errorMsg = "Es ist eine Datenbankinkonsistenz aufgetreten.";
					return DeleteResult.Error;
				}
				errorMsg = null;
				return DeleteResult.Ok;
			}
			catch(Exception e)
			{
				log.Error("Exception occured during try to delete user from database.", e);
				errorMsg = "Es ist ein Problem beim Datenbankzugriff aufgetreten.";
				return DeleteResult.Error;
			} 
			finally 
			{
				try { if(con != null) con.Close(); } 
				catch(Exception ex) 
				{
					log.Warn("Unable to close connection - exception occured.", ex);
				}
			}
		}

		internal static UsersMediator.UpdateResult UpdateUsersCheckCounter(int userId, int newCheckCounter, out string errorMsg) 
		{
			if(userId < 0)
			{
				log.Error("Tried to update Userse CheckCounter with invalid userid of : " + userId);
				errorMsg = "Es wurde eine ungltige Userid bergeben.";
				return UsersMediator.UpdateResult.InvalidParams;
			}
			string conStr = ConfigurationSettings.AppSettings["DbConnectionString"];
			IDbConnection con = null;
			try 
			{
				con = new SqlConnection(conStr);
				con.Open();
				IDbCommand cmd = con.CreateCommand();
				cmd.CommandText = "UPDATE " + DbTableOwnerPrefix + "Users SET CheckCounter = @CC WHERE Id = @ID";
				cmd.Parameters.Add(new SqlParameter("@CC", newCheckCounter));
				cmd.Parameters.Add(new SqlParameter("@ID", userId));
				int updateCount = cmd.ExecuteNonQuery();
				if(updateCount != 1) 
				{
					log.Error("Update of Users table for CheckCounter update didn't affect exactly one row (#affected: " + updateCount + "].");
					errorMsg = "Es ist eine Datenbankinkonsistenz aufgetreten.";
					return UpdateResult.Error;
				}
				errorMsg = null;
				return UpdateResult.Ok;
			}
			catch(Exception e)
			{
				log.Error("Exception occured during update of CheckCounter in Users with Id=" + userId, e);
				errorMsg = "Problem beim Datenbankzugriff aufgetreten.";
				return UpdateResult.Error;
			} 
			finally 
			{
				try { if(con != null) con.Close(); } 
				catch(Exception ex) 
				{
					log.Warn("Unable to close connection - exception occured.", ex);
				}
			}
		} // end: method Update UsersCheckCounter
		
		internal static UsersMediator.ReadResult LoadUsers(string[] roles, out Hashtable users, out string errorMsg) 
		{
			Hashtable tmp = new Hashtable();
			string conStr = ConfigurationSettings.AppSettings["DbConnectionString"];
			IDbConnection con = null;
			try 
			{
				con = new SqlConnection(conStr);
				con.Open();
				IDbCommand cmd = con.CreateCommand();
				cmd.CommandText = "SELECT * FROM " + DbTableOwnerPrefix + "Users";
				IDataReader reader = cmd.ExecuteReader();
				int rowCount = 0;
				while(reader.Read()) 
				{
					bool userInRole = false;
					// check if user is in an expected role
					string roleName = (string)reader["Role"];
					if(roles!= null) 
					{
						for(int i=0; i<roles.Length; i++) 
						{ 
							if(roles[i] == null) 
							{
								if(roleName == null) 
								{
									userInRole = true;
									break;
								}
							} 
							else 
							{
								if(roles[i]==roleName) 
								{
									userInRole = true;
									break;
								}
							}
						}
					}
					if(userInRole)
						tmp.Add((int)reader["ID"], (string)reader["LoginName"]);
					rowCount++;
				}
				if(rowCount <= 0) 
				{
					log.Error("No User found during read of Users.");
					users = null;
					errorMsg = "Es konnte kein entsprechender Benutzer gefunden werden.";
					return UsersMediator.ReadResult.InvalidParams;
				}
				users = tmp;
				errorMsg = null;
				return UsersMediator.ReadResult.Ok;
			} 
			catch(Exception e)
			{
				log.Error("Exception occured during read of Users", e);
				users = null;
				errorMsg = "Problem beim Datenbankzugriff aufgetreten.";
				return ReadResult.Error;
			} 
			finally 
			{
				try { if(con != null) con.Close(); } 
				catch(Exception ex) 
				{
					log.Warn("Unable to close connection - exception occured.", ex);
				}
			}
		}

		internal static UsersMediator.ReadResult ReadUsersCheckCounter(int userId, out int checkCounter, out string errorMsg) 
		{
			if(userId < 0) 
			{
				log.Error("Tryed to read Users CheckCounter with invalid userid of : " + userId);
				checkCounter = -1;
				errorMsg = "Es wurde eine ungltige Userid bergeben.";
				return UsersMediator.ReadResult.InvalidParams;
			}
			string conStr = ConfigurationSettings.AppSettings["DbConnectionString"];
			IDbConnection con = null;
			try 
			{
				con = new SqlConnection(conStr);
				con.Open();
				IDbCommand cmd = con.CreateCommand();
				cmd.CommandText = "SELECT CheckCounter FROM " + DbTableOwnerPrefix + "Users WHERE Id = @ID";
				cmd.Parameters.Add(new SqlParameter("@Id", userId));
				IDataReader reader = cmd.ExecuteReader();
				int rowCount = 0;
				checkCounter = -1;
				while(reader.Read()) 
				{
					if(rowCount > 0) 
					{
						log.Error("Multiple Users with same userid " + userId + " found during read of CheckCounter. Inconsistence.");
						checkCounter = -1;
						errorMsg = "Es ist eine Datenbankinkonsistenz aufgetreten.";
						return UsersMediator.ReadResult.Error;
					}
					checkCounter = (int)reader["CheckCounter"];
					rowCount++;
				}
				if(rowCount != 1) 
				{
					log.Error("No User with userid of " + userId + " found during read of CheckCounter.");
					errorMsg = "Es konnte kein entsprechender Benutzer gefunden werden.";
					return UsersMediator.ReadResult.InvalidParams;
				}
				errorMsg = null;
				return UsersMediator.ReadResult.Ok;
			} 
			catch(Exception e)
			{
				log.Error("Exception occured during read of CheckCounter in Users with Id=" + userId, e);
				checkCounter = -1;
				errorMsg = "Problem beim Datenbankzugriff aufgetreten.";
				return ReadResult.Error;
			} 
			finally 
			{
				try { if(con != null) con.Close(); } 
				catch(Exception ex) 
				{
					log.Warn("Unable to close connection - exception occured.", ex);
				}
			}
		} // end: method ReadUsersCheckCounter

	} // end: class UsersMediator
}