using System;
using System.IO;
using System.Net;
using System.Text;
using log4net;


namespace WebseitenMonitor.DomainObjects.DataSource
{	
	public abstract class DataSource 
	{	
		private static readonly ILog log = LogManager.GetLogger(typeof(DataSource));

		public enum GetSourceReaderResult {Ok, NoUrlDefined, NoSourceStreamObtainable, ExceptionOccured}

		protected string sourceURL;
		protected Encoding sourceEncoding;

		public string SourceURL 
		{
			get { return this.sourceURL; }
		}
		
		public Encoding SourceEncoding 
		{
			get { return this.sourceEncoding; }
		}

		public abstract Uri RetrievedSourceURI 
		{
			get;
		}

		public abstract bool SetSourceURL(string sourceURL, out string errorMessage); 

		public abstract bool SetSourceEncoding(Encoding sourceEncoding, out string errorMessage);
		public abstract bool SetSourceEncoding(string sourceEncodingName, out string errorMessage);
		public abstract bool SetSourceEncoding(int sourceEncodingCodePage, out string errorMessage);

		public abstract GetSourceReaderResult GetSourceReader(out TextReader sourceReader, out string errorMessage);

	}

	public class HttpDataSource : DataSource 
	{	
		private static readonly ILog log = LogManager.GetLogger(typeof(HttpDataSource));

		private HttpWebRequest httpRequest;
		private HttpWebResponse httpResponse;

		public override Uri RetrievedSourceURI 
		{
			get 
			{
				if(this.httpResponse == null)
					return null;
				return httpResponse.ResponseUri;
			}
		}

		public override bool SetSourceURL(string sourceURL, out string errorMessage) 
		{
			if(sourceURL == null || sourceURL.Length < 7 || !sourceURL.Substring(0,7).ToLower().Equals("http://")) 
			{
				errorMessage = "URL einer HTTP-Datenquelle muss mit \"http://\" beginnen.";
				return false;
			}
			errorMessage = "";
			this.sourceURL = sourceURL;
			return true;
		}

		public override bool SetSourceEncoding(Encoding sourceEncoding, out string errorMessage) 
		{
			if(sourceEncoding == null) 
			{
				log.Error("Attempt to explicitly set source encoding to null - not allowed.");
				errorMessage = "SourceEncoding darf nicht explizit auf null gesetzt werden.";
				return false;
			}
			this.sourceEncoding = sourceEncoding;
			errorMessage = "";
			return true;
		}

		public override bool SetSourceEncoding(string sourceEncodingName, out string errorMessage) 
		{
			if(sourceEncodingName == null) 
			{
				log.Error("Invalid parameter - sourceEncodingName is null.");
				errorMessage = "Ungltiger Parameter - SourceEncoding-Name ist null.";
				return false;
			}
			try 
			{
				this.sourceEncoding = Encoding.GetEncoding(sourceEncodingName);
			} 
			catch(Exception e) 
			{	
				log.Error("Exception occured during try to set source encoding by name.",e);
				errorMessage = "Ungltiger Parameter - Entsprechendes Encoding nicht untersttzt.";
				return false;
			}
			errorMessage = "";
			return true;
		}
		
		public override bool SetSourceEncoding(int sourceEncodingCodePage, out string errorMessage) 
		{
			try 
			{
				log.Error("Invalid parameter - sourceEncodingCodePage is null.");
				this.sourceEncoding = Encoding.GetEncoding(sourceEncodingCodePage);
			} 
			catch(Exception e) 
			{
				log.Error("Exception occured during try to set source encoding by codepage.",e);
				errorMessage = "Ungltiger Parameter - Entsprechendes Encoding nicht untersttzt.";
				return false;
			}
			errorMessage = "";
			return true;
		}
		
		/// <summary>
		/// Note: The TextReader obtained by this method uses the Encoding specified by this.sourceEncoding.
		///       If this.sourceEncoding is null and WebResponse.Characterset holds a characterset, this would
		///       be used. Otherwise it is tried to use "iso-8859-1" and if this also failed the platform default
		///       characterset is used. 
		/// </summary>
		/// <param name="sourceReader"></param>
		/// <param name="errorMessage"></param>
		/// <returns></returns>
		public override DataSource.GetSourceReaderResult GetSourceReader(out TextReader sourceReader, out string errorMessage) 
		{
			if(this.sourceURL == null || this.sourceURL.Length == 0) 
			{
				sourceReader = null;
				errorMessage = "Fr die Datenquelle ist keine URL definiert.";
				return DataSource.GetSourceReaderResult.NoUrlDefined;
			}
			
			Stream s = null;
			TextReader r = null;
			try 
			{
				this.httpRequest = (HttpWebRequest)WebRequest.Create(this.sourceURL);
				this.httpResponse = (HttpWebResponse)this.httpRequest.GetResponse();
				
				if(this.httpResponse != null) 
				{	
					// log.Debug("HTTP-Charset: " + httpResponse.CharacterSet);

					s = this.httpResponse.GetResponseStream();
					if(sourceEncoding == null) 
					{	
						Encoding myEncoding = null;
						try { myEncoding = Encoding.GetEncoding(httpResponse.CharacterSet); } 
						catch {}
						if(myEncoding == null) 
						{
							try { myEncoding = Encoding.GetEncoding("iso-8859-1"); } 
							catch {}
							if(myEncoding == null)
								r = new StreamReader(s);
							else
								r = new StreamReader(s, myEncoding);
						}
						else
							r = new StreamReader(s,myEncoding);
					}
					else 
					{
						r = new StreamReader(s,sourceEncoding);
					}
				}
			} 
			catch(Exception e)
			{	
				log.Warn("Exception occured while obtaining DataSource stream", e);
				sourceReader = null;
				errorMessage = "Zugriff auf Datenquelle fehlgeschlagen. Exception aufgetreten.";
				try 
				{
					if(r != null)
						r.Close();
					else if(s != null)
						s.Close();
					else if(this.httpResponse != null)
						this.httpResponse.Close();
				} 
				catch(Exception ee) {log.Warn("Exception occured while trying to close DataSource stream in response to an exception", ee);}
				return DataSource.GetSourceReaderResult.ExceptionOccured;
			}
			if(r != null) 
			{
				sourceReader = r;
				errorMessage = "";
				return DataSource.GetSourceReaderResult.Ok;
			} 
			else 
			{
				sourceReader = null;
				errorMessage = "Zugriff auf Datenquelle ist fehlgeschlagen.";
				return DataSource.GetSourceReaderResult.NoSourceStreamObtainable;
			}
		}

	}
}
		