using System;
using log4net;
using mshtml;
using System.Collections;
using WebseitenMonitor.HtmlUtilities;
using WebseitenMonitor.HttpUtilities;
using WebseitenMonitor.UriUtilities;
using System.Security.Cryptography;
using System.Text;

namespace WebseitenMonitor.DomainObjects.Behaviour 
{	
	/// <summary>
	/// This class represents a domain object in the WebseitenMonitor application.
	/// It is a special observation, as it is used to observe a certain Http-retrievable
	/// Html webpage up to a certain depth.
	/// In case of a depth greater than zero all Documents hyper-referenced by the observed
	/// document are observed too recursively up to the specified depth.
	/// </summary>
	/// <author>Stefan Preuer, Manuel Wallnfer</author>
	public class ChangeObservation : Observation
	{
		private static readonly ILog log = LogManager.GetLogger(typeof(ChangeObservation));
		
		// Enumeration of returncodes used by methods of this class
		internal enum BuildObservedDocumentTreeResult {Ok, Error, InvalidParams}
		
		// Enumerations used to categorize individual observed documents.
		internal enum ObservedDocumentKind // Kind of document corresponding to its role in a tree of observed documents
		{
			Original = 0,	// First occurence of this document in tree of observed documents
			Delegate = 1	// If a document occures multiple times in the tree of observed documents, only
							// the first occurence gets represented as original. All further occurances only
							// reference to the original and therefore are delegates.
		}
		internal enum ObservedDocumentState // State of the observation of an individual observed document
		{
			Unchanged = 0, 
			Changed = 1, 
			New = 2, 
			InvalidMimeType = 3, 
			Unretrievable = 4, 
			NotHashable = 5
		}
		
		// State-fields of the domain-object ChangeBehaviour
		private ChangeObservation.ObservedDocument rootDocument; // The root of ObservedDocuments in the tree
		private int depth; // [0,...[ - the depth of observation (i.e. of the tree)
		
		/// <summary>
		/// This is the default ChangeObservation constructor for an initial ChangeObservation.
		/// Default settings by this constructor:
		///		Date : Observation.InitDateTime
		///		CheckCounterTimeStamp : -1
		///		ChangeState : ChangeStateEnum.Init
		///		RootDocument : null
		///	Note:	In case of an inconsistence of the passed parameters an Exception is thrown.
		///			Consistency Requirements: id > 0, name != null, url != null, depth >= 0
		/// </summary>
		internal ChangeObservation(
			int id,
			string name,
			Uri url,
			int depth) : base(id, name, url, Observation.InitDateTime, -1, ChangeStateEnum.Init) 
		{
			// check arguments for consistency
			if(id < 0 || name == null || url == null || depth < 0)
				throw new Exception("Invalid Arguments passed to ChangeObservation constructor.");
			this.depth = depth;
			this.rootDocument = null;
		}
	
		/// <summary>
		/// This constructor lets you pass all possible arguments.
		/// Note:	In case of an inconsistence of the passed parameters an Exception is thrown.
		///			Consistency Requirements: 
		///				id > 0, name != null, url != null, depth >= 0
		///				In the changeStates New,Unchanged,Changed date must not be Observation.InitDateTime,
		///				checkCounterTimeStamp must be non-negative and rootDocument must not be null.
		///				In case of the changeState Init date must be Observation.InitDateTime, checkCounterTimeStamp
		///				must be -1 and rootDocument must be null.
		///				In the changeState Error no further consistency Requirements must get regarded.
		/// </summary>
		internal ChangeObservation(
			int id,
			string name,
			Uri url,
			DateTime date,
			int checkCounterTimeStamp,
			ChangeStateEnum changeState,
			int depth,
			ObservedDocument rootDocument) : base(id, name, url, date, checkCounterTimeStamp, changeState)
		{
			//log.Debug("ChangeObservatin constructor : " + id + " " + name + " " + url + " " + date + " " + checkCounterTimeStamp + " " + changeState + " " + depth); 
			// general argument consistency checks
			if(id < 0 || name == null || url == null || depth < 0)
				throw new Exception("Invalid Arguments passed to ChangeObservation constructor.");
			// changeState dependent argument consistency checks
			switch(changeState) 
			{
				case ChangeStateEnum.Init:
					if(date != Observation.InitDateTime || checkCounterTimeStamp != -1 || rootDocument != null)
						throw new Exception("Inconsistent Arguments passed to ChangeObservation constructor.");
					break;
				case ChangeStateEnum.New: case ChangeStateEnum.Unchanged: case ChangeStateEnum.Changed:
					if(date == Observation.InitDateTime || checkCounterTimeStamp < 0 || rootDocument == null)
						throw new Exception("Inconsistent Arguments passed to ChangeObservation constructor.");
					break;
				case ChangeStateEnum.Error:
					break;
			}
			// assertion: passed arguments are valid/consistent
			this.depth = depth;
			this.rootDocument = rootDocument;
		}

		// Getter Properties for ChangeObservation fields
		internal ChangeObservation.ObservedDocument RootDocument 
		{
			get { return this.rootDocument; }
		}

		internal int Depth 
		{
			get { return this.depth; }
		}

		internal enum ChangePropertiesEnum {Change, NoChange, Inconsistence}

		/// <summary>
		/// This method changes Properties of the ChangeObservation
		/// Note: It only changes the properties if something is different to the current properties.
		///       If only the name is different the ChangeState of the ChangeObservation doesn't change.
		///       If the url or the depth are different the ChangeObservation becomes Init ChangeState and
		///       therefore all observedDocuments get lost
		/// </summary>
		/// <param name="name"></param>
		/// <param name="url"></param>
		/// <param name="depth"></param>
		/// <returns>
		/// true ... Change was successfull (consistent)
		/// false ... Change didn't happen because of consistence violation
		/// </returns>
		internal ChangeObservation.ChangePropertiesEnum ChangeProperties(string name, Uri url, int depth) 
		{
			if(name == null || url == null || depth < 0)
				return ChangePropertiesEnum.Inconsistence;
			if(!this.url.Equals(url) || this.depth != depth) 
			{
				// reset to init with changed properties
				this.name = name;
				this.url = url;
				this.date = Observation.InitDateTime;
				this.checkCounterTimeStamp = -1;
				this.changeState = ChangeStateEnum.Init;
				this.depth = depth;
				this.rootDocument = null;
				return ChangePropertiesEnum.Change;
			}
			if(!this.name.Equals(name)) 
			{
				this.name = name; 
				return ChangePropertiesEnum.Change;
			}
			return ChangePropertiesEnum.NoChange;
		} // end: method ChangeProperties
		
		internal override CheckResult Check(int newCheckCounterTimeStamp, out string errorMsg) 
		{
			// Determine Hashtable of old-Original ObservedDocuments
			Hashtable oldOriginals = new Hashtable();
			if(this.rootDocument != null) 
			{
				this.recursiveDetermineOldOriginals(this.rootDocument, oldOriginals);
			}
			// BuildObservedDocumentsTree
			string errorMessage;
			ObservedDocument root;
			ChangeStateEnum changeState;
			BuildObservedDocumentTreeResult res = this.BuildObservedDocumentTree(oldOriginals, out root, out changeState, out errorMessage);
			if(res == BuildObservedDocumentTreeResult.Ok) 
			{
				this.rootDocument = root;
				this.changeState = changeState;
				this.date = DateTime.Now;
				this.checkCounterTimeStamp = newCheckCounterTimeStamp;
				errorMsg = null;
				return CheckResult.Ok;
			} 
			else if(res == BuildObservedDocumentTreeResult.InvalidParams) 
			{
				errorMsg = errorMessage;
				return CheckResult.InvalidParams;
			} 
			else 
			{
				errorMsg = errorMessage;
				return CheckResult.Error;
			}
		}

		private void recursiveDetermineOldOriginals(ObservedDocument doc, Hashtable oldOriginals) 
		{
			if(doc == null)
				return;
			if(doc.kind == ObservedDocumentKind.Original) 
			{
				if(oldOriginals[doc.url] == null)
					oldOriginals.Add(doc.url, doc);
				if(doc.childDocuments != null)
					foreach(ObservedDocument tmpDoc in doc.childDocuments)
						this.recursiveDetermineOldOriginals(tmpDoc, oldOriginals);
			}
		}

		/// <summary>
		/// This class is used to represent one observed-Document.
		/// </summary>
		internal class ObservedDocument 
		{
			internal int id = -1;	// Database Id-column-value, need (can) not be set by Check method of ChangeObservation.
									// but is used by database-access-methods.
			// kind: Original .. first occurence in tree of observed-documents
			//       Delegate .. delegate for an already observed-document in the tree of observed-documents
			internal ObservedDocumentKind kind;
			// state: -- Ok states:
			//        Unchanged .. observed-document didn't change in hashValue since last observation
			//        Changed .. observed-document changed in hashValue since last observation
			//        New .. observed-document didn't occure in last observation
			//        -- Failure states:
			//        InvalidMimeType .. retrieved-document wasn't HTML-Webpage by response-MimeType
			//        Unretrievable .. it wasn't possible to retrieve the document by it's url or because of network-problems
			//        NotHashable .. it wasn't possible to compute a hashValue out of the retrieved-document
			//   Note: In case of a Delegate-kinded observed-document, only the states New or Unchanged are possible. New indicates,
			//         that the father-node in the old observation didn't hyper-reference the delegated document. Therefore the whole subtree
			//         is new in this observation relative to the father-node. Unchanged means that the father-node in the old observation
			//         already hyper-referenced the delegated document, and therefore the state of the delegated document indicates the 
			//         change state.
			//         It is also important that the state of Original-kinded node is not meant in relation to there position in the whole tree
			//         of observed-documents. It only indicates the state relative to an observed-document in the old observation with the same url.
			//         Paticularly a New state is not propagated explicit to the child-observed-documents.
			internal ObservedDocumentState state;
			// title (only in Original-kinded with OK states): The HTML-title of the observed-document or "untitled" if no HTML-title found
			internal string title;
			// url: The absolute URL of the observed-document. Also used as identifier and for equality-checks.
			internal Uri url;
			// hashValue (only in Original-kinded with OK states): A cryptographic-hashValue of the observed-document,
			//                                                     which is used do determine changes between observations.
			internal byte[] hashValue;
			// childDocuments (only in Original-kinded): 
			//                 null .. observation-depth doesn't include further hyper-references
			//                 empty .. no hyper-referenced documents in this observed-document
			//                 not-empty .. list of hyper-referenced documents out of this one 
			internal ArrayList childDocuments;
			internal ObservedDocument delegateFor; // only relevant if kind is Delegate

			// Default is Original-kinded
			internal ObservedDocument() 
			{
				this.kind = ObservedDocumentKind.Original;
			}
		}
		
		/// <summary>
		/// This method sets the state of an original-kind non-error ObservedDocument in relation to the old version of the
		/// ObservedDocument tree represented by a Hashtable of all original-kind non-error ObservedDocuments of this old
		/// tree where the url is used as key.
		/// </summary>
		/// <param name="DocumentToChangeSetState">Original-kind non-error ObservedDocument to set state for.</param>
		/// <param name="oldOriginals">
		/// Hashtable of old original-kind non-error ObservedDocuments.
		/// If null is passed it is interpreted as empty Hashtable.
		/// </param>
		private void setStateForObservedOriginalDocument(ref ObservedDocument documentToChangeSetState, Hashtable oldOriginals) 
		{
			if(oldOriginals == null)
				oldOriginals = new Hashtable();
			ObservedDocument oldObservedDocument = (ObservedDocument)oldOriginals[documentToChangeSetState.url];
			if(oldObservedDocument == null) 
				documentToChangeSetState.state = ObservedDocumentState.New; 
			else 
			{
				if(oldObservedDocument.hashValue == null)
					documentToChangeSetState.state = ObservedDocumentState.Changed;
				else if(oldObservedDocument.hashValue.Length != documentToChangeSetState.hashValue.Length)
					documentToChangeSetState.state = ObservedDocumentState.Changed;
				else 
				{
					for(int i = 0; i < documentToChangeSetState.hashValue.Length; i++)
						if(documentToChangeSetState.hashValue[i] != oldObservedDocument.hashValue[i]) 
						{
							documentToChangeSetState.state = ObservedDocumentState.Changed;
							return;
						}
					documentToChangeSetState.state = ObservedDocumentState.Unchanged;
				}
			}
		} // end: method setStateForObservedOriginalDocument
		
		/// <summary>
		/// This method builds a Tree of ObservedDocument objects corresponding to the objects 
		/// url and depth to observe and classifies the observed-documents relative to an old
		/// observation by the states of the ObservedDocument objects.
		/// 
		/// The Documents are retrieved by a WebseitenMonitor.HttpUtilities.HttpHtmlRetriever object.
		/// 
		/// It is important to note that the resulting tree of observed-documents, with the root entry
		/// point returned by the out parameter root, gets build in breadth-first style. The reason
		/// for this is, that observed-documents with the same url only get retrieved once. In possible
		/// further occurences of the same observed document (same url) only a delegate-kinded
		/// ObservedDocuments gets inserted in the tree (without reretrieval of the original). Therefore
		/// it is necessary that the original observed-document is that one with the lowest depth among
		/// all observed-documents with the same url. In this case it is guaranteed that for each
		/// delegate-kinded observed-documents there are enough decendant observed-documents corresponding
		/// to the whole observation-depth.
		/// 
		/// It is also important to explain the role of the Hashtable oldOriginals (input parameter for the
		/// method). This Hashtable should hold all original-kind ObservedDocuments objects of an old
		/// observation to which this observation should be relative too. If instead null or an empty
		/// Hashtable is passed an absolute new observation is started. The pass of an oldOriginals Hashtable
		/// doesn't influence the way the tree is built, it only influences the states of the individual
		/// observed-documents.
		/// </summary>
		/// <param name="oldOriginals">Hashtable of old original-kind ObservedDocuments with their Uri as key.</param>
		/// <param name="root">Root ObservedDocument object of determined Tree.</param>
		/// <param name="changeState">Aggregated ChangeState of Observation.</param>
		/// <param name="errorMsg">Possible errormessage</param>
		/// <returns>Resultstatus</returns>
		internal BuildObservedDocumentTreeResult BuildObservedDocumentTree(
			Hashtable oldOriginals,
			out ObservedDocument root,
			out ChangeStateEnum changeState,
			out string errorMsg) 
		{
			if(this.Url == null) 
			{
				root = null;
				changeState = ChangeStateEnum.Error;
				errorMsg = "Es ist keine Html-Quellurl gesetzt.";
				return BuildObservedDocumentTreeResult.InvalidParams;
			}
			if(this.depth < 0) 
			{
				root = null;
				changeState = ChangeStateEnum.Error;
				errorMsg = "Es ist eine negative berprfungstiefe gesetzt.";
				return BuildObservedDocumentTreeResult.InvalidParams;
			}
			// retrieve root document
			HttpHtmlRetriever retriever = new HttpHtmlRetriever();
			string sourceText, errorMessage;
			HttpHtmlRetriever.RetrieveResult retrieveResult = retriever.Retrieve(this.Url.ToString(), out sourceText, out errorMessage);
			if(retrieveResult != HttpHtmlRetriever.RetrieveResult.Ok) 
			{
				root = null;
				changeState = ChangeStateEnum.Error;
				errorMsg = errorMessage;
				return BuildObservedDocumentTreeResult.Error;
			}
			if(sourceText == null) 
			{
				root = null;
				changeState = ChangeStateEnum.Error;
				errorMsg = "Es konnte kein Html-Quelltext bezogen werden.";
				return BuildObservedDocumentTreeResult.Error;
			}
			// assertion: source text of root document successfully retrieved
			// remember if oldOriginals was null or empty, because in this case the aggregated ChangeState
			// gets NEW
			bool isNewBuildOfObservedDocumentsTree = (oldOriginals == null || oldOriginals.Count == 0);
			ChangeStateEnum aggregatedChangeState = ChangeStateEnum.Unchanged;
			// if no oldOriginals passed - regard them as empty
			if(oldOriginals == null)
				oldOriginals = new Hashtable();
			Hashtable alreadyFoundDocuments = new Hashtable(); // only for original-kind ObservedDocuments
			ObservedDocument myRootDocument = new ObservedDocument();
			ArrayList curOpenDocumentList = new ArrayList();
			ArrayList tmpOpenDocumentList = new ArrayList();
			string tmpTitle, tmpErrorMsg;
			// determine root-document parts
			myRootDocument.url = this.Url;
			tmpTitle = this.parseDocumentForTitle(sourceText);
			myRootDocument.title = (tmpTitle == null)?"Untitled":tmpTitle;
			if(!this.createDocumentHash(sourceText, out myRootDocument.hashValue, out tmpErrorMsg)) 
			{
				root = null;
				changeState = ChangeStateEnum.Error;
				errorMsg = "Aus dem bergebenen Quelltext konnte kein Hashwert berechnet werden.";
				return BuildObservedDocumentTreeResult.Error;
			}
			if(this.depth > 0) // determine child-urls only if depth > 0
			{
				// Note: childDocuments field is only used temporarly for ParsedLinkDescription-List.
				//       Finally holds ObservedDocument-List of childs.
				if(!this.parseDocumentForLinks(sourceText, this.Url, out myRootDocument.childDocuments)) 
				{
					root = null;
					changeState = ChangeStateEnum.Error;
					errorMsg = "Suche nach links im Quelldokument verursachte einen Fehler.";
					return BuildObservedDocumentTreeResult.Error;
				}
			}
			this.setStateForObservedOriginalDocument(ref myRootDocument,oldOriginals);
			// if myRootDocument isn't unchanged since last observation -> something changed
			if(myRootDocument.state != ObservedDocumentState.Unchanged)
				aggregatedChangeState = ChangeStateEnum.Changed;
			root = myRootDocument;
			curOpenDocumentList.Add(myRootDocument);
			alreadyFoundDocuments.Add(myRootDocument.url,myRootDocument);
			ArrayList tmpDocumentList;
			ObservedDocument tmpObservedDocument;
			// create Tree in breadth-first style
			for(int i = 0; i < this.depth; i++) 
			{
				// determine list of child-urls of currently open document list
				tmpOpenDocumentList = new ArrayList();
				foreach(ObservedDocument curDocument in curOpenDocumentList) 
				{
					// Only non-delegate documents with non-error state are considered in further depth
					if(curDocument.kind == ObservedDocumentKind.Delegate)
						continue;
					else if(
						curDocument.state == ObservedDocumentState.InvalidMimeType ||
						curDocument.state == ObservedDocumentState.Unretrievable ||
						curDocument.state == ObservedDocumentState.NotHashable)
						continue;

					tmpDocumentList = new ArrayList();
					foreach(ParsedLinkDescription curLink in curDocument.childDocuments) 
					{
						tmpObservedDocument = new ObservedDocument();
						tmpObservedDocument.url = curLink.absoluteLinkUri;
						// check whether document referenced by curLink already gets retrieved
						tmpObservedDocument.delegateFor = (ObservedDocument)alreadyFoundDocuments[curLink.absoluteLinkUri];
						if(tmpObservedDocument.delegateFor != null) 
						{
							// assertion: ObservedDocument-node is delegate for already computed document
							tmpObservedDocument.kind = ObservedDocumentKind.Delegate;
							// check whether this delegate-child is new (no corresponding child-node in oldDocuments, uri is key)
							// and set state accordingly. New if corresponding fathernode in oldOriginals doesn't have a childnode
							// of current url otherwise Unchanged - note that in this case referenced original node of delegate 
							// determines change-state.
							ObservedDocument tmpOldObservedDocument = (ObservedDocument)oldOriginals[curDocument];
							if(tmpOldObservedDocument != null) 
							{
								if(tmpOldObservedDocument.childDocuments == null)
									tmpObservedDocument.state = ObservedDocumentState.New;
								else 
								{
									foreach(ObservedDocument tmpOldChildObservedDocument in tmpOldObservedDocument.childDocuments) 
										if(tmpObservedDocument.url.Equals(tmpOldChildObservedDocument.url)) 
										{
											tmpObservedDocument.state = ObservedDocumentState.Unchanged;
										}
								}
							} 
							else 
								tmpObservedDocument.state = ObservedDocumentState.New;
							tmpDocumentList.Add(tmpObservedDocument);
							continue;
						}
						// retrieve document referenced by curLink
						retriever = new HttpHtmlRetriever();
						retrieveResult = retriever.Retrieve(curLink.absoluteLinkUri.ToString(), out sourceText, out errorMessage);
						if(retrieveResult == HttpHtmlRetriever.RetrieveResult.InvalidResponseMimeType) 
						{
							tmpObservedDocument.state = ObservedDocumentState.InvalidMimeType;
							tmpDocumentList.Add(tmpObservedDocument);
							alreadyFoundDocuments.Add(tmpObservedDocument.url,tmpObservedDocument);
							// if in last observation the url wasn't an InvalidMimeType -> something changed
							if(oldOriginals != null) 
							{
								ObservedDocument oldObservedDocument = (ObservedDocument)oldOriginals[tmpObservedDocument.url];
								if(oldObservedDocument != null && oldObservedDocument.state != ObservedDocumentState.InvalidMimeType)
									aggregatedChangeState = ChangeStateEnum.Changed;
							}
							continue;
						} 
						else if(retrieveResult != HttpHtmlRetriever.RetrieveResult.Ok) 
						{
							tmpObservedDocument.state = ObservedDocumentState.Unretrievable;
							tmpDocumentList.Add(tmpObservedDocument);
							alreadyFoundDocuments.Add(tmpObservedDocument.url,tmpObservedDocument);
							// if in last observation the url wasn't Unretrievable -> something changed
							if(oldOriginals != null) 
							{
								ObservedDocument oldObservedDocument = (ObservedDocument)oldOriginals[tmpObservedDocument.url];
								if(oldObservedDocument != null && oldObservedDocument.state != ObservedDocumentState.Unretrievable)
									aggregatedChangeState = ChangeStateEnum.Changed;
							}
							continue;
						}
						// assertion: retrieval of Html-document was successful - sourceText holds the document-text
						tmpTitle = this.parseDocumentForTitle(sourceText);
						tmpObservedDocument.title = (tmpTitle == null)?"Untitled":tmpTitle;
						if(!this.createDocumentHash(sourceText, out tmpObservedDocument.hashValue, out tmpErrorMsg)) 
						{
							tmpObservedDocument.state = ObservedDocumentState.NotHashable;
							tmpDocumentList.Add(tmpObservedDocument);
							alreadyFoundDocuments.Add(tmpObservedDocument.url,tmpObservedDocument);
							// if in last observation the url wasn't NotHashable -> something changed
							if(oldOriginals != null) 
							{
								ObservedDocument oldObservedDocument = (ObservedDocument)oldOriginals[tmpObservedDocument.url];
								if(oldObservedDocument != null && oldObservedDocument.state != ObservedDocumentState.NotHashable)
									aggregatedChangeState = ChangeStateEnum.Changed;
							}
							continue;
						}
						if(i < this.depth - 1) // don't determine child-urls in deepest layer
						{
							// Note: childDocuments field is only used temporarly for ParsedLinkDescription-List.
							//       Finally holds ObservedDocument-List of childs.
							if(!this.parseDocumentForLinks(sourceText, curLink.absoluteLinkUri, out tmpObservedDocument.childDocuments)) 
							{
								tmpObservedDocument.childDocuments = new ArrayList(); // no childes in case of error in determining childs.
							}
						}
						this.setStateForObservedOriginalDocument(ref tmpObservedDocument,oldOriginals);
						tmpDocumentList.Add(tmpObservedDocument);
						alreadyFoundDocuments.Add(tmpObservedDocument.url,tmpObservedDocument);
						// if tmpObservedDocument isn't unchanged since last observation -> something changed
						if(tmpObservedDocument.state != ObservedDocumentState.Unchanged)
							aggregatedChangeState = ChangeStateEnum.Changed;
					} // end: foreach curDocument.childDocuments
					curDocument.childDocuments = tmpDocumentList;
					tmpOpenDocumentList.AddRange(tmpDocumentList);
				} // end: foreach curOpenDocumentList
				curOpenDocumentList = tmpOpenDocumentList;
			}
			if(isNewBuildOfObservedDocumentsTree)
				changeState = ChangeStateEnum.New;
			else
				changeState = aggregatedChangeState;
			errorMsg = "";
			return BuildObservedDocumentTreeResult.Ok;
		} // end: method BuildObservedDocumentTree
		
		/// <summary>
		/// This method is a helper-method, which parses the passed text for a
		/// HTML-title element and returns this.
		/// </summary>
		/// <param name="text">HTML-sourcetext</param>
		/// <returns>
		/// Title corresponding to found HTML-title element or null if no
		/// regular HTML-title element found.
		/// </returns>
		private string parseDocumentForTitle(string text) 
		{
			if(text == null) 
				return null;
			HtmlTagScanner myScanner = new HtmlTagScanner();
			myScanner.Init(text);
			HtmlTagToken curToken = myScanner.NextTag();
			while(curToken.kind != HtmlTagToken.eof) 
			{
				if(curToken.kind == HtmlTagToken.openingTag && curToken.name.Equals("title")) 
				{
					int titleTextStartIndex = curToken.endPos + 1;
					int titleTextOuterEndIndex;
					curToken = myScanner.NextTag();
					if(curToken.kind == HtmlTagToken.closingTag && curToken.name.Equals("title")) 
					{
						// assertion: valid title found
						titleTextOuterEndIndex = curToken.beginPos;
						return text.Substring(titleTextStartIndex, titleTextOuterEndIndex - titleTextStartIndex);
						
					}
					continue; // no closing tag found - go on with search
				}
				curToken = myScanner.NextTag();
			}
			return null; // no title found
		} // end: method parseDocumentForTitle

		// ArrayList links contains ParsedLinkDescription elements all with distinct absoluteLinkUri.
		/// <summary>
		/// This method parses a Html-Document for hyper-references (a-elements with href attribute).
		/// 
		/// It is important to know that the resulting list of links (ArrayList of ParsedLinkDescription objects)
		/// doesn't hold duplicate Urls in case of an multible occurence of the same Url.
		/// Urls are distinguished by the equals or rather GetHashCode implementation of the System.Uri class.
		/// Therefore Uri's which only differ in a fragment-identifier are seen as equal and therefore also yield
		/// to a same hashCode.
		/// 
		/// Clients of this method can also rely on the circumstance that the resulting list of Uri's, where each
		/// is encapsulated by a ParsedLinkDescription only hold absolute Uris. Relative Uri's which occure as
		/// hyper-references in the document are resolved to the corresponding absolute one, where the passed
		/// parameter named documentUri indicates the base for relative-to-absolute resolution.
		/// </summary>
		/// <param name="text">Html-sourcetext.</param>
		/// <param name="documentUri">Uri of the Html-document with the sourcetext passed with text.</param>
		/// <param name="links">List of ParsedLinkDescription objects for each distinct link in the parsed document.</param>
		/// <returns>true - ok, false - critical problem occured.</returns>
		private bool parseDocumentForLinks(string text, Uri documentUri, out ArrayList links) 
		{
			if(text == null) 
			{
				links = null;
				return false;
			}
			Hashtable alreadyFoundUrisAsKeys = new Hashtable();
			alreadyFoundUrisAsKeys.Add(documentUri, true); // avoid self-referencing links
			ArrayList tmpList = new ArrayList();
			HtmlTagScanner myScanner = new HtmlTagScanner();
			myScanner.Init(text);
			HtmlTagToken curToken = myScanner.NextTag();
			// step through all Html-Tags of the document
			while(curToken.kind != HtmlTagToken.eof) 
			{
				// Note: HtmlTagScanner converts tagnames to lowercase
				if(curToken.kind == HtmlTagToken.openingTag && curToken.name.Equals("a"))
				{
					// assertion: curToken is opening a-Tag
					// search for href-attribute
					HtmlTagAttributeScanner attrScanner = new HtmlTagAttributeScanner();
					attrScanner.Init(text.Substring(curToken.beginPos, curToken.endPos - curToken.beginPos + 1));
					HtmlTagAttributeToken curAttrToken = attrScanner.NextAttribute();
					bool hrefFound = false;
					while(curAttrToken.kind != HtmlTagAttributeToken.eof) 
					{
						if(curAttrToken.name.Equals("href")) 
						{
							hrefFound = true;
							break;
						}
						curAttrToken = attrScanner.NextAttribute();
					}
					if(hrefFound) 
					{
						// assertion: curToken is opening a-Tag with href-attribute (held in curAttrToken)
						// Get the absolut Uri of the link, because maybe the href-attribute holds a relative one.
						Uri absoluteLinkUri = UriConvertion.GetAbsoluteUriCorrToContext(documentUri, curAttrToken.val);
						// If it wasn't possible to get an absolute Uri out of the href-attribute value or if the determined
						// absolte Uri already was discovered don't regard this link (again).
						if(absoluteLinkUri == null || alreadyFoundUrisAsKeys.ContainsKey(absoluteLinkUri)) 
						{
							curToken = myScanner.NextTag();
							continue;
						}
						// assertion: it was possible to convert the found href-value in a wellformed absolute uri
						// try to determine link-text
						int linkTextStartIndex = curToken.endPos + 1;
						int linkTextOuterEndIndex;
						curToken = myScanner.NextTag();
						if(curToken.name.Equals("a") && curToken.kind == HtmlTagToken.closingTag) 
						{

							linkTextOuterEndIndex = curToken.beginPos;
							ParsedLinkDescription tmpDescr = new ParsedLinkDescription(
								text.Substring(linkTextStartIndex, linkTextOuterEndIndex - linkTextStartIndex),
								absoluteLinkUri);
							tmpList.Add(tmpDescr);
							alreadyFoundUrisAsKeys.Add(absoluteLinkUri,true);
							
						} else
							continue;
					}
				}
				curToken = myScanner.NextTag();
			} // end: loop while
			links = tmpList;
			return true;
		} // end: method parseDocumentForLinks
		
		/// <summary>
		/// This class encapsulates data of a hyper-reference found in a HTML-document.
		/// It holds the name of the hyper-reference (CDATA of a-element) and the hyper-reference
		/// Uri.
		/// </summary>
		internal class ParsedLinkDescription 
		{
			internal string linkName;
			internal Uri absoluteLinkUri;

			internal ParsedLinkDescription(string linkName, Uri absoluteLinkUri) 
			{
				this.linkName = linkName;
				this.absoluteLinkUri = absoluteLinkUri;
			}
		}
		
		/// <summary>
		/// This method is used to calculate the hash-value of an observed-document.
		/// The current implementation calculates a MD5 Hash-value.
		/// </summary>
		/// <param name="text">source</param>
		/// <param name="hashValue">calculated hash</param>
		/// <param name="errorMsg">Error message in case of false returned</param>
		/// <returns>true .. ok, false .. an error occured, unable to calculate hash.</returns>
		private bool createDocumentHash(string text, out byte[] hashValue, out string errorMsg) 
		{
			// Generate MD5 Hash of text
			UTF8Encoding enc = new UTF8Encoding(); 
			byte[] textUtf8Bytes = enc.GetBytes(text);
			MD5 hashAlgorithm = new MD5CryptoServiceProvider();
			byte[] tmpHashValue = hashAlgorithm.ComputeHash(textUtf8Bytes);
			if(tmpHashValue == null) 
			{
				hashValue = null;
				errorMsg = "Es konnte kein Hash-Wert berechnet werden.";
				return false;
			}
			hashValue = tmpHashValue;
			errorMsg = "";
			return true;
		}
		
		/// <summary>
		/// This method can be used to log a tree of observed-documents e.g. for debugging
		/// reasons.
		/// </summary>
		/// <param name="root"></param>
		internal void logObservedDocumentTree(ObservedDocument root) 
		{
			StringBuilder sb = new StringBuilder();
			sb.Append("\n");
			if(root == null) 
			{
				sb.Append("Root node is null");
			}
			this.recursiveLogObservedDocumentTree(root, ref sb, 0);
			log.Debug(sb.ToString());
		}
		
		// recursive helper-method for logging a tree of observed-documents.
		private void recursiveLogObservedDocumentTree(ObservedDocument node, ref StringBuilder sb, int depth) 
		{
			string indent = "";
			for(int i = 0; i < depth; i++)
				indent = indent + "          ";
			if(node == null) 
			{
				sb.Append("\n" + indent + "null\n");
				return;
			}
			sb.Append(
				"\n" +
					indent + "kind: " + node.kind + "\n" +
					indent + "state: " + node.state + "\n" +
					indent + "title: " + node.title + "\n" +
					indent + "url: " + node.url + "\n" +
					indent + "hash: " + this.byteArrayToString(node.hashValue) + "\n" +
					indent + "delegateFor: " + ((node.delegateFor == null)?"null":node.delegateFor.url.ToString()) + "\n" +
				"\n");
			if(node.childDocuments != null)
				foreach(ObservedDocument child in node.childDocuments)
					recursiveLogObservedDocumentTree(child, ref sb, depth + 1);
		}

		// Used for logging the hash-value of an observed-document.
		private string byteArrayToString(byte[] a) 
		{
			if(a == null)
				return "null";
			StringBuilder sb = new StringBuilder();
			foreach(byte b in a) 
			{
				sb.Append(b);
			}
			return sb.ToString();
		}
	}
}