using System;
using System.Text.RegularExpressions;
using System.Collections;
using log4net;
using System.IO;
using System.Text;
using System.Xml.XPath;
using mshtml;
using WebseitenMonitor.UserSelection;
using WebseitenMonitor.HttpUtilities;

namespace WebseitenMonitor.DomainObjects.Behaviour 
{	
	/// <summary>
	/// This class represents a domain object in the WebseitenMonitor application.
	/// It is a special observation, as it is used to observe a certain Http-retrievable
	/// Html webpage limited to user defined parts of it. This parts get used to parameterize
	/// a user defined message template. The result is a user defined message, that gets observed
	/// by such an observation.
	/// The user defined selection of parts of the Html page under observation is realized in two
	/// consecutive steps, where the second step complements the first one. Therefore the second step
	/// is an optional one.
	/// So the first selection is based on the DOM-view of an Html document. The user can select parts
	/// of the document based on there position in the DOM tree, where the garnularity is based on Html
	/// elements or DOM subtrees to express it in the terminology of DOM. Multible such selections within a
	/// Html document are possible, and the order of the selection is significant as the consecution of the
	/// individual parts represents the base for the second step of selection. By default the body of the Html
	/// document gets the only part selected in this first step.
	/// In the second step the selection gets refined based on a regular expression. This second step is 
	/// optional. The extraction of parts matched by the regular expression is done with the concept of
	/// capturing groups. There are named and numbered capturing groups possible, and the matches of the
	/// capturing groups form the actual values of the possible parameters for the user defined message
	/// template.
	/// </summary>
	/// <author>Stefan Preuer, Manuel Wallnfer</author>
	public class UserDefinedMessageObservation : Observation
	{	
		private static readonly ILog log = LogManager.GetLogger(typeof(UserDefinedMessageObservation));

		public enum UpdateSearchPatternParameterResult {Match, NoMatch, Error}
		public enum GetUserDefinedMessageResult {Ok, NoRegExMatch, RegExParamNotFound, NoSelectionMatch, Error}
		public enum GetUserSelectedElementsResult {Match, NoMatch, Error}
		public enum ApplySearchPatternResult {Ok, NoMatch, ParamNotFound, Error}
		public enum CreateSelectedElementsPositionDescriptorResult {Ok, NoMatch, InvalidParam, Error}
		public enum ServerSearchDirection {Left, Right, Both}
		public enum CreateUserDefinedMessageObservationResult {Ok, NoMatch, InvalidParam, Error}
		public enum PreviewResult {Ok, NoRegExMatch, RegExParamNotFound, NoSelectionMatch, RetrieveError, InvalidParams, Error}
		
		// Common options for the .Net regex engine.
		private static readonly RegexOptions myRegexOptions = RegexOptions.Singleline | RegexOptions.Multiline;
		
		/// <summary>
		/// This enumeration is used to specify the detail result state of the UserDefinedMessageObservation.
		/// Its important to note, that this state doesn't mean the changeState of the Observation. The changeState
		/// indicates how the observed document changed in the definition of the Observation.
		/// The resultState indicates the interpretation of the result, which is the resultText in case of the
		/// UserDefinedMessageObservation.
		/// Undefined            ... No result obtained yet - initial state
		/// Ok                   ... Result successfully determined (resultText holds the result)
		/// NoSearchPatternMatch ... Check didn't find a match corresponding to the searchPattern of the UserDefinedMessageObservation.
		///                          Therefore resultText doesn't hold a value.
		/// MessageTemplateParamNotFound ... Check didn't find a mapping to the messageTemplate of the UserDefinedMessageObservation.
		///                                  Therefore resultText doesn't hold a value.
		/// NoSelectionMatch     ... Check didn't find a match corresponding to the selectedElementsPositionDescriptor of the UserDefinedMessageObservation.
		///                          Note: NoSelectionMatch dominates a potentially NoSearchPatternMatch or MessageTemplateParamNotFound.
		///                          Therefore resultText doesn't hold a value.
		/// Error                ... An error occured during check - The changeState also holds the error-state.
		/// </summary>
		public enum ResultStateEnum 
		{
			Undefined = 0, 
			Ok = 1, 
			NoSearchPatternMatch = 2, 
			MessageTemplateParamNotFound = 3, 
			NoSelectionMatch = 4, 
			Error = 5
		}

		// State-fields of the domain-object UserDefinedMessageObservation
		private bool searchPatternEnabled; // false - searchPattern and messageTemplate are not used
		private string searchPattern; // Regex-search-pattern
		private string messageTemplate; // message-template for userdefined message
		private string selectedElementsPositionDescriptor; // note: "-" indicates whole Html-body
		private string resultText; // not null only if resultState == Ok
		private UserDefinedMessageObservation.ResultStateEnum resultState; // Undefined in init-state

		private Regex searchPatternRegex; // object-version of searchPattern for faster processing
		
		/// <summary>
		/// This is the default UserDefinedMessageObservation constructor for an initial UserDefinedMessageObservation.
		/// Default settings by this constructor:
		///		Date : Observation.InitDateTime
		///		CheckCounterTimeStamp : -1
		///		ChangeState : ChangeStateEnum.Init
		///		RootDocument : null
		///		resultText : null
		///		resultState : Undefined
		///	Note:	In case of an inconsistence of the passed parameters an Exception is thrown.
		///			Consistency Requirements:	id > 0, name != null, url != null
		///										searchPattern, messageTemplate syntactically correct if searchPatternEnabled
		///										selectedElementsPositionDescriptor syntactically correct or null for whole body.
		///			If selectedElementsPositionDescritor is passed as null the "-" selectedElementsPositionDescriptor is assumed,
		///			which means the whole html-body.
		///	
		/// </summary>
		internal UserDefinedMessageObservation(
			int id,
			string name,
			Uri url,
			bool searchPatternEnabled,
			string searchPattern,
			string messageTemplate,
			string selectedElementsPositionDescriptor) : base(id, name, url, Observation.InitDateTime, -1, ChangeStateEnum.Init)
		{
			// check general arguments for consistency
			if(id < 0 || name == null || url == null)
				throw new Exception("Invalid Arguments passed to UserDefinedMessageObservation constructor.");
			// check UserDefinedMessageObservation specific arguments for consistency
			this.searchPatternEnabled = searchPatternEnabled;
			string errorMsg;
			if(searchPatternEnabled) 
			{
				if(!this.SetSearchPattern(searchPattern, out errorMsg))
					throw new Exception("Invalid searchPattern argument passed to UserDefinedMessageObservation constructor: " + errorMsg);
				if(!this.SetMessageTemplate(messageTemplate, out errorMsg))
					throw new Exception("Invalid messageTemplate argument passed to UserDefinedMessageObservation constructor: " + errorMsg);
			} 
			else 
			{
				this.searchPattern = null;
				this.messageTemplate = null;
			}
			if(selectedElementsPositionDescriptor == null || selectedElementsPositionDescriptor.Equals("")) 
			{
				if(!this.SetSelectedElementsPositionDescriptor("-", out errorMsg))
					throw new Exception("Invalid selectedElementsPositionDescriptor argument passed to UserDefinedMessageObservation constructor: " + errorMsg);
			} 
			else 
			{
				if(!this.SetSelectedElementsPositionDescriptor(selectedElementsPositionDescriptor, out errorMsg))
					throw new Exception("Invalid messageTemplate argument passed to UserDefinedMessageObservation constructor: " + errorMsg);
			}
			this.resultText = null;
			this.resultState = UserDefinedMessageObservation.ResultStateEnum.Undefined;
		}

		/// <summary>
		/// This constructor lets you pass all possible arguments.
		/// Note:	In case of an inconsistence of the passed parameters an Exception is thrown.
		///			Consistency Requirements: 
		///				id > 0, name != null, url != null
		///				searchPattern, messageTemplate syntactically correct if searchPatternEnabled
		///				selectedElementsPositionDescriptor syntactically correct or null for whole body.
		///				In the changeStates New,Unchanged,Changed date must not be Observation.InitDateTime,
		///				checkCounterTimeStamp must be non-negative.
		///				In case of the changeState Init date must be Observation.InitDateTime, checkCounterTimeStamp
		///				must be -1, resultState must be undefined and resultText must be null.
		///			If selectedElementsPositionDescritor is passed as null the "-" selectedElementsPositionDescriptor is assumed,
		///			which means the whole html-body.
		/// </summary>
		internal UserDefinedMessageObservation(
			int id,
			string name,
			Uri url,
			DateTime date,
			int checkCounterTimeStamp,
			ChangeStateEnum changeState,
			bool searchPatternEnabled,
			string searchPattern,
			string messageTemplate,
			string selectedElementsPositionDescriptor,
			string resultText,
			UserDefinedMessageObservation.ResultStateEnum resultState) : base(id, name, url, date, checkCounterTimeStamp, changeState)
		{
			// check general arguments for consistency
			if(id < 0 || name == null || url == null)
				throw new Exception("Invalid Arguments passed to UserDefinedMessageObservation constructor.");
			// changeState dependent argument consistency checks
			switch(changeState) 
			{
				case ChangeStateEnum.Init:
					if(date != Observation.InitDateTime || checkCounterTimeStamp != -1 || resultText != null || resultState != UserDefinedMessageObservation.ResultStateEnum.Undefined)
						throw new Exception("Inconsistent Arguments passed to UserDefinedMessageObservation constructor.");
					break;
				case ChangeStateEnum.New: case ChangeStateEnum.Unchanged: case ChangeStateEnum.Changed:
					if(date == Observation.InitDateTime || checkCounterTimeStamp < 0 || resultState == UserDefinedMessageObservation.ResultStateEnum.Undefined)
						throw new Exception("Inconsistent Arguments passed to UserDefinedMessageObservation constructor.");
					break;
				case ChangeStateEnum.Error:
					break;
			}
			// assertion: passed arguments are valid/consistent
			// check UserDefinedMessageObservation specific arguments for consistency
			this.searchPatternEnabled = searchPatternEnabled;
			string errorMsg;
			if(searchPatternEnabled) 
			{
				if(!this.SetSearchPattern(searchPattern, out errorMsg))
					throw new Exception("Invalid searchPattern argument passed to UserDefinedMessageObservation constructor: " + errorMsg);
				if(!this.SetMessageTemplate(messageTemplate, out errorMsg))
					throw new Exception("Invalid messageTemplate argument passed to UserDefinedMessageObservation constructor: " + errorMsg);
			} 
			else 
			{
				this.searchPattern = null;
				this.messageTemplate = null;
			}
			if(selectedElementsPositionDescriptor == null || selectedElementsPositionDescriptor.Equals("")) 
			{
				if(!this.SetSelectedElementsPositionDescriptor("-", out errorMsg))
					throw new Exception("Invalid selectedElementsPositionDescriptor argument passed to UserDefinedMessageObservation constructor: " + errorMsg);
			} 
			else 
			{
				if(!this.SetSelectedElementsPositionDescriptor(selectedElementsPositionDescriptor, out errorMsg))
					throw new Exception("Invalid selectedElementsPositionDescriptor argument passed to UserDefinedMessageObservation constructor: " + errorMsg);
			}
			this.resultState = resultState;
			this.resultText = resultText;
		}

		/// <summary>
		/// This constructor directly sets the passed values.
		/// Note: Can only be used in same class (private) - because no checks are done.
		/// </summary>
		private UserDefinedMessageObservation(
			int id,
			string name,
			Uri url,
			DateTime date,
			int checkCounterTimeStamp,
			ChangeStateEnum changeState,
			bool searchPatternEnabled,
			string searchPattern,
			string messageTemplate) : base(id, name, url, date, checkCounterTimeStamp, changeState) 
		{
			this.searchPatternEnabled = searchPatternEnabled;
			this.selectedElementsPositionDescriptor = null;
			this.resultText = null;
			this.resultState = ResultStateEnum.Undefined;
			if(SearchPatternEnabled) 
			{
				string errorMsg;
				if(!this.SetSearchPattern(searchPattern, out errorMsg))
					throw new Exception("Invalid searchPattern passed to UserDefinedMessageObservation constructor!");
				if(!this.SetMessageTemplate(messageTemplate, out errorMsg))
					throw new Exception("Invalid messageTemplate passed to UserDefinedMessageObservation constructor!");
			}
		}
		
		/// <summary>
		/// This constructor creates a minimum inconsistent UserDefinedMessageObservation object.
		/// Note: Can only be used in same class for individual reasons - private (no checks are done.
		/// </summary>
		/// <param name="url"></param>
		private UserDefinedMessageObservation() : base(-1,"",null,Observation.InitDateTime,-1,ChangeStateEnum.Init) {}
		
		/// <summary>
		/// This method is used to create a new UserDefinedMessageObservation out of the SelectedElementsPositionDescriptor
		/// obtained from the client-script.
		/// The SelectedElementsPositionDescriptor is based on a path in the DOM-tree of the source-html-document.
		/// Since there can happen a problem in the synchronization of the clients DOM-tree and the server DOM-engined DOM-tree
		/// it is checked if the clients SelectedElementsPositionDescriptor yields to a match by the servers DOM-engine too.
		/// Only in case of a match a corresponding UserDefinedMessageObservation-object is returned.
		/// 
		/// Note: If no problem occured the resulting UserDefinedMessageObservation object is in the changeState New.
		///       The date of the check is automatically set to now.
		/// 
		/// The clientside selectedElementsPositionDescriptor must be of the following syntactical structure:
		///     ElementsPositionDescriptor = id ";" depth ";" LayerDescriptor{depth} ">".
		///     LayerDescriptor = nodeName "," numSamePrevSib "," numSameNextSib "," elemCountRelToBody "," sameElemCountRelToBody "," sourceIndexRelToBody ";".
		///   where:
		///     id .. steady increasing non-negative integer identifier (the earlier selected the lower the id).
		///     depth .. depth of selected element in DOM-Tree with Body-node as root-node.
		///     nodeName .. corresponding to DOM-node-attribute nodeName.
		///     numSamePrevSib .. number of same named previous siblings in current layer.
		///     numSameNextSib .. number of same named following siblings in current layer.
		///     elemCountRelToBody .. count-index of current-layer-element relative to the Body-node (nodes of type elementnode (typeid 1) count - body itself has index 0)
		///     sameElemCountRelToFather .. count-index of current-layer-element relative to Father-node (only same named elementnodes count - father itself has index 0 whether same named or not)
		///     sourceIndexRelToBody .. DOM-node-attribute sourceIndex relative to that of document.body (only available in MS Internet Explorer), if not defined sourceIndexRelToBody is -1. 
		/// </summary>
		/// <param name="name">Name of UserDefinedMessageObservation - must not be null.</param>
		/// <param name="url">Url of UserDefinedMessageObservation - must not be null.</param>
		/// <param name="checkCounterTimeStamp">CheckCounterTimeStamp of UserDefinedMessageObservation.</param>
		/// <param name="searchPattern">SearchPattern of UserDefinedMessageObservation - if null searchPatternEnabled becomes false.</param>
		/// <param name="messageTemplate">MessageTemplate of UserDefinedMessageObservation - if null searchPatternEnabled becomes false.</param>
		/// <param name="clientElementsPositionDescriptor">
		/// The SelectedElementsPositionDescriptor constructed from the client-Script (UserSelectScript.js) based on the users selection.
		/// Must not be null.
		/// </param>
		/// <param name="text">
		/// html-source-text on which the clientSelectedElementsPositionDescriptor is based. Must not be null.
		/// </param>
		/// <param name="searchDir">Synchronization direction for siblings at one layer in the DOM-Tree.</param>
		/// <param name="errorMsg">ErrorMessage in case of a result different from Ok.</param>
		/// <returns>Ok, NoMatch, InvalidParam, Error</returns>
		internal static UserDefinedMessageObservation.CreateUserDefinedMessageObservationResult CreateUserDefinedMessageObservation(
			string name,
			Uri url,
			int checkCounterTimeStamp,
			string searchPattern,
			string messageTemplate,
			string clientElementsPositionDescriptor,
			string text,
			ServerSearchDirection searchDir,
			out UserDefinedMessageObservation resultUdm,
			out string errorMsg) 
		{
			// Checks of Parameters
			if(name == null || url == null || clientElementsPositionDescriptor == null || text == null) 
			{
				resultUdm = null;
				errorMsg = "Ungltige Parameter";
				return CreateUserDefinedMessageObservationResult.InvalidParam;
			}
			bool searchPatternEnabled = (searchPattern != null && messageTemplate != null);
			UserDefinedMessageObservation udm = null;
			try 
			{
				udm = new UserDefinedMessageObservation(
					-1, 
					name, 
					url,
					DateTime.Now,
					checkCounterTimeStamp,
					ChangeStateEnum.New,
					searchPatternEnabled, 
					searchPatternEnabled?searchPattern:null,
					searchPatternEnabled?messageTemplate:null);
			} 
			catch 
			{
				resultUdm = null;
				errorMsg = "Ungltige Parameter";
				return CreateUserDefinedMessageObservationResult.InvalidParam;
			}
			// Set server-SelectedElementsPositionDescriptor according to the client-SelectedElementsPositionDescriptor
			IList selectedElements;
			string errorMessage;
			UserDefinedMessageObservation.CreateSelectedElementsPositionDescriptorResult res = udm.CreateSelectedElementsPositionDescriptor(
				clientElementsPositionDescriptor,
				text,
				searchDir,
				out selectedElements,
				out errorMessage);
			if(res != CreateSelectedElementsPositionDescriptorResult.Ok) 
			{
				resultUdm = null;
				errorMsg = "Fr die clientseitige Selektionsbeschreibung konnte serverseitig keine bereinstimmung gefunden werden.";
				return CreateUserDefinedMessageObservationResult.NoMatch;
			}
			// Do selection and apply searchPattern if enabled
			string resultText;
			UserDefinedMessageObservation.GetUserDefinedMessageResult getRes = udm.GetUserDefinedMessage(text, out resultText,out errorMessage);
			switch(getRes) 
			{
				case UserDefinedMessageObservation.GetUserDefinedMessageResult.Ok:
					udm.changeState = ChangeStateEnum.New;
					udm.resultState = UserDefinedMessageObservation.ResultStateEnum.Ok;
					break;
				case UserDefinedMessageObservation.GetUserDefinedMessageResult.NoRegExMatch:
					udm.changeState = ChangeStateEnum.New;
					udm.resultState = UserDefinedMessageObservation.ResultStateEnum.NoSearchPatternMatch;
					break;
				case UserDefinedMessageObservation.GetUserDefinedMessageResult.RegExParamNotFound:
					udm.changeState = ChangeStateEnum.New;
					udm.resultState = UserDefinedMessageObservation.ResultStateEnum.MessageTemplateParamNotFound;
					break;
				case UserDefinedMessageObservation.GetUserDefinedMessageResult.NoSelectionMatch:
					udm.changeState = ChangeStateEnum.New;
					udm.resultState = UserDefinedMessageObservation.ResultStateEnum.NoSelectionMatch;
					break;
				case UserDefinedMessageObservation.GetUserDefinedMessageResult.Error:
					udm.changeState = ChangeStateEnum.Error;
					udm.resultState = UserDefinedMessageObservation.ResultStateEnum.Error;
					break;
			}
			udm.resultText = resultText;
			resultUdm = udm;
			errorMsg = null;
			return CreateUserDefinedMessageObservationResult.Ok;
		} // end: CreateUserDefinedMessageObservation

		internal static PreviewResult GetPreview(
			string url,
			bool searchPatternEnabled,
			string searchPattern,
			string messageTemplate,
			string selectedElementsPositionDescriptor,
			out string previewText,
			out string errorMsg) 
		{
			// CheckParameters
			if(url == null) 
			{
				errorMsg = "Es muss eine Url angegeben werden.";
				previewText = null;
				return PreviewResult.InvalidParams;
			}
			if(searchPatternEnabled && (searchPattern == null || messageTemplate == null)) 
			{
				errorMsg = "Es muss ein Suchmuster und eine Meldungsvorlage angegeben werden - da aktiviert.";
				previewText = null;
				return PreviewResult.InvalidParams;
			}
			try { Uri myUrl = new Uri(url); } 
			catch 
			{
				errorMsg = "Es muss eine gltige Url angegeben werden.";
				previewText = null;
				return PreviewResult.InvalidParams;
			}
			// assertion: parameters passed initial check phase
			// set parameters as UserDefinedMessageObservation properties
			UserDefinedMessageObservation udmObj = new UserDefinedMessageObservation();
			udmObj.searchPatternEnabled = searchPatternEnabled;
			udmObj.searchPattern = searchPattern;
			udmObj.messageTemplate = messageTemplate;
			udmObj.selectedElementsPositionDescriptor = (selectedElementsPositionDescriptor==null)?"-":selectedElementsPositionDescriptor;
			// Retrieve sourceHtmlText
			HttpHtmlRetriever retriever = new HttpHtmlRetriever();
			string sourceText, errorMessage;
			HttpHtmlRetriever.RetrieveResult retrieveResult = retriever.Retrieve(url, out sourceText, out errorMessage);
			if(retrieveResult != HttpHtmlRetriever.RetrieveResult.Ok) 
			{
				errorMsg = errorMessage;
				previewText = null;
				return PreviewResult.RetrieveError;
			}
			string resultText;
			UserDefinedMessageObservation.GetUserDefinedMessageResult res = udmObj.GetUserDefinedMessage(
				sourceText, 
				out resultText,
				out errorMessage);
			switch(res) 
			{
				case GetUserDefinedMessageResult.Ok:
					errorMsg = null;
					previewText = resultText;
					return PreviewResult.Ok;
				case GetUserDefinedMessageResult.NoRegExMatch:
					errorMsg = errorMessage;
					previewText = null;
					return PreviewResult.NoRegExMatch;
				case GetUserDefinedMessageResult.RegExParamNotFound:
					errorMsg = errorMessage;
					previewText = null;
					return PreviewResult.RegExParamNotFound;
				case GetUserDefinedMessageResult.NoSelectionMatch:
					errorMsg = errorMessage;
					previewText = null;
					return PreviewResult.NoSelectionMatch;
				case GetUserDefinedMessageResult.Error:
					errorMsg = errorMessage;
					previewText = null;
					return PreviewResult.Error;
			}
			errorMsg = "Es ist ein Fehler aufgetreten.";
			previewText = null;
			return PreviewResult.Error;
		} // end: method GetPreview

		internal override CheckResult Check(int newCheckCounterTimeStamp, out string errorMsg) 
		{
			HttpHtmlRetriever retriever = new HttpHtmlRetriever();
			string sourceText, errorMessage;
			HttpHtmlRetriever.RetrieveResult retrieveResult = retriever.Retrieve(this.url.ToString(), out sourceText, out errorMessage);
			if(retrieveResult != HttpHtmlRetriever.RetrieveResult.Ok) 
			{
				this.changeState = ChangeStateEnum.Error;
				this.date = DateTime.Now;
				this.checkCounterTimeStamp = newCheckCounterTimeStamp;
				this.resultText = null;
				this.resultState = UserDefinedMessageObservation.ResultStateEnum.Error;
				errorMsg = errorMessage;
				return CheckResult.Error;
			}
			string resultText;
			UserDefinedMessageObservation.GetUserDefinedMessageResult res = this.GetUserDefinedMessage(sourceText, out resultText,out errorMessage);
			switch(res) 
			{
				case GetUserDefinedMessageResult.Ok:
					if(this.resultText == null) // if no resultText to compare to - New Observation
						this.changeState = ChangeStateEnum.New;
					else 
					{
						// Check if resultText and therefore changeState changed
						if(this.resultText.Equals(resultText))
							this.changeState = ChangeStateEnum.Unchanged;
						else
							this.changeState = ChangeStateEnum.Changed;
					}
					this.resultState = UserDefinedMessageObservation.ResultStateEnum.Ok;
					break;
				case GetUserDefinedMessageResult.NoRegExMatch:
					this.changeState = ChangeStateEnum.Changed;
					this.resultState = UserDefinedMessageObservation.ResultStateEnum.NoSearchPatternMatch;
					break;
				case GetUserDefinedMessageResult.RegExParamNotFound:
					this.changeState = ChangeStateEnum.Changed;
					this.resultState = UserDefinedMessageObservation.ResultStateEnum.MessageTemplateParamNotFound;
					break;
				case GetUserDefinedMessageResult.NoSelectionMatch:
					this.changeState = ChangeStateEnum.Changed;
					this.resultState = UserDefinedMessageObservation.ResultStateEnum.NoSelectionMatch;
					break;
				case GetUserDefinedMessageResult.Error:
					this.changeState = ChangeStateEnum.Error;
					this.resultState = UserDefinedMessageObservation.ResultStateEnum.Error;
					break;
			}
			this.resultText = resultText;
			this.date = DateTime.Now;
			this.checkCounterTimeStamp = newCheckCounterTimeStamp;
			errorMsg = null;
			return CheckResult.Ok;
		}

		/******************************************************************************************************
		 *****                Getter Properties for UserDefinedMessageObservation fields                 ******
		 ******************************************************************************************************/
		internal bool SearchPatternEnabled 
		{
			get { return this.searchPatternEnabled; }
		}

		internal string SearchPattern 
		{
			get { return this.searchPattern; }
		}
		
		internal string MessageTemplate
		{
			get { return this.messageTemplate; }
		}

		internal string SelectedElementsPositionDescriptor 
		{
			get { return this.selectedElementsPositionDescriptor; }
		}

		internal string ResultText 
		{
			get { return this.resultText; }
		}

		internal UserDefinedMessageObservation.ResultStateEnum ResultState 
		{
			get { return this.resultState; }
		}

		internal void SetId(int id) 
		{
			if(id < 0)
				throw new Exception("Invalid Arguments passed to SetId.");
			this.id = id;
		}

        /// <summary>
		/// Sets/changes searchPattern only in case of successful check.
		/// </summary>
		/// <param name="pattern"></param>
		/// <param name="errorMessage"></param>
		/// <returns></returns>
		private bool SetSearchPattern(string pattern, out string errorMessage) 
		{	
			try 
			{
				this.searchPatternRegex = new Regex(pattern,myRegexOptions);
			} 
			catch(Exception e) 
			{
				errorMessage = e.Message;
				return false;
			}
			errorMessage = "";
			this.searchPattern = pattern;
			return true;
		}
		
		/// <summary>
		/// Sets the messageTemplate to the one passed, but only if it is syntactically correct.
		/// The syntactical correctnes is checked by CheckMessageTemplate.
		/// Basically each string can be used as message-template, but what get's checked is the correct
		/// syntax of used search-pattern-parameters. Note, that only the syntax get's checked, but not if
		/// a corresponding search-pattern-parameter exits at all.
		/// Search-pattern-parameters can be indicated by the following syntax:
		/// $0,$1, ... ,$10, ...   in case of unnamed capture groups
		/// ${name} in case of named capture groups, where each character of name is in the characterclass [-_+ a-zA-z0-9]
		/// </summary>
		/// <param name="template"></param>
		/// <param name="errorMessage"></param>
		/// <returns>
		/// true .. messageTemplate syntactical correct - got set.
		/// false .. messageTemplate not syntactical correct - errorMessage holds error description.
		/// </returns>
		private bool SetMessageTemplate(string messageTemplate, out string errorMessage) 
		{	
			string errorMsg;
			bool correct = CheckMessageTemplate(messageTemplate, out errorMsg);
			if(!correct) 
			{
				errorMessage = errorMsg;
				return false;
			}
			// assertion: messageTemplate is syntactically correct
			this.messageTemplate = messageTemplate;
			errorMessage = null;
			return true;
		} // end: method SetMessageTemplate

		private bool SetSelectedElementsPositionDescriptor(string elementsPositionDescriptor, out string errorMessage) 
		{
			if(elementsPositionDescriptor == null)
				elementsPositionDescriptor = "-";
			if(!CheckSelectedElementsPositionDescriptor(elementsPositionDescriptor, out errorMessage))
				return false;
			this.selectedElementsPositionDescriptor = elementsPositionDescriptor;
			return true;
		}

		internal enum ChangePropertiesEnum {Change, NoChange, Inconsistence}

		internal UserDefinedMessageObservation.ChangePropertiesEnum ChangeProperties(
			string name, 
			Uri url,
			bool searchPatternEnabled,
			string searchPattern,
			string messageTemplate,
			string selectedElementsPositionDescriptor) 
		{
			if(name == null || url == null)
				return ChangePropertiesEnum.Inconsistence;
			if(searchPatternEnabled && (searchPattern == null || messageTemplate == null))
				return ChangePropertiesEnum.Inconsistence;
			bool changed = false;
			if(!this.name.Equals(name) || !this.url.Equals(url))
				changed = true;
			if(this.searchPatternEnabled != searchPatternEnabled)
				changed = true;
			if(searchPatternEnabled)
				if(!searchPattern.Equals(this.searchPattern) || !messageTemplate.Equals(this.messageTemplate))
					changed = true;
			// Note: this.selectedElementsPositionDescriptor becomes "-" in case of set to null by SetSelectedElementsPositionDescriptor
			if(this.selectedElementsPositionDescriptor.Equals("-") && selectedElementsPositionDescriptor != null && !selectedElementsPositionDescriptor.Equals("-"))
				changed = true;
			if(!this.selectedElementsPositionDescriptor.Equals(selectedElementsPositionDescriptor))
				changed = true;
			if(!changed)
				return ChangePropertiesEnum.NoChange;
			// assertion: something changed
			string errorMsg;
			string tmpSearchPattern = this.searchPattern;
			string tmpMessageTemplate = this.messageTemplate;
			if(searchPatternEnabled) 
			{
				if(!this.SetSearchPattern(searchPattern, out errorMsg))
					return ChangePropertiesEnum.Inconsistence;
				if(!this.SetMessageTemplate(messageTemplate, out errorMsg)) 
				{
					this.searchPattern = tmpSearchPattern; // restore searchPattern
					return ChangePropertiesEnum.Inconsistence;
				}
			}
			if(!this.SetSelectedElementsPositionDescriptor(selectedElementsPositionDescriptor, out errorMsg)) 
			{
				this.searchPattern = tmpSearchPattern;
				this.messageTemplate = tmpMessageTemplate;
				return ChangePropertiesEnum.Inconsistence;
			}
			this.searchPatternEnabled = searchPatternEnabled;
			if(!this.searchPatternEnabled) 
			{
				this.searchPattern = null;
				this.messageTemplate = null;
			}
			this.name = name;
			this.url = url;
			this.date = Observation.InitDateTime;
			this.checkCounterTimeStamp = -1;
			this.changeState = ChangeStateEnum.Init;
			this.resultState = ResultStateEnum.Undefined;
			this.resultText = null;
			return ChangePropertiesEnum.Change;
		} // end: method ChangeProperties

		/*****************************************************************************************
		 ******************************* Syntactical check stuff *********************************
		 *****************************************************************************************/
		
		/// <summary>
		/// This method checks a searchPattern for syntactical correctness.
		/// Checks if a valid Regex-object can be obtained from the searchPattern string.
		/// </summary>
		/// <param name="searchPattern"></param>
		/// <param name="errorMessage"></param>
		/// <returns></returns>
		internal static bool CheckSearchPattern(string searchPattern, out string errorMessage) 
		{
			try 
			{
				new Regex(searchPattern,myRegexOptions);
			} 
			catch(Exception e) 
			{
				errorMessage = e.Message;
				return false;
			}
			errorMessage = "";
			return true;
		} // end: method CheckSearchPattern
		
		/// <summary>
		/// Checks the passed messageTemplate for syntactical correctness to be used as MessageTemplate for
		/// a UserDefinedMessageObservation.
		/// Basically each string can be used as message-template, but what get's checked is the correct
		/// syntax of used search-pattern-parameters. Note, that only the syntax get's checked, but not if
		/// a corresponding search-pattern-parameter exits at all.
		/// Search-pattern-parameters can be indicated by the following syntax:
		/// $0,$1, ... ,$10, ...   in case of unnamed capture groups
		/// ${name} in case of named capture groups, where each character of name is in the characterclass [-_+ a-zA-z0-9]
		/// </summary>
		/// <param name="messageTemplate"></param>
		/// <param name="errorMessage"></param>
		/// <returns></returns>
		internal static bool CheckMessageTemplate(string messageTemplate, out string errorMessage)  
		{
			bool lastWasDollar = false;
			bool inParameter = false;
			bool inBracelessParameter = false;
			char cur = '\0';
			for(int i=0; i<messageTemplate.Length; i++) 
			{
				cur = messageTemplate[i];
				if(Char.IsLetter(cur) || cur==' ' || cur=='-' || cur=='_' || cur=='+')  
				{	
					if(inBracelessParameter) {inBracelessParameter = false;}
					else if(lastWasDollar) {errorMessage="[Position: " + (i+1) + "] Dieses Zeichen ist nicht erlaubt als Parametername."; return false;}
				}
				else if(Char.IsDigit(cur)) 
				{
					if(lastWasDollar) {lastWasDollar = false; inBracelessParameter = true;}
				}
				else if(cur == '$') 
				{
					if(inParameter) {errorMessage="[Position: " + (i+1) + "] $-Zeichen nicht erlaubt in Parameternamen."; return false;}
					else if(inBracelessParameter) {inBracelessParameter = false;}
					else if(lastWasDollar) {lastWasDollar = false;}
					else lastWasDollar = true;
				} 
				else if(cur == '{') 
				{
					if(inParameter) {errorMessage="[Position: " + (i+1) + "] {-Zeichen nicht erlaubt in Parameternamen."; return false;}
					else if(inBracelessParameter) {inBracelessParameter = false;}
					else if(lastWasDollar) {lastWasDollar = false; inParameter = true;}
				}
				else if(cur == '}') 
				{
					if(inParameter) {inParameter = false;}
					if(inBracelessParameter) {inBracelessParameter = false;}
					if(lastWasDollar) {errorMessage="[Position: " + (i+1) + "] }-Zeichen nicht erlaubt als Parametername."; return false;}
				} 
				else 
				{	
					if(inBracelessParameter) {inBracelessParameter = false;}
					else 
					{
						if(inParameter) {errorMessage="[Position: " + (i+1) + "] Dieses Zeichen ist nicht erlaubt in Parameternamen."; return false;}
						if(lastWasDollar) {errorMessage="[Position: " + (i+1) + "] Dieses Zeichen ist nicht erlaubt als Parametername."; return false;}
					}
				}
			}
			if(lastWasDollar) 
			{
				errorMessage="[Position: " + messageTemplate.Length + "] Parameter muss einen Bezeichner haben."; return false;
			}
			else if(inParameter) 
			{
				errorMessage="[Position: " + messageTemplate.Length + "] Parameterklammer wurde nicht abgeschlossen."; return false;
			}
			errorMessage = "";
			return true;
		} // end: method CheckMessageTemplate

		/// <summary>
		/// This method checks a (server-side) selectedElementsPositionDescriptor for syntactical correctness
		/// and some basic (dependency) constraints of parts of the descriptor.
		/// 
		/// The selectedElementsPositionDescriptor must be of the following syntactical structure:
		///     ElementsPositionDescriptor = "-" | (id ";" depth ";" LayerDescriptor{depth} ">").
		///     LayerDescriptor = nodeName "," SearchDirection "," numSamePrevSib "," numSameNextSib ";".
		///     SearchDirection = "r" | "l" | "b".
		///   where:
		///     ElementsPositionDescriptor = "-" .. empty positionDescriptor which means Html-DOM-Body is selected as whole.
		///     id .. steady increasing non-negative integer identifier (the earlier selected the lower the id).
		///     depth .. depth of selected element in DOM-Tree with Body-node as root-node.
		///     nodeName .. corresponding to DOM-node-attribute nodeName.
		///     searchDirection .. direction to use when trying to find node from same-siblings count in corresponding layer.
		///                        "r" .. right (up from lower indexed siblings - numSamePrevSib to be used).
		///                        "l" .. left (down from higher indexed siblings - numSameNextSib to be used).
		///                        "b" .. both (up form lower and down from higher indexe siblings - numSamePrevSib and numSameNextSib to be used).
		///     numSamePrevSib .. number of same named previous siblings in current layer.
		///     numSameNextSib .. number of same named following siblings in current layer.
		/// </summary>
		/// <param name="elementsPositionDescriptor"></param>
		/// <param name="errorMessage"></param>
		/// <returns></returns>
		internal static bool CheckSelectedElementsPositionDescriptor(string elementsPositionDescriptor, out string errorMessage) 
		{
			if(elementsPositionDescriptor == null || elementsPositionDescriptor.Length == 0) 
			{
				errorMessage = "Es wurden keine Elemente als Selektion angegeben.";
			}
			if(elementsPositionDescriptor.Length == 1 && elementsPositionDescriptor[0] == '-') // whole Html-Body selected
			{
				errorMessage = "";
				return true;
			}
			// get elementsPositionDescriptor-parts for each selected element
			string[] wholeDescriptorParts = elementsPositionDescriptor.Split(new char[] {'>'});
			if(!wholeDescriptorParts[wholeDescriptorParts.Length - 1].Equals("")) 
			{
				errorMessage = "Element-Selektionsbeschreibung muss mit einem '>' abschlieen.";
				return false;
			}
			// check each selected element - note selectedElementsParts[length - 1] must be an empty string.
			for(int i = 0; i < wholeDescriptorParts.Length - 1; i++) 
			{
				string[] curSelectedElemParts = wholeDescriptorParts[i].Split(new char[] {';'});
				// path for each element must have an id and a depth
				if(curSelectedElemParts.Length < 3) 
				{
					errorMessage = "Element-Selektionsbeschreibung fr " + (i+1) + "-tes Element muss eine Id und eine Tiefenangabe haben.";
					return false;
				}
				// path for each element must end with a ';'
				if(curSelectedElemParts[curSelectedElemParts.Length - 1] != "")
				{
					errorMessage = "Element-Selektionsbeschreibung schliet fr " + (i+1) + "-tes Element nicht mit einem ';' ab";
					return false;
				}
				// path-id of each selected element must be a non-negative integer
				try{ int id = Int32.Parse(curSelectedElemParts[0]); if(id < 0) throw new Exception(); }
				catch 
				{
					errorMessage = "Element-Selektionsbeschreibung fr " + (i+1) + "-tes Element hat keine ganzzahlige nicht-negative Id.";
					return false;
				}
				// depth of each selected element must be a non-negative integer
				try { int sourceIndex = Int32.Parse(curSelectedElemParts[1]); if(sourceIndex < 0) throw new Exception(); }
				catch 
				{
					errorMessage = "Element-Selektionsbeschreibung fr " + (i+1) + "-tes Element hat keine ganzzahlige nicht-negative Tiefenangabe.";
					return false;
				}
				// assertion: id and depth are ok
				// check each layer of currently checked element
				for(int j = 2; j < curSelectedElemParts.Length - 1; j++) 
				{
					string[] curSelectedElemLayerParts = curSelectedElemParts[j].Split(new char[] {','});
					// layer-data must have 4 values:
					// nodeName, searchDirection, numSamePrevSib, numSameNextSib
					if(curSelectedElemLayerParts.Length != 4) 
					{
						errorMessage = "Element-Selektionsbeschreibung fr" + (i+1) + "-tes Element ist auf " + j + "-ter Ebene ungltig.";
						return false;
					}
					// searchDirection must be either 'r' or 'l' or 'b'
					if(curSelectedElemLayerParts[1].Length != 1 ||
						(curSelectedElemLayerParts[1][0] != 'r' &&
						curSelectedElemLayerParts[1][0] != 'l' &&
						curSelectedElemLayerParts[1][0] != 'b')) 
					{
						errorMessage = "Element-Suchrichtung fr " + (i+1) + "-tes Element hat auf " + j + "-ter Ebene einen undefinierten Wert.";
						return false;
					}
					int nsps = 0, nsns = 0;
					// numSamePrevSiblings must be a non-negative integer on currently checked layer
					try { nsps = Int32.Parse(curSelectedElemLayerParts[2]); if(nsps < 0) throw new Exception(); }
					catch 
					{ 
						errorMessage = "Element-Selektionsbeschreibung fr " + (i+1) + "-tes Element hat auf " + j + "-ter Ebene keine ganzzahlige nicht-negative Anzahl von Vorgngerzwillingen.";
						return false;
					}
					// numSameNextSiblings must be a non-negative integer on currently checked layer
					try { nsns = Int32.Parse(curSelectedElemLayerParts[3]); if(nsns < 0) throw new Exception(); }
					catch 
					{ 
						errorMessage = "Element-Selektionsbeschreibung fr " + (i+1) + "-tes Element hat auf " + j + "-ter Ebene keine ganzzahlige nicht-negative Anzahl von Nachfolgerzwillingen.";
						return false;
					}
				}
			}
			errorMessage = "";
			return true;
		} // end: method CheckSelectedElementsPositionDescriptor
		

		/******************************************************************************************
		 ******************************** Applicationlogic stuff **********************************
		 ******************************************************************************************/

		internal class MessageTemplateParameterToken 
		{	
			internal const int none		= 0, // no parameter-token
			                   numbered	= 1, // numbered parameter-token
			                   named	= 2, // named parameter-token
			                   eof		= 3; // eof reached-token

			internal int kind;
			internal string name;	// only if kind is numbered or named
			internal int number;	// only if kind is numbered
			internal int pos;		
			internal int length;	// only if kind is numbered or named
		}

		internal class MessageTemplateScanner 
		{	
			private const char eofCh = '\u0080';

			private TextReader r;
			private char ch;
			private int pos;

			internal void Init(TextReader r) 
			{
				this.r = r;
				this.pos = -1;
				this.nextCh();
			}

			private void nextCh() 
			{	
				int c = 0;
				if(this.ch != UserDefinedMessageObservation.MessageTemplateScanner.eofCh) 
				{
					if((c = this.r.Read()) >= 0)  
					{
						this.pos++;
						this.ch = (char)c;
					} 
					else 
					{	
						this.pos++;
						this.ch = MessageTemplateScanner.eofCh;
						r.Close();
					}
				}
			}

			internal UserDefinedMessageObservation.MessageTemplateParameterToken NextParameter() 
			{	
				UserDefinedMessageObservation.MessageTemplateParameterToken t = null;
				do 
				{
					while(this.ch != '$' && this.ch != UserDefinedMessageObservation.MessageTemplateScanner.eofCh)
						this.nextCh();
					t = new UserDefinedMessageObservation.MessageTemplateParameterToken();
					t.pos = this.pos;
					this.nextCh();
					switch(this.ch) 
					{	
						case '0': case '1': case '2': case '3': case '4': case '5': case '6': case '7': case '8':
						case '9':	
							t.kind = UserDefinedMessageObservation.MessageTemplateParameterToken.numbered;
							this.readNumbered(t);
							break;
						case '{':	
							t.kind = UserDefinedMessageObservation.MessageTemplateParameterToken.named;
							this.readNamed(t);
							break;
						case '$':	
							t.kind = UserDefinedMessageObservation.MessageTemplateParameterToken.none;
							this.nextCh();
							break;
						case UserDefinedMessageObservation.MessageTemplateScanner.eofCh:
							t.kind = UserDefinedMessageObservation.MessageTemplateParameterToken.eof;
							break;
						default:	
							t.kind = UserDefinedMessageObservation.MessageTemplateParameterToken.none;
							break;
					}
				} while(t.kind == UserDefinedMessageObservation.MessageTemplateParameterToken.none);
				return t;
			}

			internal void readNumbered(UserDefinedMessageObservation.MessageTemplateParameterToken t) 
			{
				StringBuilder sb = new StringBuilder();
				sb.Append(this.ch);
				this.nextCh();
				while(Char.IsDigit(this.ch)) 
				{
					sb.Append(this.ch);
					this.nextCh();
				}
				t.name = sb.ToString();
				t.number = Int32.Parse(t.name);
				t.length = this.pos - t.pos;
				
			}

			internal void readNamed(UserDefinedMessageObservation.MessageTemplateParameterToken t) 
			{
				StringBuilder sb = new StringBuilder();
				this.nextCh();
				while(Char.IsLetterOrDigit(this.ch) || this.ch == ' ' || this.ch == '-' || this.ch == '_' || this.ch == '+') 
				{
					sb.Append(this.ch);
					this.nextCh();
				}
				if(this.ch == '}') 
				{
					t.name = sb.ToString();
					this.nextCh();
					t.length = this.pos - t.pos;
				}
				else 
					t.kind = UserDefinedMessageObservation.MessageTemplateParameterToken.none;
			}
		} // end: class MessageTemplateScanner

		/// <summary>
		/// This method gets the user defined message defined by this object. First it applies the method
		/// GetUserSelectedElements whichs result determines the base-string for the second run in which
		/// the search pattern and the message template are applied.
		/// This second run only gets executed if this.searchPatternEnabled is set, otherwise the resulting
		/// string of GetUserSelectedElements forms the result.
		/// 
		/// The returned GetUserDefinedMessageResult indicates what problem occured if any occured or Ok if
		/// the method runs successfully.
		/// </summary>
		/// <param name="text"></param>
		/// <param name="userDefMsg">resulting string</param>
		/// <param name="errorMsg"></param>
		/// <returns></returns>
		private GetUserDefinedMessageResult GetUserDefinedMessage(string text, out string userDefMsg, out string errorMsg) 
		{
			userDefMsg = null;
			if(this.selectedElementsPositionDescriptor == null || this.selectedElementsPositionDescriptor.Length == 0) 
			{
				errorMsg = "Es konnte nur eine leere Element-Selektionsbeschreibung gefunden werden.";
				return GetUserDefinedMessageResult.Error;
			}
			if(this.searchPatternEnabled && this.searchPattern == null) 
			{
				errorMsg = "Es ist kein Suchmuster definiert, obwohl dieses aktiviert ist.";
				return GetUserDefinedMessageResult.Error;
			}
			if(this.searchPatternEnabled && this.messageTemplate == null) 
			{
				errorMsg = "Es ist keine Meldungsvorlage definiert, obwohl diese aktiviert ist.";
				return GetUserDefinedMessageResult.Error;
			}
			IList mySelectedElements = null;
			string errorMessage = null;
			GetUserSelectedElementsResult selectionResult =  this.GetUserSelectedElements(
				text, 
				out mySelectedElements, 
				out errorMessage);
			if(selectionResult == GetUserSelectedElementsResult.Error) 
			{
				errorMsg = errorMessage;
				return GetUserDefinedMessageResult.Error;
			}
			if(selectionResult != GetUserSelectedElementsResult.Match) 
			{
				errorMsg = errorMessage;
				return GetUserDefinedMessageResult.NoSelectionMatch;
			}
			StringBuilder tmpStringBuilder = new StringBuilder();
			IEnumerator mySelectedEnumerator = mySelectedElements.GetEnumerator();
			while(mySelectedEnumerator.MoveNext()) 
			{
				tmpStringBuilder.Append((string)mySelectedEnumerator.Current);
			}
			if(!searchPatternEnabled) 
			{
				userDefMsg = tmpStringBuilder.ToString();
				errorMsg = "";
				return GetUserDefinedMessageResult.Ok;
			}
			// assertion: searchPattern is enabled and must therefore also be used
			string resString = null;
			ApplySearchPatternResult applyResult = this.ApplySearchPattern(
				tmpStringBuilder.ToString(), 
				out resString, 
				out errorMessage);
			if(applyResult == ApplySearchPatternResult.NoMatch) 
			{
				errorMsg = errorMessage;
				return GetUserDefinedMessageResult.NoRegExMatch;
			}
			if(applyResult == ApplySearchPatternResult.ParamNotFound) 
			{
				errorMsg = errorMessage;
				return GetUserDefinedMessageResult.RegExParamNotFound;
			}
			if(applyResult != ApplySearchPatternResult.Ok) 
			{
				errorMsg = errorMessage;
				return GetUserDefinedMessageResult.Error;
			}
			// searchPattern and messageTemplate successfully applied.
			userDefMsg = resString;
			errorMsg = "";
			return GetUserDefinedMessageResult.Ok;
		} // end: method GetUserDefinedMessage

		/// <summary>
		/// Applies the regex corresponding to this.searchPattern to param text and sets the out param userDefMsg
		/// to a string resulting when each parameter in this.messageTemplate is replaced by the corresponding
		/// matched capturing group value of the regex.
		/// Numbered capturing groups correspond to message-template-parameters like $n1 or ${n2}, where n1 is a
		/// non-negative number with can contain leading zeros and n2 is a non-negative number which must not contain
		/// leading zeros.
		/// Named capturing groups correspond to message-template-parameters like ${x} where x is a name containing
		/// Letters, Digits, '-', '_', ' ' or '+'. 
		/// 
		/// Possible Errors:
		///		- this.searchPattern is null
		///		- this.messageTemplate is null
		///		- Regex object can't get generated (possibly a syntax error)
		///		- no match found
		///		- message template containes parameter which doesn't correspond to a capturing group of the regex
		///		- any error during apply of the Regex object to param text
		/// 
		/// Sideeffect:
		///		If this.searchPatternRegex is null, but this.searchPatter is not null then the corresponding Regex
		///		object is generated and assigned to this.searchPatternRegex.
		/// </summary>
		/// <param name="text">string on which the regex should applied to</param>
		/// <param name="userDefMsg">resulting user-defined-message</param>
		/// <param name="errorMsg">error message if NoMatch, Error or ParamNotFound</param>
		/// <returns>Ok or NoMatch or  ParamNotFound or Error</returns>
		public ApplySearchPatternResult ApplySearchPattern(string text, out string userDefMsg, out string errorMsg) 
		{
			Match matchObj;

			if(this.searchPattern == null) 
			{
				userDefMsg = null;
				errorMsg = "Kein Suchmuster definiert.";
				return ApplySearchPatternResult.Error;
			}
			if(this.messageTemplate == null) 
			{
				userDefMsg = null;
				errorMsg = "Keine Meldungsvorlage definiert.";
				return ApplySearchPatternResult.Error;
			}
			Regex regex = null;
			if(this.searchPatternRegex == null) 
			{
				try 
				{
					this.searchPatternRegex = regex = new Regex(this.searchPattern, myRegexOptions);
				} 
				catch(Exception e) 
				{
					userDefMsg = null;
					errorMsg = e.Message;
					return ApplySearchPatternResult.Error;
				}
			}
			regex = this.searchPatternRegex;
			try 
			{
				matchObj = regex.Match(text);
				if(matchObj.Success) 
				{	
					StringReader sr = new StringReader(this.messageTemplate);
					UserDefinedMessageObservation.MessageTemplateScanner scanner = new UserDefinedMessageObservation.MessageTemplateScanner();
					scanner.Init(sr);
					UserDefinedMessageObservation.MessageTemplateParameterToken t = scanner.NextParameter();
					StringBuilder sb = new StringBuilder();
					int pos = 0;
					string paramReplaceString = null;
					Group tmp = null;
					while(t.kind != UserDefinedMessageObservation.MessageTemplateParameterToken.eof) 
					{
						sb.Append(this.messageTemplate.Substring(pos,t.pos - pos));
						if(t.kind == UserDefinedMessageObservation.MessageTemplateParameterToken.numbered) 
						{
							tmp = matchObj.Groups[t.number];
							paramReplaceString = (tmp == null)?null:tmp.Value;
						}
						else 
						{
							tmp = matchObj.Groups[t.name];
							paramReplaceString = (tmp == null)?null:tmp.Value;
						}
						if(paramReplaceString == null) 
						{
							userDefMsg = null;
							errorMsg = "Parameter " + t.name + " hat keine Entsprechung im Suchmuster";
							return ApplySearchPatternResult.ParamNotFound;
						}
						sb.Append(paramReplaceString);
						pos = t.pos + t.length;
						t = scanner.NextParameter();
					}
					// append text after last parameter
					sb.Append(this.messageTemplate.Substring(pos,this.messageTemplate.Length - pos));
					userDefMsg = sb.ToString();
					errorMsg = "";
					return ApplySearchPatternResult.Ok;
				} 
				else 
				{
					userDefMsg = null;
					errorMsg = "Keine Musterbereinstimmung gefunden.";
					return ApplySearchPatternResult.NoMatch;
				}
			} 
			catch(Exception e)
			{	
				userDefMsg = null;
				errorMsg = e.Message;
				return ApplySearchPatternResult.Error;
			}
		} // end: method ApplySearchPatternResult
		
		/// <summary>
		/// Applies the regex corresponding to this.searchPattern to param text and sets the out param dictionary
		/// to a dictinary containing with the found capturing groups (key: capturing group name, value: capturing group match).
		/// In case of a unnamed and therefore numerical capturing group the corresponding index-number-string is the key, the whole
		/// match is therefore stored under the key "0".
		/// 
		/// Possible Errors:
		///		- this.searchPattern is null
		///		- Regex object can't get generated (possibly a syntax error)
		///		- no match found
		///		- any error during apply of the Regex object to param text
		/// 
		/// Sideeffect:
		///		If this.searchPatternRegex is null, but this.searchPatter is not null then the corresponding Regex
		///		object is generated and assigned to this.searchPatternRegex.
		/// </summary>
		/// <param name="text">string on which the regex should applied to</param>
		/// <param name="dictionary">parameter dictionary or null if no match or error</param>
		/// <param name="errorMsg">error message if no match or error</param>
		/// <returns>Match or NoMatch or Error</returns>
		internal UpdateSearchPatternParameterResult updateSearchPatternParameterDictionary(string text, out IDictionary dictionary, out string errorMsg) 
		{	
			Match matchObj;

			if(this.searchPattern == null) 
			{
				dictionary = null;
				errorMsg = "Kein Suchmuster definiert.";
				return UpdateSearchPatternParameterResult.Error;
			}
			Regex regex = null;
			if(this.searchPatternRegex == null) 
			{
				try 
				{
					this.searchPatternRegex = regex = new Regex(this.searchPattern, myRegexOptions);
				} 
				catch(Exception e) 
				{
					dictionary = null;
					errorMsg = e.Message;
					return UpdateSearchPatternParameterResult.Error;
				}
			}
			regex = this.searchPatternRegex;
			try 
			{
				matchObj = regex.Match(text);
				if(matchObj.Success) 
				{
					string[] parameters = regex.GetGroupNames();
					IDictionary tmpDictionary = new Hashtable(parameters.Length);
					for(int i=0; i<parameters.Length; i++) 
						tmpDictionary.Add(parameters[i],matchObj.Groups[parameters[i]].Value);
					dictionary = tmpDictionary;
					errorMsg = "";
					return UpdateSearchPatternParameterResult.Match;
				} 
				else 
				{
					dictionary = null;
					errorMsg = "Keine Musterbereinstimmung gefunden.";
					return UpdateSearchPatternParameterResult.NoMatch;
				}
			} 
			catch(Exception e) 
			{	
				dictionary = null;
				errorMsg = e.Message;
				return UpdateSearchPatternParameterResult.Error;
			}
		} // end: method updateSearchPatternParameterDictionary
		
		/// <summary>
		/// This method selects the Elements of a passed Html-text corresponding to the this.selectedElementsPositionDescriptor set.
		/// If any of the selectedElements specified by this.selectedElementsPositionDescriptor couldn't be found it returns with
		/// GetUserSelectedElementsResult.NoMatch and a corresponding error message.
		/// Otherwise if the search was successful it returns with GetUserSelectedElementsResult.Match and the out-parameter
		/// selectedElements references the List of selected Elements found in the this.selectedElementsPositionDescriptor
		/// corresponding order.
		/// If the Html-text passed is null GetUserSelectedElementsResult.Error and a corresponding error message is returned.
		/// 
		/// Note: This method assumes a synactically valid this.selectedElementsPositionDescriptor.
		/// 
		/// If there isn't an ElementSelectionScope set it returns with GetUserSelectedElementsResult.Error and a
		/// corresponding error message. The same happens if the Html-text passed is null.
		/// </summary>
		/// <param name="text">source Html.</param>
		/// <param name="selectedElements">List of selected Html-Elements in string-representation (with tags).</param>
		/// <param name="errorMsg">error message in case of not GetUserSelectedElementsResult.Match returned.</param>
		/// <returns></returns>
		internal GetUserSelectedElementsResult GetUserSelectedElements(string text, out IList selectedElements, out string errorMsg) 
		{
			selectedElements = null;
			if(this.selectedElementsPositionDescriptor == null || this.selectedElementsPositionDescriptor == "") 
			{
				errorMsg = "Es ist nur eine leere Element Selektionsbeschreibung vorhanden.";
				return GetUserSelectedElementsResult.Error;
			}
			if(text == null) 
			{
				errorMsg = "Es wurde kein Html-Quelltext bergeben.";
				return GetUserSelectedElementsResult.Error;
			}
			// assumption: this.selectedElementsPositionDescriptor is syntactically ok
			// generate SelectedElementsWrapper for comfortable access
			SelectedElementsWrapper wholeDescriptor = new SelectedElementsWrapper(this.selectedElementsPositionDescriptor);
			// Generate Html-DOM-Tree of html-source-text
			// Note: mshtml (Microsoft.mshtml.dll) wraps native COM-calls by using InteropServices.
			//       To catch possible exceptions tue to this the following code is put into a try/catch block.
			ArrayList mySelectedElements = new ArrayList();
			try 
			{
				// Load HtmlDocument from string
				IHTMLDocument2 htmlDoc = new HTMLDocumentClass();
				htmlDoc.write(new object[] {text});
				htmlDoc.close();

				IHTMLDOMNode bodyNode = (IHTMLDOMNode)htmlDoc.body;
				IHTMLDOMNode tmpNode = null;

				if(bodyNode == null) 
				{
					selectedElements = null;
					errorMsg = "Es konnte kein Html-Body-Element gefunden werden.";
					return GetUserSelectedElementsResult.NoMatch;
				}
				// Decomment next statement to log Html-Body-DOM-Tree
				// NOTE: Only decomment if debugging - because it's very time and space-consuming
				// log.Debug(HtmlTreeToString(bodyNode));
				
				// If whole body is selected return this as only selected Element
				if(wholeDescriptor.bodyOnlySelected) 
				{
					mySelectedElements.Add(((IHTMLElement)bodyNode).outerHTML);
					selectedElements = mySelectedElements;
					errorMsg = "";
					return GetUserSelectedElementsResult.Match;
				}

				// Search for each selected Element
				for(int i = 0; i < wholeDescriptor.selectedElements.Length; i++) 
				{
					tmpNode = bodyNode;
					SelectedElementWrapper curSelElem = wholeDescriptor.selectedElements[i];
					for(int j = 0; j < curSelElem.layers.Length; j++) 
					{
						SelectedElementLayerWrapper curLayer = curSelElem.layers[j];
						if(!this.findSelectedElementPartAtNextLayerByNumSibCorrToSearchDir(ref tmpNode, curLayer)) 
						{
							// No match for selected element on current layer
							selectedElements = null;
							errorMsg = (i+1) + "-tes selektiertes Element in " + j + "-ter Ebene nicht gefunden.";
							return GetUserSelectedElementsResult.NoMatch;
						}		
					}
					// Add selected element to result-list
					mySelectedElements.Add(((IHTMLElement)tmpNode).outerHTML);
				}
			}
			catch(System.Runtime.InteropServices.COMException ce) 
			{
				log.Error("A COM related exception occured within Microsoft.mshtml.",ce);
				selectedElements = null;
				errorMsg = "Es ist ein Verarbeitungsfehler aufgetreten.";
				return GetUserSelectedElementsResult.Error;
			} 
			catch(Exception e)
			{
				log.Error("Exception occured.",e);
				selectedElements = null;
				errorMsg = "Es ist ein Verarbeitungsfehler aufgetreten.";
				return GetUserSelectedElementsResult.Error;
			}
			selectedElements = mySelectedElements;
			errorMsg = "";
			return GetUserSelectedElementsResult.Match;
		} // end: method GetUserSelectedElements

		/// <summary>
		/// Helper-method for GetUserSelectedElements which finds the next-layer-DOM-node by the
		/// numSamePrevSib and numSameNextSib parts corresponding to the searchDir part of that
		/// layer (see selectedElementsPositionDescriptor).
		/// </summary>
		/// <param name="fatherNode">Gets changed to the next-layer-node if successfully found.</param>
		/// <param name="nextLayer"></param>
		/// <returns></returns>
		private bool findSelectedElementPartAtNextLayerByNumSibCorrToSearchDir(
			ref IHTMLDOMNode fatherNode,
			SelectedElementLayerWrapper nextLayer) 
		{
			IHTMLDOMChildrenCollection myChildCollection = (IHTMLDOMChildrenCollection)fatherNode.childNodes;
			int numChildes = myChildCollection.length;
			int leftIndex = 0;
			int rightIndex = 0;
			int i = 0;
			if(nextLayer.searchDirection != 'r') 
			{
				while(i < numChildes) 
				{
					if(leftIndex >= nextLayer.numSamePrevSib && ((IHTMLDOMNode)myChildCollection.item(i)).nodeName.ToLower().Equals(nextLayer.nodeName.ToLower()))
						break;
					if(((IHTMLDOMNode)myChildCollection.item(i)).nodeName.ToLower().Equals(nextLayer.nodeName.ToLower()))
						leftIndex++;
					i++;
				}
				if(i >= numChildes) // loop terminated because not enough previous siblings found - no match
					return false;
				if(nextLayer.searchDirection == 'l') // if searchDirection is left this is already sufficient 
				{
					fatherNode = (IHTMLDOMNode)myChildCollection.item(i);
					return true;
				}
				leftIndex = i; // remember index - later necessary to check both search Direction
			}
			if(nextLayer.searchDirection != 'l') 
			{
				i = numChildes - 1;
				while( i >= 0) 
				{
					if(rightIndex >= nextLayer.numSameNextSib && ((IHTMLDOMNode)myChildCollection.item(i)).nodeName.ToLower().Equals(nextLayer.nodeName.ToLower()))
						break;
					if(((IHTMLDOMNode)myChildCollection.item(i)).nodeName.ToLower().Equals(nextLayer.nodeName.ToLower()))
						rightIndex++;
					i--;
				}
				if(i < 0) // loop terminated because not enough next siblings found - no match
					return false;
				if(nextLayer.searchDirection == 'r') // if searchDirection is right this is already sufficient
				{
					fatherNode = (IHTMLDOMNode)myChildCollection.item(i);
					return true;
				}
				rightIndex = i; // remember index - later necessary to check both search Direction
			}
			// assertion: searchDirection is both ('b' of not 'l' or 'r' because 'b' is default) - otherwise method already returned
			if(leftIndex == rightIndex) 
			{ // left- and right-search found same Element - both Direction ok
				fatherNode = (IHTMLDOMNode)myChildCollection.item(leftIndex);
				return true;
			}
			return false;
		}

		/******************************************************************************************
		 ***************** Client-side selectedElementsPositionDescriptor stuff *******************
		 ******************************************************************************************/
		
		/// <summary>
		/// This method checks a clientSelectedElementsPositionDescriptor for syntactical correctness and
		/// some basic (dependency) constraints of parts of the descriptor.
		/// 
		/// The clientSelectedElementsPositionDescriptor must be of the following syntactical structure:
		///     ClientElementsPositionDescriptor = id ";" depth ";" ClientLayerDescriptor{depth} ">".
		///     ClientLayerDescriptor = nodeName "," numSamePrevSib "," numSameNextSib "," elemCountRelToBody "," sameElemCountRelToBody "," sourceIndexRelToBody ";".
		///   where:
		///     id .. steady increasing non-negative integer identifier (the earlier selected the lower the id).
		///     depth .. depth of selected element in DOM-Tree with Body-node as root-node.
		///     nodeName .. corresponding to DOM-node-attribute nodeName.
		///     numSamePrevSib .. number of same named previous siblings in current layer.
		///     numSameNextSib .. number of same named following siblings in current layer.
		///     elemCountRelToBody .. count-index of current-layer-element relative to the Body-node (nodes of type elementnode (typeid 1) count - body itself has index 0)
		///     sameElemCountRelToFather .. count-index of current-layer-element relative to Father-node (only same named elementnodes count - father itself has index 0 whether same named or not)
		///     sourceIndexRelToBody .. DOM-node-attribute sourceIndex relative to that of document.body (only available in MS Internet Explorer), if not defined sourceIndexRelToBody is -1. 
		/// </summary>
		/// <param name="template"></param>
		/// <param name="errorMessage"></param>
		/// <returns></returns>
		internal static bool CheckClientSelectedElementsPositionDescriptor(string clientElementsPositionDescriptor, out string errorMessage) 
		{
			if(clientElementsPositionDescriptor == null || clientElementsPositionDescriptor.Length == 0) 
			{
				errorMessage = "Es wurden keine Elemente als Selektion angegeben.";
				return false;
			}
			// get elementsPositionDescriptor-parts for each selected element
			string[] wholeDescriptorParts = clientElementsPositionDescriptor.Split(new char[] {'>'});
			if(!wholeDescriptorParts[wholeDescriptorParts.Length - 1].Equals("")) 
			{
				errorMessage = "Element-Selektionsbeschreibung muss mit einem '>' abschlieen.";
				return false;
			}
			// check each selected element - note selectedElementsParts[length - 1] must be an empty string.
			for(int i = 0; i < wholeDescriptorParts.Length - 1; i++) 
			{
				string[] curSelectedElemParts = wholeDescriptorParts[i].Split(new char[] {';'});
				// path for each element must have an id and a depth
				if(curSelectedElemParts.Length < 3) 
				{
					errorMessage = "Element-Selektionsbeschreibung fr " + (i+1) + "-tes Element muss eine Id und eine Tiefenangabe haben.";
					return false;
				}
				// path for each element must end with a ';'
				if(curSelectedElemParts[curSelectedElemParts.Length - 1] != "") 
				{
					errorMessage = "Element-Selektionsbeschreibung schliet fr " + (i+1) + "-tes Element nicht mit einem ';' ab";
					return false;
				}
				// path-id of each selected element must be a non-negative integer
				try { int id = Int32.Parse(curSelectedElemParts[0]); if(id < 0) throw new Exception(); }
				catch 
				{ 
					errorMessage = "Element-Selektionsbeschreibung fr " + (i+1) + "-tes Element hat keine ganzzahlige nicht-negative Id.";
					return false;
				}
				// depth of each selected element must be a non-negative integer
				try { int sourceIndex = Int32.Parse(curSelectedElemParts[1]); if(sourceIndex < 0) throw new Exception(); }
				catch 
				{
					errorMessage = "Element-Selektionsbeschreibung fr " + (i+1) + "-tes Element hat keine ganzzahlinge nicht-negative Tiefenangabe.";
					return false;
				}
				// assertion: id and depth are ok
				// check each layer of currently checked element
				for(int j = 2; j < curSelectedElemParts.Length -1; j++) 
				{
					string[] curSelectedElemLayerParts = curSelectedElemParts[j].Split(new char[] {','});
					// layer-data must have 6 values:
					//   nodeName, numSamePrevSib, numSameNextSib, elemCountRelToBody, sameElemCountRelToBody, sourceIndexRelToBody
					if(curSelectedElemLayerParts.Length != 6) 
					{
						errorMessage = "Element-Selektionsbeschreibung fr " + (i+1) + "-tes Element ist auf " + j + "-ter Ebene ungltig.";
						return false;
					}
					int nsps = 0, nsns = 0, ecrtb = 0, secrtf = 0, sirtb = 0;
					// numSamePrevSiblings must be a non-negative integer on currently checked layer
					try { nsps = Int32.Parse(curSelectedElemLayerParts[1]); if(nsps < 0) throw new Exception(); }
					catch 
					{ 
						errorMessage = "Element-Selektionsbeschreibung fr " + (i+1) + "-tes Element hat auf " + j + "-ter Ebene keine ganzzahlige nicht-negative Anzahl von Vorgngerzwillingen.";
						return false;
					}
					// numSameNextSiblings must be a non-negative integer on currently checked layer
					try { nsns = Int32.Parse(curSelectedElemLayerParts[2]); if(nsns < 0) throw new Exception(); }
					catch 
					{ 
						errorMessage = "Element-Selektionsbeschreibung fr " + (i+1) + "-tes Element hat auf " + j + "-ter Ebene keine ganzzahlige nicht-negative Anzahl von Nachfolgerzwillingen.";
						return false;
					}
					// elemCountRelToBody must be a non-negative integer on currently checked layer
					try { ecrtb = Int32.Parse(curSelectedElemLayerParts[3]); if(ecrtb < 0) throw new Exception(); }
					catch 
					{ 
						errorMessage = "Element-Selektionsbeschreibung fr " + (i+1) + "-tes Element hat auf " + j + "-ter Ebene keinen ganzzahlige nicht-negativen Index relativ zum Html-Body-Tag.";
						return false;
					}
					// sameElemCountRelToFather must be a non-negative integer on currently checked layer
					try { secrtf = Int32.Parse(curSelectedElemLayerParts[4]); if(secrtf < 0) throw new Exception(); }
					catch 
					{ 
						errorMessage = "Element-Selektionsbeschreibung fr " + (i+1) + "-tes Element hat auf " + j + "-ter Ebene keinen ganzzahlige nicht-negativen Index gleicher Elemente relativ zum Html-Body-Tag.";
						return false;
					}
					// sameElemCountRelToBody must be smaller or equal to elemCountRelToBody
					if(secrtf > ecrtb) 
					{
						errorMessage = "Element-Selektionsbeschreibung fr " + (i+1) + "-tes Element hat auf " + j + "-ter Ebene inkonsistente Indexangaben bezglich relativer Position zum Html-Body-Tag.";
						return false;
					}
					// sourceIndexRelToBody must be a non-negative integer or -1 on currently checked layer
					try { sirtb = Int32.Parse(curSelectedElemLayerParts[5]); if(sirtb < -1) throw new Exception(); }
					catch 
					{ 
						errorMessage = "Element-Selektionsbeschreibung fr " + (i+1) + "-tes Element hat auf " + j + "-ter Ebene keine erlaubte Quellindexangabe.";
						return false;
					}
				}
			}
			errorMessage = "";
			return true;
		} // end: method CheckClientSelectedElementsPositionDescriptor
		
		/// <summary>
		/// Wrapper-class for a whole client-side selectedElementsPositionDescriptor.
		/// </summary>
		internal class ClientSelectedElementsWrapper 
		{
			internal ClientSelectedElementWrapper[] selectedElements;
			// Note: selectedElementsPositionDescriptor must be syntactical valid
			internal ClientSelectedElementsWrapper(string clientSelectedElementsPositionDescriptor) 
			{
				string[] wholeDescriptorParts = clientSelectedElementsPositionDescriptor.Split(new char[] {'>'});
				this.selectedElements = new ClientSelectedElementWrapper[wholeDescriptorParts.Length - 1];
				for(int i = 0; i < wholeDescriptorParts.Length - 1; i++) 
				{
					string[] curSelectedElemParts = wholeDescriptorParts[i].Split(new char[] {';'});
					ClientSelectedElementWrapper tmpSEW = new ClientSelectedElementWrapper();
					tmpSEW.id = Int32.Parse(curSelectedElemParts[0]);
					tmpSEW.depth = Int32.Parse(curSelectedElemParts[1]);
					tmpSEW.layers = new ClientSelectedElementLayerWrapper[curSelectedElemParts.Length - 3];
					for(int j = 2; j < curSelectedElemParts.Length - 1; j++) 
					{
						string[] curSelectedElemLayerParts = curSelectedElemParts[j].Split(new char[] {','});
						ClientSelectedElementLayerWrapper tmpSELW = new ClientSelectedElementLayerWrapper();
						tmpSELW.nodeName = curSelectedElemLayerParts[0];
						tmpSELW.numSamePrevSib = Int32.Parse(curSelectedElemLayerParts[1]);
						tmpSELW.numSameNextSib = Int32.Parse(curSelectedElemLayerParts[2]);
						tmpSELW.elemCountRelToBody = Int32.Parse(curSelectedElemLayerParts[3]);
						tmpSELW.sameElemCountRelToFather = Int32.Parse(curSelectedElemLayerParts[4]);
						tmpSELW.sourceIndexRelToBody = Int32.Parse(curSelectedElemLayerParts[5]);
						tmpSEW.layers[j - 2] = tmpSELW;
					}
					this.selectedElements[i] = tmpSEW;
				}
			}
			public override string ToString() 
			{
				StringBuilder sb = new StringBuilder();
				for(int i = 0; i < selectedElements.Length; i++)
					sb.Append(selectedElements[i].ToString());
				return sb.ToString();
			}
		}

		/// <summary>
		/// Wrapper class for one selected element in a whole client-side selectedElementsPositionDescriptor.
		/// </summary>
		internal class ClientSelectedElementWrapper 
		{	
			internal int id;
			internal int depth;
			internal ClientSelectedElementLayerWrapper[] layers;

			public override string ToString() 
			{
				StringBuilder sb = new StringBuilder();
				sb.Append(id + ";" + depth + ";");
				for(int i = 0; i < layers.Length; i++)
					sb.Append(layers[i].ToString());
				sb.Append(">");
				return sb.ToString();
			}
		}

		/// <summary>
		/// Wrapper class for one selection-layer in a selected element as part of a whole
		/// client-side selectedElementsPositionDescriptor.
		/// </summary>
		internal class ClientSelectedElementLayerWrapper 
		{
			internal string nodeName;
			internal int numSamePrevSib;
			internal int numSameNextSib;
			internal int elemCountRelToBody;
			internal int sameElemCountRelToFather;
			internal int sourceIndexRelToBody;

			public override string ToString() 
			{
				return nodeName + "," + numSamePrevSib + "," + numSameNextSib + "," + elemCountRelToBody + "," + sameElemCountRelToFather + "," + sourceIndexRelToBody + ";";
			}
		}

		/// <summary>
		/// This method creates the server-side selectedElementsPositionDescriptor out of the client-side
		/// clientSelectedElementsPositionDescriptor.
		/// Note: The resulting server-side selectedElementsPositionDescriptor gets stored in the field
		///       this.selectedElementsPositionDescriptor.
		/// 
		/// The clientSelectedElementsPositionDescriptor is the one submited by the client-side browser-scipt, and
		/// determines the user selected elements of a html-page in a redundant way.
		/// This redundancy allows it to avoid wrong selections because of possible differences between the 
		/// DOM-Tree of the client (Browser-DOM-engine) and that one created at the server side (mshtml).
		/// 
		/// The clientside selectedElementsPositionDescriptor must be of the following syntactical structure:
		///     ElementsPositionDescriptor = id ";" depth ";" LayerDescriptor{depth} ">".
		///     LayerDescriptor = nodeName "," numSamePrevSib "," numSameNextSib "," elemCountRelToBody "," sameElemCountRelToBody "," sourceIndexRelToBody ";".
		///   where:
		///     id .. steady increasing non-negative integer identifier (the earlier selected the lower the id).
		///     depth .. depth of selected element in DOM-Tree with Body-node as root-node.
		///     nodeName .. corresponding to DOM-node-attribute nodeName.
		///     numSamePrevSib .. number of same named previous siblings in current layer.
		///     numSameNextSib .. number of same named following siblings in current layer.
		///     elemCountRelToBody .. count-index of current-layer-element relative to the Body-node (nodes of type elementnode (typeid 1) count - body itself has index 0)
		///     sameElemCountRelToFather .. count-index of current-layer-element relative to Father-node (only same named elementnodes count - father itself has index 0 whether same named or not)
		///     sourceIndexRelToBody .. DOM-node-attribute sourceIndex relative to that of document.body (only available in MS Internet Explorer), if not defined sourceIndexRelToBody is -1. 
		/// </summary>
		/// <param name="clientElementsPositionDescriptor"></param>
		/// <param name="text"></param>
		/// <param name="selectedElements"></param>
		/// <param name="errorMsg"></param>
		/// <returns></returns>
		private CreateSelectedElementsPositionDescriptorResult CreateSelectedElementsPositionDescriptor(
			string clientElementsPositionDescriptor, 
			string text, 
			ServerSearchDirection searchDir,
			out IList selectedElements, 
			out string errorMsg) 
		{
			selectedElements = null;
			if(clientElementsPositionDescriptor == null) 
			{
				errorMsg = "Es wurde keine Client-Selektionsbeschreibung bergeben.";
				return CreateSelectedElementsPositionDescriptorResult.InvalidParam;
			}
			if(text == null) 
			{
				errorMsg = "Es wurde kein Html-Quelltext bergeben.";
				return CreateSelectedElementsPositionDescriptorResult.InvalidParam;
			}
			string tmpErrorMsg = null;
			if(!CheckClientSelectedElementsPositionDescriptor(clientElementsPositionDescriptor, out tmpErrorMsg)) 
			{
				errorMsg = "Client-Selektionsbeschreibung ist syntaktisch inkorrekt.";
				return CreateSelectedElementsPositionDescriptorResult.InvalidParam;
			}
			// assumption: clientElementsPositionDescriptor is syntactically ok
			// generate SelectedElementsWrapper for comfortable access
			ClientSelectedElementsWrapper wholeDescriptor = new ClientSelectedElementsWrapper(clientElementsPositionDescriptor);
			// Generate Html-DOM-Tree of html-source-text
			// Note: mshtml (Microsoft.mshtml.dll) wraps native COM-calls by using InteropServices.
			//       To catch possible exceptions tue to this the following code is put into a try/catch block.
			ArrayList mySelectedElements = new ArrayList();
			try 
			{
				// Load HtmlDocument from string
				IHTMLDocument2 htmlDoc = new HTMLDocumentClass();
				htmlDoc.write(new object[] {text});
				htmlDoc.close();
				
				IHTMLDOMNode bodyNode = (IHTMLDOMNode)htmlDoc.body;
				IHTMLDOMNode tmpNode = null;

				if(bodyNode == null) 
				{
					selectedElements = null;
					errorMsg = "Es konnte kein Html-Body-Element gefunden werden.";
					return CreateSelectedElementsPositionDescriptorResult.NoMatch;
				}
				// Decomment next statement to log Html-Body-DOM-Tree
				// NOTE: Only decomment if debugging - because it's very time and space-consuming
				// log.Debug(HtmlTreeToString(bodyNode));
		
				SelectedElementsWrapper serverSelElems = new SelectedElementsWrapper();
				serverSelElems.selectedElements = new SelectedElementWrapper[wholeDescriptor.selectedElements.Length];

				// Search for each selected Element and create server-side elementsPositionDescriptor by the way.
				for(int i = 0; i < wholeDescriptor.selectedElements.Length; i++) 
				{
					tmpNode = bodyNode;
					ClientSelectedElementWrapper curSelElem = wholeDescriptor.selectedElements[i];
					for(int j = 0; j < curSelElem.layers.Length; j++) 
					{
						ClientSelectedElementLayerWrapper curLayer = curSelElem.layers[j];
						if(!this.findClientSelectedElementPartAtNextLayerByNumSib(ref tmpNode, curLayer))
							if(!this.findClientSelectedElementPartAtNextLayerBySameElemCountRelToFather(ref tmpNode, curLayer))
								if(!this.findClientSelectedElementPartAtNextLayerBySourceIndex(ref tmpNode, bodyNode, curLayer)) 
								{
									// No match for selected element on current layer
									selectedElements = null;
									errorMsg = (i+1) + "-tes selektiertes Element in " + j + "-ter Ebene nicht gefunden.";
									return CreateSelectedElementsPositionDescriptorResult.NoMatch;
								}
					}
					// Create server-side selectedElementWrapper for current selected Element and append it to selectedElementsWrapper.
					if(!createSelectedElementWrapperForNode(out serverSelElems.selectedElements[i], tmpNode, bodyNode, curSelElem.id, searchDir)) 
					{
						selectedElements = null;
						errorMsg = "Beim serverseitigen erstellen der Selektionsbeschreibung ist ein Fehler aufgetreten.";
						return CreateSelectedElementsPositionDescriptorResult.Error;
					}
					// Add selected element to result-list
					mySelectedElements.Add(((IHTMLElement)tmpNode).outerHTML);
				}
				// set whole server-side selectedElementsDescriptor by wrappers ToString method.
				this.selectedElementsPositionDescriptor = serverSelElems.ToString();
			}
			catch(System.Runtime.InteropServices.COMException ce) 
			{
				log.Error("A COM related exception occured within Microsoft.mshtml.",ce);
				selectedElements = null;
				errorMsg = "Es ist ein Verarbeitungsfehler aufgetreten.";
				return CreateSelectedElementsPositionDescriptorResult.Error;
			} 
			catch(Exception e)
			{
				log.Error("Exception occured.",e);
				selectedElements = null;
				errorMsg = "Es ist ein Verarbeitungsfehler aufgetreten.";
				return CreateSelectedElementsPositionDescriptorResult.Error;
			}
			selectedElements = mySelectedElements;
			errorMsg = "";
			return CreateSelectedElementsPositionDescriptorResult.Ok;
		} // end: method CreateSelectedElementsPositionDescriptor
		
		/// <summary>
		/// Wrapper-class for a whole server-side selectedElementsPositionDescriptor.
		/// </summary>
		internal class SelectedElementsWrapper 
		{
			internal bool bodyOnlySelected;
			internal SelectedElementWrapper[] selectedElements;

			internal SelectedElementsWrapper() { this.bodyOnlySelected = false;}
			// Note: selectedElementsPositionDescriptor must be syntactical valid
			internal SelectedElementsWrapper(string selectedElementsPositionDescriptor) 
			{
				if(selectedElementsPositionDescriptor.Length == 1 && selectedElementsPositionDescriptor[0] == '-') 
				{
					this.selectedElements = new SelectedElementWrapper[0];
					this.bodyOnlySelected = true;
				}
				string[] wholeDescriptorParts = selectedElementsPositionDescriptor.Split(new char[] {'>'});
				this.selectedElements = new SelectedElementWrapper[wholeDescriptorParts.Length - 1];
				for(int i = 0; i < wholeDescriptorParts.Length - 1; i++) 
				{
					string[] curSelectedElemParts = wholeDescriptorParts[i].Split(new char[] {';'});
					SelectedElementWrapper tmpSEW = new SelectedElementWrapper();
					tmpSEW.id = Int32.Parse(curSelectedElemParts[0]);
					tmpSEW.depth = Int32.Parse(curSelectedElemParts[1]);
					tmpSEW.layers = new SelectedElementLayerWrapper[curSelectedElemParts.Length - 3];
					for(int j = 2; j < curSelectedElemParts.Length - 1; j++) 
					{
						string[] curSelectedElemLayerParts = curSelectedElemParts[j].Split(new char[] {','});
						SelectedElementLayerWrapper tmpSELW = new SelectedElementLayerWrapper();
						tmpSELW.nodeName = curSelectedElemLayerParts[0];
						tmpSELW.searchDirection = curSelectedElemLayerParts[1][0];
						tmpSELW.numSamePrevSib = Int32.Parse(curSelectedElemLayerParts[2]);
						tmpSELW.numSameNextSib = Int32.Parse(curSelectedElemLayerParts[3]);
						tmpSEW.layers[j - 2] = tmpSELW;
					}
					this.selectedElements[i] = tmpSEW;
				}
			}
			public override string ToString() 
			{
				if(this.bodyOnlySelected)
					return "-";

				StringBuilder sb = new StringBuilder();
				for(int i = 0; i < selectedElements.Length; i++)
					sb.Append(selectedElements[i].ToString());
				return sb.ToString();
			}
		}

		/// <summary>
		/// Wrapper class for one selected element in a whole server-side selectedElementsPositionDescriptor.
		/// </summary>
		internal class SelectedElementWrapper 
		{	
			internal int id;
			internal int depth;
			internal SelectedElementLayerWrapper[] layers;

			public override string ToString() 
			{
				StringBuilder sb = new StringBuilder();
				sb.Append(id + ";" + depth + ";");
				for(int i = 0; i < layers.Length; i++)
					sb.Append(layers[i].ToString());
				sb.Append(">");
				return sb.ToString();
			}
		}

		/// <summary>
		/// Wrapper class for one selection-layer in a selected element as part of a whole
		/// server-side selectedElementsPositionDescriptor.
		/// </summary>
		internal class SelectedElementLayerWrapper 
		{
			internal string nodeName;
			internal char searchDirection;
			internal int numSamePrevSib;
			internal int numSameNextSib;

			public override string ToString() 
			{
				return nodeName + "," + searchDirection + "," + numSamePrevSib + "," + numSameNextSib + ";";
			}
		}
		
		/// <summary>
		/// This method creates a server-side SelectedElementWrapper for a Html-DOM-Node relative to the corresponding
		/// Html-DOM-Body-Node.
		/// </summary>
		/// <param name="resWrapper">Resulting wrapper</param>
		/// <param name="selNode">Html-DOM-Node for which the server-side SelectedElementWrapper should get created.</param>
		/// <param name="bodyNode">Html-DOM-Body-Node to which the wrapped SelectedElementPositionDescripter is relative to.</param>
		/// <param name="selId">Id for the selectedElement.</param>
		/// <param name="searchDir">SearchDirection for the selectedElement (left('l'),right('r') or both ('b') - default is 'b'</param>
		/// <returns>true - success, false - failed</returns>
		private bool createSelectedElementWrapperForNode(
			out SelectedElementWrapper resWrapper, 
			IHTMLDOMNode selNode, 
			IHTMLDOMNode bodyNode,
			int selId, 
			ServerSearchDirection searchDir) 
		{
			IHTMLDOMNode tmpNode = selNode;
			SelectedElementWrapper mySelElemWrapper = new SelectedElementWrapper();
			mySelElemWrapper.id = selId;
			ArrayList tmpList = new ArrayList();
			while(tmpNode != null && tmpNode != bodyNode) 
			{
				SelectedElementLayerWrapper curLayer = new SelectedElementLayerWrapper();
				curLayer.nodeName = tmpNode.nodeName.ToLower();
				if(searchDir == ServerSearchDirection.Left)
					curLayer.searchDirection = 'l';
				else if(searchDir == ServerSearchDirection.Right)
					curLayer.searchDirection = 'r';
				else
					curLayer.searchDirection = 'b';
				int numSamePrevSibs = 0;
				int numSameNextSibs = 0;
				IHTMLDOMNode tmpSibTraverseNode = tmpNode.previousSibling;
				while(tmpSibTraverseNode != null) 
				{
					if(tmpSibTraverseNode.nodeName.ToLower().Equals(curLayer.nodeName))
						numSamePrevSibs++;
					tmpSibTraverseNode = tmpSibTraverseNode.previousSibling;
				}
				tmpSibTraverseNode = tmpNode.nextSibling;
				while(tmpSibTraverseNode != null) 
				{
					if(tmpSibTraverseNode.nodeName.ToLower().Equals(curLayer.nodeName))
						numSameNextSibs++;
					tmpSibTraverseNode = tmpSibTraverseNode.nextSibling;
				}
				curLayer.numSamePrevSib = numSamePrevSibs;
				curLayer.numSameNextSib = numSameNextSibs;
				tmpList.Add(curLayer);

				tmpNode = tmpNode.parentNode;
			}
			mySelElemWrapper.depth = tmpList.Count;
			mySelElemWrapper.layers = new SelectedElementLayerWrapper[tmpList.Count];
			for(int i = tmpList.Count - 1; i >= 0; i--) 
				mySelElemWrapper.layers[tmpList.Count - i - 1] = (SelectedElementLayerWrapper)tmpList[i];
			resWrapper = mySelElemWrapper;
			return true;
		}

		/// <summary>
		/// Helper-method for CreateSelectedElementsPositionDescriptor which finds the next-layer-DOM-node
		/// by the numSamePrevSib and numSameNextSib parts of that layer (see clientSelectedElementsPositionDescriptor).
		/// </summary>
		/// <param name="fatherNode">Gets changed to the next-layer-node if successfully found.</param>
		/// <param name="nextLayer"></param>
		/// <returns></returns>
		private bool findClientSelectedElementPartAtNextLayerByNumSib(
			ref IHTMLDOMNode fatherNode, 
			ClientSelectedElementLayerWrapper nextLayer) 
		{
			IHTMLDOMNode myTmpNode = null;
			IHTMLDOMChildrenCollection myChildCollection = (IHTMLDOMChildrenCollection)fatherNode.childNodes;
			IEnumerator myChildEnumerator = myChildCollection.GetEnumerator();
			int k = 0;
			bool moveNextSuccessful = false;
			
			// check for correct count of numSamePrevSib
			while(moveNextSuccessful = myChildEnumerator.MoveNext()) 
			{
				if(k >= nextLayer.numSamePrevSib && ((IHTMLDOMNode)myChildEnumerator.Current).nodeName.ToLower().Equals(nextLayer.nodeName.ToLower()))
					break;
				if(((IHTMLDOMNode)myChildEnumerator.Current).nodeName.ToLower().Equals(nextLayer.nodeName.ToLower()))
					k++;
			}
			if(!moveNextSuccessful) // loop terminated because of MoveNext() returned false - not enough previous siblings found - no match
				return false;
			// check for correct count of numSameNextSib
			myTmpNode = (IHTMLDOMNode)myChildEnumerator.Current;
			// assertion: moveNextSuccessful = true
			k = 0;
			while(k < nextLayer.numSameNextSib) 
			{
				moveNextSuccessful = myChildEnumerator.MoveNext();
				if(!moveNextSuccessful) // not enough next siblings found - no match
					return false;
				if(((IHTMLDOMNode)myChildEnumerator.Current).nodeName.ToLower().Equals(nextLayer.nodeName.ToLower()))
					k++;
			}
			// assertion: moveNextSuccessful == true
			// node found - traverse to this in HTML-DOM-Tree and return match indicating true
			fatherNode = myTmpNode;
			return true;
		}

		
		/// <summary>
		/// Helper-method for CreateSelectedElementsPositionDescriptor which finds the next-layer-DOM-node
		/// by the sameElemCountRelToFather parts of that layer (see clientSelectedElementsPositionDescriptor).
		/// </summary>
		/// <param name="fatherNode">Gets changed to the next-layer-node if successfully found.</param>
		/// <param name="nextLayer"></param>
		/// <returns></returns>
		private bool findClientSelectedElementPartAtNextLayerBySameElemCountRelToFather(
			ref IHTMLDOMNode fatherNode, 
			ClientSelectedElementLayerWrapper nextLayer) 
		{
			// assign fatherNode to temporally variable - to not manipulate reference-parameter in case of fail
			IHTMLDOMNode tmpNode = fatherNode;
			IHTMLDOMNode retNode = null;
			int steps = nextLayer.sameElemCountRelToFather; // steps to go
			if(steps <= 0)
				return false;
			retNode = TraverseDOMForeward(ref tmpNode, ref steps, nextLayer.nodeName);
			if(steps == 0 && tmpNode != null) 
			{
				// assertion: it was possible to do the number of steps, and at least 1 step, because otherwise
				//            tmpNode would be null if not steps was initially 0
				fatherNode = tmpNode;
				return true;
			}
			return false;
		}
		
		/// <summary>
		/// Helper-method for CreateSelectedElementsPositionDescriptor which finds the next-layer-DOM-node
		/// by the sourceIndexRelToBody part of that layer (see clientSelectedElementsPositionDescriptor).
		/// </summary>
		/// <param name="fatherNode">Gets changed to the next-layer-node if successfully found.</param>
		/// <param name="bodyNode"></param>
		/// <param name="nextLayer"></param>
		/// <returns></returns>
		private bool findClientSelectedElementPartAtNextLayerBySourceIndex(
			ref IHTMLDOMNode fatherNode,
			IHTMLDOMNode bodyNode,
			ClientSelectedElementLayerWrapper nextLayer) 
		{
			if(nextLayer.sourceIndexRelToBody < 0) // no sourceIndexRelToBody set by client (only available in MS Internet Explorer)
				return false;
			IHTMLElementCollection allElementsCollection = (IHTMLElementCollection)((IHTMLElement)fatherNode).all;
			IEnumerator allElementsEnumerator = allElementsCollection.GetEnumerator();
			IHTMLElement curElement = null;
			int indexToSearchFor = curElement.sourceIndex;
			int bodyNodeIndex = ((IHTMLElement)bodyNode).sourceIndex;
			bool found = false;
			while(allElementsEnumerator.MoveNext()) 
			{
				curElement = (IHTMLElement)allElementsEnumerator.Current;
				if(curElement.sourceIndex - bodyNodeIndex == indexToSearchFor) 
				{	
					if(curElement.tagName.ToLower().Equals(nextLayer.nodeName)) // check if also nodename matches
						found = true;
					break;
				}
			}
			if(found) {
				fatherNode = (IHTMLDOMNode)curElement;
				return true;
			}
			return false;
		}
		
		/******************************************************************************************
		 *********************************** HTML-DOM-TREE-UTILS **********************************
		 ******************************************************************************************/
		
		/// <summary>
		/// Traverses a HTML-DOM-Tree in the forward direction (preorder).
		/// It does this in a recursive way by using the recursiveTraverseDOMForeward method.
		/// NOTE: Only HTML-DOM-Nodes of type "elementnode" with DOM-nodeType-number 1 are counted.
		/// </summary>
		/// <param name="initNode">
		/// Before the call it is the initial starting node for traversial.
		/// After the call it is the resulting node of the search, or if not so much steps were possible it
		/// is the last found element named "lowerElement", or null if not even one step was possible and steps
		/// was greater than 0.
		/// </param>
		/// <param name="steps">steps to traverse, gets decremented by each successful step</param>
		/// <param name="lowerElementName">lowercase element name of elements to count - if null, all elementtype nodes are counted</param>
		/// <returns>Last Node visited during traversal - if node found same as initNode, else last node in tree</returns>
		internal static IHTMLDOMNode TraverseDOMForeward(ref IHTMLDOMNode initNode, ref int steps, string lowerElementName) 
		{
			IHTMLDOMNode curNode = initNode;
			bool posSteps = steps > 0;
			recursiveTraverseDOMForeward(ref initNode, ref curNode, ref steps, lowerElementName);
			// not even one step was possible but steps was initialy > 0 -> set initNode to null.
			if(posSteps && lowerElementName != null && !initNode.nodeName.ToLower().Equals(lowerElementName))
				initNode = null;
			return curNode;
		}

		/// <summary>
		/// NOTE: This method only counts HTML-DOM-Nodes of type "elementnode" with DOM-nodeType-number 1.
		/// </summary>
		private static void recursiveTraverseDOMForeward(ref IHTMLDOMNode resNode, ref IHTMLDOMNode curNode, ref int stepsLeft, string lowerElementName) 
		{
			if(stepsLeft <= 0)
				return;
			for(IHTMLDOMNode tmpNode = curNode.firstChild; tmpNode != null; tmpNode = tmpNode.nextSibling)
			{
				if(stepsLeft > 0 && tmpNode.nodeType == 1) // only Elementnodes are recognized DOM-Nodetype 1
				{
					curNode = tmpNode;
					// only count elementnodes named like lowerElementName - or all elementnodes if lowerElementName is null
					if( lowerElementName == null || curNode.nodeName.ToLower().Equals(lowerElementName)) 
					{
						resNode = curNode;
						stepsLeft = stepsLeft - 1;
					}
					recursiveTraverseDOMForeward(ref resNode, ref curNode, ref stepsLeft, lowerElementName);
				}
			}
		}

		/// <summary>
		/// Get Tree-representation of HTML-DOM-Tree - for debugging reasons (logging).
		/// This method uses the private method recursivePreOrderHtmlElementToString() method to recursively
		/// do the job.
		/// </summary>
		/// <param name="rootNode">rootnode of the (sub)tree to represent</param>
		/// <returns></returns>
		internal string HtmlTreeToString(IHTMLDOMNode rootNode) 
		{
			StringBuilder sb = new StringBuilder();
			if(rootNode == null) 
			{
				sb.Append("\nnull");
			} 
			else 
			{
				sb.Append("\n");
				this.recursivePreOrderHtmlDomNodeToString(rootNode,0,sb);
			}
			return sb.ToString();
		} // end: method HtmlTreeToString

		private void recursivePreOrderHtmlDomNodeToString(IHTMLDOMNode myNode, int myDepth, StringBuilder sb) 
		{
			if(myNode == null)
				return;
			for(int i=0; i<myDepth; i++) 
			{
				if(i < (myDepth -1))
					sb.Append("      ");
				else
					sb.Append("   |_ ");
			}

			sb.Append("<" + myNode.nodeName + "," + myNode.nodeType + ">\n");
			IHTMLDOMChildrenCollection myChildNodes = (IHTMLDOMChildrenCollection)myNode.childNodes;
			if(myChildNodes == null || myChildNodes.length == 0)
				return;
			foreach(IHTMLDOMNode currChild in myChildNodes)
				recursivePreOrderHtmlDomNodeToString(currChild,myDepth + 1, sb);
		} // end: method recursivePreOrderHtmlElementToString
	}
}