using System;
using log4net;
using System.Text;

namespace WebseitenMonitor.HtmlUtilities 
{
	internal class HtmlTagAttributeToken 
	{
		internal const int none = 0, // non relevant or syntactical invalid one
			quotedAttribute = 1, // attrute-value is embraced by quotes
			nonQuotedAttribute = 2, // attribute-value isn't embraced by quotes
			eof = 3; // eof reached-token
		internal int kind;
		internal string name; // in lower case
		internal string val;
		internal int nameBeginPos;
		internal int nameOuterEndPos;
		internal int valBeginPos; // in case of quotedAttribute - pos of char after '"', else first non-Whitespace-char after '="
		internal int valOuterEndPos; // in case of quotedAttribute - pos of '"'-char, else last non-Whitespace-char in attr-val
	} // end: class HtmlTagAttributeToken

	internal class HtmlTagAttributeScanner 
	{
		private static readonly ILog log = LogManager.GetLogger(typeof(HtmlTagAttributeScanner));

		private const char eofCh = '\u0080';

		private string s;
		private char ch;
		private int pos;

		internal void Init(string s) 
		{
			this.s = s;
			this.pos = -1;
			this.nextCh();
		}

		private void nextCh() 
		{
			if(this.pos >= this.s.Length)
				this.ch = HtmlTagAttributeScanner.eofCh;
			else 
			{
				this.pos++;
				if(this.pos >= this.s.Length)
					this.ch = HtmlTagAttributeScanner.eofCh;
				else
					this.ch = this.s[this.pos];
			}
		}

		internal HtmlTagAttributeToken ScanForAttribute(string name) 
		{
			HtmlTagAttributeToken tmpToken = this.NextAttribute();
			while(tmpToken.kind != HtmlTagAttributeToken.eof) 
			{
				if(tmpToken.name.Equals(name))
					return tmpToken;
			}
			return null;
		}

		internal HtmlTagAttributeToken NextAttribute() 
		{
			HtmlTagAttributeToken t = null;
			do 
			{
			while(!Char.IsLetter(this.ch) && this.ch != HtmlTagAttributeScanner.eofCh)
				this.nextCh();
				// assertion: this.ch is letter or eofCh
				t = new HtmlTagAttributeToken();
				t.nameBeginPos = this.pos;
				switch(this.ch) 
				{
					case 'a': case 'b': case 'c': case 'd': case 'e': case 'f': case 'g': case 'h': case 'i':
					case 'j': case 'k': case 'l': case 'm': case 'n': case 'o': case 'p': case 'q': case 'r':
					case 's': case 't': case 'u': case 'v': case 'w': case 'x': case 'y': case 'z':
					case 'A': case 'B': case 'C': case 'D': case 'E': case 'F': case 'G': case 'H': case 'I':
					case 'J': case 'K': case 'L': case 'M': case 'N': case 'O': case 'P': case 'Q': case 'R':
					case 'S': case 'T': case 'U': case 'V': case 'W': case 'X': case 'Y': case 'Z':
						this.readTagAttribute(t);
						break;
					case HtmlTagAttributeScanner.eofCh:
						t.kind = HtmlTagAttributeToken.eof;
						break;
					default:
						t.kind = HtmlTagAttributeToken.none;
						break;
				}

			} while(t.kind == HtmlTagAttributeToken.none);
			return t;
		} // end: method NextAttribute

		internal void readTagAttribute(HtmlTagAttributeToken t) 
		{
			StringBuilder sb = new StringBuilder();
			sb.Append(this.ch);
			this.nextCh();
			while(Char.IsLetter(this.ch))
			{
				sb.Append(this.ch);
				this.nextCh();
			}
			// After attribute-name there must follow whitespace or the equality-sign
			if(this.ch != '=' && !Char.IsWhiteSpace(this.ch)) 
			{
				t.kind = HtmlTagAttributeToken.none;
				return;
			}
			t.name = sb.ToString().ToLower();
			t.nameOuterEndPos = this.pos;
			// overread whitespace between attribute-name and =
			while(Char.IsWhiteSpace(this.ch))
				this.nextCh();
			// after overridden whitespace between attribute name and = there must occure the =
			if(this.ch != '=') 
			{
				t.kind = HtmlTagAttributeToken.none;
				return;
			}
			// assertion: this.ch == '='
			this.nextCh();
			// overread whitespace between = and attribute-value
			while(Char.IsWhiteSpace(this.ch))
				this.nextCh();
			// there must follow an attribute-value
			if(this.ch == HtmlTagAttributeScanner.eofCh) 
			{
				t.kind = HtmlTagAttributeToken.none;
				return;
			}
			// assertion: this.ch is attribute-value-begin-guote or first sign of non-quoted attribute-value
			if(this.ch == '"') 
			{
				t.kind = HtmlTagAttributeToken.quotedAttribute;
				this.nextCh();
			}
			else
				t.kind = HtmlTagAttributeToken.nonQuotedAttribute;
			t.valBeginPos = this.pos;
			sb = new StringBuilder();
			while(this.ch != HtmlTagAttributeScanner.eofCh && this.ch != '>') 
			{
				if(t.kind == HtmlTagAttributeToken.quotedAttribute) // quoted attribute-value
				{
					if(this.ch == '"') 
					{
						t.valOuterEndPos = this.pos;
						break;
					}
					sb.Append(this.ch);
				}
				else // non-quoted attribute-value
				{
					if(Char.IsWhiteSpace(this.ch) ) 
					{
						t.valOuterEndPos = this.pos;
						break;
					}
					sb.Append(this.ch);
				}
				this.nextCh();
			}
			t.val = sb.ToString();
			if(this.ch == HtmlTagAttributeScanner.eofCh) 
			{
				t.valOuterEndPos = this.pos;
				if(t.kind == HtmlTagAttributeToken.quotedAttribute)
					t.kind = HtmlTagAttributeToken.none;
			} 
			else if(this.ch == '>') 
			{
				t.valOuterEndPos = this.pos;
				if(t.kind == HtmlTagAttributeToken.quotedAttribute)
					t.kind = HtmlTagAttributeToken.none;
			}
		} // end: method readTagAttribute
	} // end: class HtmlTagAttributeScanner
	
	internal class HtmlTagToken 
	{	
		internal const int none		= 0, // non relevant
			openingTag = 1, // Html opening tag
			closingTag = 2, // Html closing tag
			bothTag = 3, // Html opening-closing tag e.g.: <BR/>
			eof		= 4; // eof reached-token

		internal int kind;
		internal string name;	// name of Html-Tag in lower-case!!!
		internal int beginPos;
		internal int endPos;
	} // end: class HtmlTagToken

	internal class HtmlTagScanner 
	{	
		private static readonly ILog log = LogManager.GetLogger(typeof(HtmlTagScanner));

		private const char eofCh = '\u0080';

		private StringBuilder sb; // if initialized with stringBuilder
		private string s; // if initialized with string
		private bool baseIsFixedLength; // true - if initialized with string, false if initialized with StringBuilder
		private char ch;
		private int pos;

		internal void Init(StringBuilder sb) 
		{
			this.baseIsFixedLength = false;
			this.sb = sb;
			this.pos = -1;
			this.nextCh();
		}

		internal void Init(string s) 
		{
			this.baseIsFixedLength = true;
			this.s = s;
			this.pos = -1;
			this.nextCh();
		}

		private void nextCh() 
		{	
			if(this.baseIsFixedLength) // if base is string
			{
				if(this.pos >= this.s.Length)
					this.ch = HtmlTagScanner.eofCh;
				else 
				{
					this.pos++;
					if(this.pos >= this.s.Length)
						this.ch = HtmlTagScanner.eofCh;
					else
						this.ch = this.s[this.pos];
				}
			} 
			else // if base is StringBuilder
			{
				if(this.pos >= this.sb.Length)
					this.ch = HtmlTagScanner.eofCh;
				else 
				{
					this.pos++;
					if(this.pos >= this.sb.Length)
						this.ch = HtmlTagScanner.eofCh;
					else
						this.ch = this.sb[this.pos];
				}
			}
		}

		internal void jumpToPos(int pos) 
		{
			if(this.baseIsFixedLength) // if base is string
			{
				this.pos = pos;
				if(this.pos >= this.s.Length)
					this.ch = HtmlTagScanner.eofCh;
				else
					this.ch = this.s[this.pos];
			} 
			else // if base is StringBuilder
			{
				this.pos = pos;
				if(this.pos >= this.sb.Length)
					this.ch = HtmlTagScanner.eofCh;
				else
					this.ch = this.sb[this.pos];
			}
		}

		internal void jumpFwd(int step) 
		{
			if(this.baseIsFixedLength) // if base is string
			{
				this.pos = this.pos + step;
				if(this.pos >= this.s.Length)
					this.ch = HtmlTagScanner.eofCh;
				else
					this.ch = this.s[this.pos];
			} 
			else // if base is StringBuilder
			{
				this.pos = this.pos + step;
				if(this.pos >= this.sb.Length)
					this.ch = HtmlTagScanner.eofCh;
				else
					this.ch = this.sb[this.pos];
			}
		}

		internal void jumpRwd(int step) 
		{
			if(this.baseIsFixedLength) // if base is string
			{
				this.pos = this.pos - step;
				if(this.pos >= this.s.Length)
					this.ch = HtmlTagScanner.eofCh;
				else
					this.ch = this.s[this.pos];
			} 
			else // if base is StringBuilder
			{
				this.pos = this.pos - step;
				if(this.pos >= this.sb.Length)
					this.ch = HtmlTagScanner.eofCh;
				else
					this.ch = this.sb[this.pos];
			}
		}

		internal void refreshChar() 
		{
			if(this.baseIsFixedLength) // if base is string
			{
				if(this.pos >= this.s.Length)
					this.ch = HtmlTagScanner.eofCh;
				else
					this.ch = this.s[this.pos];
			} 
			else // if base is StringBuilder
			{
				if(this.pos >= this.sb.Length)
					this.ch = HtmlTagScanner.eofCh;
				else 
					this.ch = this.sb[this.pos];
			}
		}

		internal HtmlTagToken NextTag() 
		{	
			HtmlTagToken t = null;
			do 
			{
				while(this.ch != '<' && this.ch != HtmlTagScanner.eofCh)
					this.nextCh();
				t = new HtmlTagToken();
				t.beginPos = this.pos;
				this.nextCh();
				switch(this.ch) 
				{	
					case '/': 
						t.kind = HtmlTagToken.closingTag;
						this.readClosingTagName(t);
						break;
					case 'a': case 'b': case 'c': case 'd': case 'e': case 'f': case 'g': case 'h': case 'i':
					case 'j': case 'k': case 'l': case 'm': case 'n': case 'o': case 'p': case 'q': case 'r':
					case 's': case 't': case 'u': case 'v': case 'w': case 'x': case 'y': case 'z':
					case 'A': case 'B': case 'C': case 'D': case 'E': case 'F': case 'G': case 'H': case 'I':
					case 'J': case 'K': case 'L': case 'M': case 'N': case 'O': case 'P': case 'Q': case 'R':
					case 'S': case 'T': case 'U': case 'V': case 'W': case 'X': case 'Y': case 'Z':
						t.kind = HtmlTagToken.openingTag;
						this.readOpeningTagName(t);
						break;
					case '!':
						t.kind = HtmlTagToken.none;
						this.readComment(t);
						break;
					case HtmlTagScanner.eofCh:
						t.kind = HtmlTagToken.eof;
						break;
					default:
						t.kind = HtmlTagToken.none;
						break;
				}
			} while(t.kind == HtmlTagToken.none);
			return t;
		} // end method NextTag
			
		/// <summary>
		/// This method overreads Html-comments. Is must be called when this.ch stands at the '!' character
		/// (<!-- ... -->). If after the '!' there follow two '-' then everything until after a "-->" sequence
		/// is overridden unless eof is reached before. If after the '!' there don't follow two '-' the
		/// method returns wih this.ch at the first non-'-' sign after the '!'.
		/// </summary>
		/// <param name="t"></param>
		internal void readComment(HtmlTagToken t) 
		{
			char firstHyphen = '\0';
			char secondHyphen = '\0';
			this.nextCh();
			if(this.ch != '-')
				return;
			firstHyphen = this.ch;
			this.nextCh();
			if(this.ch != '-')
				return;
			secondHyphen = this.ch;
			this.nextCh();
			if(firstHyphen == '-' && secondHyphen == '-') 
			{	
				char tmpChar = '\0';
				bool doubleHyphenFound = false;
				while(this.ch != HtmlTagScanner.eofCh) 
				{	
					if(doubleHyphenFound)
						if(this.ch == '>')
							break;
						else
							doubleHyphenFound = false;
					if(tmpChar == '-')
						if(this.ch == '-')
							doubleHyphenFound = true;
					tmpChar = this.ch;
					this.nextCh();
				}
				if(this.ch != HtmlTagScanner.eofCh)
					this.nextCh();
			}
		} // end method readComment

		internal void readClosingTagName(HtmlTagToken t) 
		{
			StringBuilder sb = new StringBuilder();
			this.nextCh();
			// No leading space allowed e.g.: </h1> ok. but </ h1> or < /h1> not a closing tag
			while(Char.IsLetterOrDigit(this.ch)) 
			{
				sb.Append(this.ch);
				this.nextCh();
			}
			t.name = sb.ToString().ToLower();
			// overread tailing whitespace
			while(Char.IsWhiteSpace(this.ch))
				this.nextCh();
			if(this.ch != '>') 
			{
				t.kind = HtmlTagToken.none;
			}
			else 
			{	
				t.endPos = this.pos;
				this.nextCh();
			}

		} // end method readClosingTagName

		internal void readOpeningTagName(HtmlTagToken t) 
		{
			StringBuilder sb = new StringBuilder();
			sb.Append(this.ch);
			this.nextCh();
			while(Char.IsLetterOrDigit(this.ch)) 
			{
				sb.Append(this.ch);
				this.nextCh();
			}
			t.name = sb.ToString().ToLower();
			char lastCh = sb[sb.Length -1];
			while(this.ch != '>' && this.ch != '<' && this.ch != HtmlTagScanner.eofCh) 
			{
				if(this.ch != ' ' && this.ch != '\t')
					lastCh = this.ch;
				this.nextCh();
			}
			if(this.ch == '>') 
			{
				if(lastCh == '/') 
				{
					t.kind = HtmlTagToken.bothTag;
					t.endPos = this.pos;
					this.nextCh();
					return;
				} 
				else 
				{
					t.kind = HtmlTagToken.openingTag;
					t.endPos = this.pos;
					this.nextCh();
					return;
				}
			} 
			else 
			{
				t.kind = HtmlTagToken.none;
				return;
			}
		} // end: method readOpeningTagName
	} // end: class HtmlTagScanner
}