using System;
using System.Text;
using log4net;
using System.Net;
using System.IO;

namespace WebseitenMonitor.HttpUtilities 
{

	public class HttpHtmlRetriever 
	{
		private static readonly ILog log = LogManager.GetLogger(typeof(HttpHtmlRetriever));

		// Request specific
		private Encoding sourceEncoding;
		private string[] acceptTypes = new string[] {"text/html","application/xhtml+xml"};
		
		internal enum RetrieveResult {Ok, InvalidUrl, NotAHttpUrl, InvalidResponseMimeType, InvalidResponseStatusCode, NoSourceStreamObtainable, ExceptionOccured, Error}

		private string createRequestAcceptValue() 
		{
			// TODO: syntax for multible accept-types
			return null;
			
			/*
			if(this.acceptTypes == null)
				return null;
			StringBuilder sb = new StringBuilder();
			for(int i = 0; i < this.acceptTypes.Length; i++) 
			{
				sb.Append(this.acceptTypes[i]);
				if(i < this.acceptTypes.Length - 1)
					sb.Append(" , ");
			}
			return sb.ToString();
			*/
		}

		private bool checkResponseMimeType(string mimeType) 
		{
			mimeType = mimeType.Trim().ToLower();
			foreach(string s in this.acceptTypes)
				if(mimeType.StartsWith(s))
					return true;
			return false;
		}

		public bool SetSourceEncoding(Encoding sourceEncoding, out string errorMessage) 
		{
			if(sourceEncoding == null) 
			{
				log.Error("Attempt to explicitly set source encoding to null - not allowed.");
				errorMessage = "SourceEncoding darf nicht explizit auf null gesetzt werden.";
				return false;
			}
			this.sourceEncoding = sourceEncoding;
			errorMessage = "";
			return true;
		}

		public bool SetSourceEncoding(string sourceEncodingName, out string errorMessage) 
		{
			if(sourceEncodingName == null) 
			{
				log.Error("Invalid parameter - sourceEncodingName is null.");
				errorMessage = "Ungltiger Parameter - SourceEncoding-Name ist null.";
				return false;
			}
			try 
			{
				this.sourceEncoding = Encoding.GetEncoding(sourceEncodingName);
			} 
			catch(Exception e) 
			{	
				log.Error("Exception occured during try to set source encoding by name.",e);
				errorMessage = "Ungltiger Parameter - Entsprechendes Encoding nicht untersttzt.";
				return false;
			}
			errorMessage = "";
			return true;
		}
		
		public bool SetSourceEncoding(int sourceEncodingCodePage, out string errorMessage) 
		{
			try 
			{
				log.Error("Invalid parameter - sourceEncodingCodePage is null.");
				this.sourceEncoding = Encoding.GetEncoding(sourceEncodingCodePage);
			} 
			catch(Exception e) 
			{
				log.Error("Exception occured during try to set source encoding by codepage.",e);
				errorMessage = "Ungltiger Parameter - Entsprechendes Encoding nicht untersttzt.";
				return false;
			}
			errorMessage = "";
			return true;
		}
		
		// only wrappes the other overloaded Retrieve method.
		internal RetrieveResult Retrieve(string sourceUrl, out string sourceString, out string errorMsg) 
		{
			TextReader r;
			RetrieveResult rRes = this.Retrieve(sourceUrl, out r, out errorMsg);
			if(r != null) 
			{
				try 
				{
					sourceString = r.ReadToEnd();
				} 
				catch 
				{
					errorMsg = "Response-Text konnte nicht erfolgreich ausgelesen werden.";
					sourceString = null;
					return RetrieveResult.ExceptionOccured;
				}
			} else
				sourceString = null;
			return rRes;

		}
		
		/// <summary>
		/// Note: The TextReader obtained by this method uses the Encoding specified by this.sourceEncoding.
		///       If this.sourceEncoding is null and WebResponse.Characterset holds a characterset, this would
		///       be used. Otherwise it is tried to use "iso-8859-1" and if this also failed the platform default
		///       characterset is used.
		///       
		///       The generated WebRequest gets a Accept-Field for the mime-types in this.acceptTypes (text/html , application/xhtml+xml).
		///       In the resulting WebResponse there is also a check for this mime-Types done, and the statusCode must be OK (200).
		/// </summary>
		/// <param name="sourceUrl"></param>
		/// <param name="sourceReader"></param>
		/// <param name="errorMsg"></param>
		/// <returns></returns>
		internal RetrieveResult Retrieve(string sourceUrl, out TextReader sourceReader, out string errorMsg) 
		{
			if(sourceUrl == null) 
			{
				sourceReader = null;
				errorMsg = "Es wurde keine Quell-Url bergeben.";
				return RetrieveResult.InvalidUrl;
			}
			Uri sourceUri = null;
			try 
			{
				sourceUri = new Uri(sourceUrl);
			}	
			catch(UriFormatException) {}
			if(sourceUri == null) 
			{
				sourceReader = null;
				errorMsg = "Es wurde eine syntaktisch inkorrekte Quell-Url bergeben.";
				return RetrieveResult.InvalidUrl;
			}
			Stream s = null;
			TextReader r = null;
			
			WebRequest headReq = null;
			HttpWebRequest httpHeadReq = null;
			WebRequest req = null;
			HttpWebRequest httpReq = null;
			HttpWebResponse httpRes = null;
			try 
			{
				headReq = WebRequest.Create(sourceUri);
				if(headReq == null) 
				{
					sourceReader = null;
					errorMsg = "Es konnte kein Web-Request erzeugt werden.";
					return RetrieveResult.Error;
				}
				if((httpHeadReq = (headReq as HttpWebRequest)) == null) 
				{
					sourceReader = null;
					errorMsg = "Bei der bergebenen Url handelt es sich um keine Http-entsprechende.";
					return RetrieveResult.NotAHttpUrl;
				}
				// set headers
				httpHeadReq.KeepAlive = true;
				httpHeadReq.Method = "HEAD";
				HttpWebResponse httpHeadRes = (HttpWebResponse)httpHeadReq.GetResponse();
				if(httpHeadRes == null) 
				{
					sourceReader = null;
					errorMsg = "Es konnte kein Web-Response bezogen werden.";
					return RetrieveResult.Error;
				}
				try 
				{
					httpHeadRes.Close();
				} 
				catch {}
				if(httpHeadRes.StatusCode == HttpStatusCode.OK 
					|| httpHeadRes.StatusCode == HttpStatusCode.NoContent
					|| httpHeadRes.StatusCode == HttpStatusCode.MethodNotAllowed) 
				{
					if(httpHeadRes.StatusCode == HttpStatusCode.MethodNotAllowed) 
					{
						log.Info("HTTP-HEAD Request answered with MethodNotAllowed-Response-Statuscode (Url: " + sourceUri.ToString() + ")");
					} 
					else 
					{
						// Check Mime-Type if HEAD Request was successfull (got answered)
						if(!checkResponseMimeType(httpHeadRes.ContentType)) 
						{
							sourceReader = null;
							errorMsg = "Der bezogene Web-Response hat einen ungltigen Mime-Type.";
							return RetrieveResult.InvalidResponseMimeType;
						}
					}
					// assertion: Head request showed valid Mime-Type
					// Do real request
					req = WebRequest.Create(sourceUri);
					if(req == null) 
					{
						sourceReader = null;
						errorMsg = "Es konnte kein Web-Request erzeugt werden.";
						return RetrieveResult.Error;
					}
					if((httpReq = (req as HttpWebRequest)) == null) 
					{
						sourceReader = null;
						errorMsg = "Bei der bergebenen Url handelt es sich um keine Http-entsprechende.";
						return RetrieveResult.NotAHttpUrl;
					}
					// set headers
					httpReq.KeepAlive = false;
					httpReq.Method = "GET";
					httpRes = (HttpWebResponse)httpReq.GetResponse();
					if(httpRes == null) 
					{
						sourceReader = null;
						errorMsg = "Es konnte kein Web-Response bezogen werden.";
						return RetrieveResult.Error;
					}
					// Check StatusCode and MimeType of response
					if(httpRes.StatusCode != HttpStatusCode.OK) 
					{
						try 
						{
							httpRes.Close();
						} 
						catch {}
						sourceReader = null;
						errorMsg = "Der bezogene Web-Response hat einen ungltigen Statuscode.";
						return RetrieveResult.InvalidResponseStatusCode;
					}
					if(!checkResponseMimeType(httpRes.ContentType)) 
					{
						try 
						{
							httpRes.Close();
						} 
						catch {}
						sourceReader = null;
						errorMsg = "Der bezogene Web-Response hat einen ungltigen Mime-Type.";
						return RetrieveResult.InvalidResponseMimeType;
					}
				} // end: if(httpHeadRes.StatusCode == HttpStatusCode.OK || httpHeadRes.StatusCode == HttpStatusCode.NoContent)
				else 
				{
					sourceReader = null;
					errorMsg = "";
					return RetrieveResult.InvalidResponseStatusCode;
				}

				// Read Response Content
				s = httpRes.GetResponseStream();
				if(this.sourceEncoding == null) 
				{	
					Encoding myEncoding = null;
					try { myEncoding = Encoding.GetEncoding(httpRes.CharacterSet); } 
					catch {}
					if(myEncoding == null) 
					{
						try { myEncoding = Encoding.GetEncoding("iso-8859-1"); } 
						catch {}
						if(myEncoding == null)
							r = new StreamReader(s);
						else
							r = new StreamReader(s, myEncoding);
					}
					else
						r = new StreamReader(s,myEncoding);
				}
				else 
				{
					r = new StreamReader(s,sourceEncoding);
				}
			} 
			catch(Exception e) 
			{
				log.Warn("Exception occured while obtaining DataSource stream", e);
				sourceReader = null;
				errorMsg = "Zugriff auf Datenquelle fehlgeschlagen. Exception aufgetreten.";
				try 
				{
					if(r != null)
						r.Close();
					else if(s != null)
						s.Close();
					else if(httpRes != null)
						httpRes.Close();
				} 
				catch(Exception ee) {log.Warn("Exception occured while trying to close DataSource stream in response to an exception", ee);}
				return RetrieveResult.ExceptionOccured;
			}
			if(r != null) 
			{
				sourceReader = r;
				errorMsg = "";
				return RetrieveResult.Ok;
			} 
			else 
			{
				sourceReader = null;
				errorMsg = "Zugriff auf Datenquelle ist fehlgeschlagen.";
				return RetrieveResult.NoSourceStreamObtainable;
			}
		}
	}
}