namespace WebseitenMonitor.Tests 
{
	using System;
	using NUnit.Framework;
	using WebseitenMonitor.UriUtilities;
	
	[TestFixture]
	public class GetAbsoluteUriCorrToContextTests 
	{
		[Test] public void SelfTest1() 
		{
			Assert.AreEqual(
				(new UriBuilder("http://www.google.at/index.html#fragment")).Path,
				"/index.html");
			Assert.AreEqual(
				(new UriBuilder("http://www.google.at")).Path,
				"/");
			Assert.AreEqual(
				(new UriBuilder("http://www.gmx.at/dir/dir2?name=value")).Path,
				"/dir/dir2");
			Assert.AreEqual(
				(new UriBuilder("http://www.gmx.at/dir/file.html")).Path,
				"/dir/file.html");
		}

		[Test] public void AlreadyAbsoluteUriTest() 
		{
			Assert.AreEqual(
				new Uri("http://www.google.at/test/hallo.html"),
				UriConvertion.GetAbsoluteUriCorrToContext(null,"http://www.google.at/test/hallo.html"));
		}

		[Test] public void NoAbsoluteUriAndBaseNullTest() 
		{
			Assert.IsNull(UriConvertion.GetAbsoluteUriCorrToContext(
				null,"/path/index.html"));
		}

		[Test] public void AlreadyAbsoluteUriAndBaseIrrelevantTest() 
		{
			Assert.AreEqual(
				new Uri("http://www.google.at/test/hallo.html"),
				UriConvertion.GetAbsoluteUriCorrToContext(new Uri("http://www.gmx.at"),"http://www.google.at/test/hallo.html"));
		}

		[Test] public void AlreadyAbsoluteUriHasFragmentIdentifierTest() 
		{
			// Note: Uri's equals method ignores fragments - so compare the uri's string representation
			Assert.AreEqual(
				(new Uri("http://www.google.at/test/hallo.html#fragment1")).ToString(),
				UriConvertion.GetAbsoluteUriCorrToContext(null,"http://www.google.at/test/hallo.html#fragment1").ToString());
		}

		[Test] public void SimpleSameDirectoryTest() 
		{
			Assert.AreEqual(
				(new Uri("http://www.gmx.at/dir/res.html")).ToString(),
				UriConvertion.GetAbsoluteUriCorrToContext(new Uri("http://www.gmx.at/dir/original.html"),"res.html").ToString());
			Assert.AreEqual(
				(new Uri("http://www.gmx.at/dir/res.html")).ToString(),
				UriConvertion.GetAbsoluteUriCorrToContext(new Uri("http://www.gmx.at/dir/original.html"),"./res.html").ToString());
		}

		[Test] public void SimpleAbsoluteDirectoryTest() 
		{
			Assert.AreEqual(
				(new Uri("http://www.gmx.at/dir1/dir2/index.html")).ToString(),
				UriConvertion.GetAbsoluteUriCorrToContext(new Uri("http://www.gmx.at/d/hallo.html"),"/dir1/dir2/index.html").ToString());
		}

		[Test] public void SimpleParentDirectoryTest() 
		{
			Assert.AreEqual(
				(new Uri("http://www.gmx.at/dir1/index.html")).ToString(),
				UriConvertion.GetAbsoluteUriCorrToContext(new Uri("http://www.gmx.at/dir1/dir2/hallo.html"),"../index.html").ToString());
		}
		
		[Test] public void SimpleSubDirectoryTest() 
		{
			Assert.AreEqual(
				(new Uri("http://www.gmx.at/dir1/dir2/index.html")).ToString(),
				UriConvertion.GetAbsoluteUriCorrToContext(new Uri("http://www.gmx.at/dir1/hallo.html"),"dir2/index.html").ToString());
			Assert.AreEqual(
				(new Uri("http://www.gmx.at/dir1/dir2/index.html")).ToString(),
				UriConvertion.GetAbsoluteUriCorrToContext(new Uri("http://www.gmx.at/dir1/hallo.html"),"./dir2/index.html").ToString());
		}

		[Test] public void SelfAndParentTest() 
		{
			Assert.AreEqual(
				(new Uri("http://www.gmx.at/dir1/index.html")).ToString(),
				UriConvertion.GetAbsoluteUriCorrToContext(new Uri("http://www.gmx.at/dir1/dir2/dir3/dir4/hallo.html"),"../.././dir3/.././../index.html").ToString());
		}

		[Test] public void NonFileExtentionSegmentTest() 
		{
			Assert.AreEqual(
				(new Uri("http://www.google.at/search?name=value")).ToString(),
				UriConvertion.GetAbsoluteUriCorrToContext(new Uri("http://www.google.at/suche?n=v"),"search?name=value").ToString());
			Assert.AreEqual(
				(new Uri("http://www.google.at/search?name=value")).ToString(),
				UriConvertion.GetAbsoluteUriCorrToContext(new Uri("http://www.google.at/suche?n=v"),"./search?name=value").ToString());
		}

		[Test] public void VoidSegmentTest() 
		{
			Assert.AreEqual(
				(new Uri("http://www.google.at/dir/hallo.html")).ToString(),
				UriConvertion.GetAbsoluteUriCorrToContext(new Uri("http://www.google.at/dir/"),"hallo.html").ToString());
			Assert.AreEqual(
				(new Uri("http://www.google.at/dir/hallo.html")).ToString(),
				UriConvertion.GetAbsoluteUriCorrToContext(new Uri("http://www.google.at/dir/"),"./hallo.html").ToString());
		}
		
	}

	[TestFixture]
	public class GetUrisBasePathUriStringTests 
	{
		[Test] public void ParamTest() 
		{
			Assert.AreEqual("http://localhost/base/",
				UriConvertion.getUrisBasePathUriString(
				new Uri("http://localhost/base/index.htm?name=val:name1=val1")));
		}
		
		[Test] public void FragmentTest() 
		{
			Assert.AreEqual("http://localhost/base/",
				UriConvertion.getUrisBasePathUriString(
				new Uri("http://localhost/base/index.htm#fragment")));
		}

		[Test] public void StandardTest() 
		{
			Assert.AreEqual("http://localhost/base/",
				UriConvertion.getUrisBasePathUriString(
				new Uri("http://localhost/base/index.htm")));
		}

		[Test] public void NoExtensionButFragmentTest() 
		{
			Assert.AreEqual("http://localhost/base/",
				UriConvertion.getUrisBasePathUriString(
				new Uri("http://localhost/base/index#fragment")));
		}
		
		[Test] public void AlreadyBaseTest() 
		{
			Assert.AreEqual("http://localhost/base/",
				UriConvertion.getUrisBasePathUriString(
				new Uri("http://localhost/base/")));
		}

		[Test] public void NoFileExtensionTest() 
		{
			Assert.AreEqual("http://localhost/",
				UriConvertion.getUrisBasePathUriString(
				new Uri("http://localhost/base")));
		}

		[Test] public void RootPathOnlyTest() 
		{
			Assert.AreEqual("http://www.google.at/",
				UriConvertion.getUrisBasePathUriString(
				new Uri("http://www.google.at/")));
		}

		[Test] public void RootPathOnlyPutNotExplicitTest() 
		{
			Assert.AreEqual("http://www.google.at/",
				UriConvertion.getUrisBasePathUriString(
				new Uri("http://www.google.at")));
		}
	} // end: class GetUrisBasePathUriStringTests
}