using System;
using System.Data;
using System.Data.SqlClient;
using System.Collections;
using System.Configuration;
using log4net;
using System.Text;

namespace WebseitenMonitor.Ui 
{
	
	internal class CategoryNavigationForest 
	{
		private static readonly ILog log = LogManager.GetLogger(typeof(CategoryNavigationForest));
		private static readonly string DbTableOwnerPrefix = ((string)ConfigurationSettings.AppSettings["DbTableOwnerPrefix"]).Trim();
		
		// CategoryNavigationForest-state specific stuff
		private int userId; // UserId of user to which the forest corresponds
		private ArrayList roots; // root-CategoryNodes of forest
		private int selectedCategoryId = -2; // -2 .. none selected, -1 .. root selected, n .. Category with Id n selected
		
		// Html-representation specific stuff
		private string selectedFolderImg;
		private string unselectedFolderImg;
		private string plusImg;
		private string minusImg;
		private string neutralImg;
		private string navigBaseUrl;
		private string contentBaseUrl;
		
		// types for individual CategoryNodes
		internal enum CategoryNodeState {Open, Close};
		internal enum CategoryNodeKind {Leaf, NonLeaf};
		
		internal class CategoryNode 
		{
			internal CategoryNode(int Id, string CategoryName, CategoryNode FatherNode) 
			{
				this.State = CategoryNodeState.Close;
				this.Kind = CategoryNodeKind.Leaf;
				this.Id = Id;
				this.CategoryName = CategoryName;
				this.ChildNodes = new ArrayList();
				this.FatherNode = FatherNode;
			}

			internal CategoryNodeState State;
			internal CategoryNodeKind Kind;
			internal int Id; // id from DB (Categories table)
			internal string CategoryName; // from DB
			internal ArrayList ChildNodes; // null not allowed - empty or it holds subcategories
			internal CategoryNode FatherNode;
		}

		/// <summary>
		/// Creates new CategoryNavigationForest for a user identified by it's Id.
		/// Note that the constructor doesn't already load the data from the database, you
		/// have to call LoadWholeForest for this purpose.
		/// </summary>
		/// <param name="userId">
		/// UserId of user for which the tree should correspond to. This id is needed for
		/// a proper database-access.
		/// </param>
		internal CategoryNavigationForest(int userId) 
		{
			this.userId = userId;
		}
		
		internal int SelectedCategoryId 
		{
			set { this.selectedCategoryId = value; }
			get { return this.selectedCategoryId; }
		}
		
		internal CategoryNavigationForest.CategoryNode SelectedCategory 
		{
			get { return this.findCategoryNode(this.selectedCategoryId); }
		}

		internal ArrayList RootNodes 
		{
			get { return (this.roots==null)?new ArrayList():this.roots; }
		}

		/// <summary>
		/// This methods opens a Category by changing it's state to Open.
		/// If no Category with the passed id exists - nothing except of an error-log happens.
		/// </summary>
		/// <param name="Id">Id of Category to open.</param>
		internal void OpenCategory(int Id) 
		{
			if(Id == -1) // root Node is always open
				return;
			CategoryNode tmpNode = this.findCategoryNode(Id);
			if(tmpNode == null) 
			{
				log.Error("Tried to open non existing Category Node! Application inconsistence or http-request-fake.");
			} 
			else 
			{
				tmpNode.State = CategoryNavigationForest.CategoryNodeState.Open;
			}
		}
		
		/// <summary>
		/// This methods closes a Category by changing it's state to Close.
		/// If no Category with the passed id exists - nothing except of an error-log happens.
		/// </summary>
		/// <param name="Id">Id of Category to close.</param>
		internal void CloseCategory(int Id) 
		{
			CategoryNode tmpNode = this.findCategoryNode(Id);
			if(tmpNode == null) 
			{
				log.Fatal("Tried to close non existing Category Node! Application inconsistence or http-request-fake");
			} 
			else 
			{
				tmpNode.State = CategoryNavigationForest.CategoryNodeState.Close;
			}
		}
		
		// find Category by forrest-traversion
		private CategoryNavigationForest.CategoryNode findCategoryNode(int Id) 
		{
			CategoryNode resultNode;
			foreach(CategoryNode tmpNode in roots) 
			{
				resultNode = this.recursiveFindCategoryNode(Id, tmpNode);
				if(resultNode != null)
					return resultNode;
			}
			return null;
		}
		
		private CategoryNavigationForest.CategoryNode recursiveFindCategoryNode(int Id, CategoryNode Node) 
		{
			CategoryNode resultNode;
			if(Node.Id == Id)
				return Node;
			foreach(CategoryNode tmpNode in Node.ChildNodes) 
			{
				resultNode = this.recursiveFindCategoryNode(Id, tmpNode);
				if(resultNode != null)
					return resultNode;
			}
			return null;
		}

		private void removeNodeFromForest(CategoryNode node) 
		{
			if(node == null || node.Id < 0)
				return;
			if(node.FatherNode == null) 
			{
				if(this.selectedCategoryId == node.Id)
					this.selectedCategoryId = -1;
				this.roots.Remove(node);
			}
			else 
			{
				if(this.selectedCategoryId == node.Id)
					this.selectedCategoryId = node.FatherNode.Id;
				node.FatherNode.ChildNodes.Remove(node);
				if(node.FatherNode.ChildNodes.Count == 0)
					node.FatherNode.Kind = CategoryNodeKind.Leaf;
			}
		}

		// insert Node into CategoryForest

		internal enum ChangeCategoryResult {Ok, InvalidParams, Error}

		internal CategoryNavigationForest.ChangeCategoryResult ChangeCategoryName(int id, string name, out string errorMsg) 
		{
			if(id < -1) 
			{
				errorMsg = "Es wurde eine ungltige Kategorie-Identifikation bergeben.";
				return ChangeCategoryResult.InvalidParams;
			}
			string conStr = ConfigurationSettings.AppSettings["DbConnectionString"];
			IDbConnection con = null;
			try 
			{
				con = new SqlConnection(conStr);
				con.Open();
				IDbCommand cmd = con.CreateCommand();
				cmd.CommandText = "UPDATE " + DbTableOwnerPrefix + "Categories SET CategoryName = @CATNAME WHERE Id = @ID";
				cmd.Parameters.Add(new SqlParameter("@CATNAME",name));
				cmd.Parameters.Add(new SqlParameter("@ID",id));
				int updateCount = cmd.ExecuteNonQuery();
				if(updateCount != 1) 
				{
					log.Error("Update of Categories table for change of categoryname didn't affect exactly one row [#affected: " + updateCount + "].");
					errorMsg = "Es ist eine Datenbankinkonsistenz aufgetreten.";
					return ChangeCategoryResult.Error;
				}
			} 
			catch(Exception e) 
			{
				log.Error("Exception occured during update of categoryname [id: " + id + "] by database access", e);
				errorMsg = "Es ist ein Problem beim Datenbankzugriff aufgetreten.";
				return ChangeCategoryResult.Error;
			}
			finally 
			{
				try { if(con != null) con.Close(); }
				catch(Exception ex) 
				{
					log.Warn("Unable to close connection - exception occured.", ex);
				}
			}
			errorMsg = null;
			this.SelectedCategory.CategoryName = name;
			return ChangeCategoryResult.Ok;
		} // ChangeCategoryName

		internal enum InsertCategoryResult {Ok, InvalidParams, Error}
		
		internal CategoryNavigationForest.InsertCategoryResult InsertCategory(int parentId, string name, out string errorMsg) 
		{
			if(parentId < -1) 
			{
				errorMsg = "Es wurde eine unglige Identifikation der Einfgeposition bergeben.";
				return InsertCategoryResult.InvalidParams;
			}
			CategoryNode parentNode;
			if(parentId == -1) // insert as rootCategory
				parentNode = null;
			else 
			{
				parentNode = this.findCategoryNode(parentId);
				if(parentNode == null) 
				{
					errorMsg = "Die bergebene Einfgeposition existiert nicht.";
					return InsertCategoryResult.InvalidParams;
				}
			}
			// assertion: parentNode == null ... insert as rootCategory, parentNode != null ... insert as child of parentNode
			string conStr = ConfigurationSettings.AppSettings["DbConnectionString"];
			IDbConnection con = null;
			try 
			{
				con = new SqlConnection(conStr);
				con.Open();
				IDbCommand cmd = con.CreateCommand();
				if(parentNode == null) 
				{
					cmd.CommandText = "INSERT INTO " + DbTableOwnerPrefix + "Categories (CategoryName,Fk_Users) VALUES (@CATNAME,@USRID);" +
						"SELECT SCOPE_IDENTITY() AS [SCOPE_IDENTITY]";
					cmd.Parameters.Add(new SqlParameter("@CATNAME",name));
					cmd.Parameters.Add(new SqlParameter("@USRID",this.userId));
				}
				else 
				{
					cmd.CommandText = "INSERT INTO " + DbTableOwnerPrefix + "Categories (CategoryName,Fk_Users,Fk_ParentCategory) VALUES (@CATNAME,@USRID,@PARCAT);" +
						"SELECT SCOPE_IDENTITY() AS [SCOPE_IDENTITY]";
					cmd.Parameters.Add(new SqlParameter("@CATNAME",name));
					cmd.Parameters.Add(new SqlParameter("@USRID",this.userId));
					cmd.Parameters.Add(new SqlParameter("@PARCAT",parentNode.Id));
				}
				int insertedCategoryId = Convert.ToInt32((decimal)cmd.ExecuteScalar());
				// update CategoryForest object
				if(parentNode == null)
					this.roots.Add(new CategoryNode(insertedCategoryId, name, null));
				else 
				{
					parentNode.Kind = CategoryNodeKind.NonLeaf;
					parentNode.ChildNodes.Add(new CategoryNode(insertedCategoryId, name, parentNode));
				}
			} 
			catch(Exception e)
			{
				log.Error("Exception occured during insert of new Category [ParentId: " + parentId + "] by database access", e);
				errorMsg = "Es ist ein Problem beim Datenbankzugriff aufgetreten.";
				return CategoryNavigationForest.InsertCategoryResult.Error;
			} 
			finally 
			{
				try { if(con != null) con.Close(); } 
				catch(Exception ex) 
				{
					log.Warn("Unable to close connection - exception occured.", ex);
				}
			}
			errorMsg = null;
			return CategoryNavigationForest.InsertCategoryResult.Ok;
		} // end: method InsertCategory

		/// <summary>
		/// This method loads this CategoryNavigationForest, corresponding to a user identified by it's id, from the database.
		/// </summary>
		/// <returns>true .. successful, false .. error occured</returns>
		internal bool LoadWholeForest()
		{
			string conStr = ConfigurationSettings.AppSettings["DbConnectionString"];
			IDbConnection con = null;
			this.roots = new ArrayList();
			ArrayList tmpNodeList = new ArrayList();
			try 
			{
				con = new SqlConnection(conStr);
				con.Open();
				// Load root-categories from Database
				IDbCommand cmd = con.CreateCommand();
				cmd.CommandText = "SELECT Id, CategoryName FROM " + DbTableOwnerPrefix + "Categories WHERE Fk_Users='" + this.userId + "' AND Fk_ParentCategory IS NULL";
				//cmd.Parameters.Add(new SqlParameter("@UID",SqlDbType.Int));
				//cmd.Parameters["@UID"] = this.userId;
				IDataReader reader = cmd.ExecuteReader();
				while(reader.Read()) 
				{
					this.roots.Add(new CategoryNode((int)reader["Id"],(string)reader["CategoryName"], null));
				}
				reader.Close();
				// recursivly load Subcategories
				foreach(CategoryNode tmpNode in this.roots) 
				{
					if(!this.LoadChilds(tmpNode, con))
						return false;
				}
			} 
			catch(Exception e)
			{
				log.Error("Exception occured during load of Categories from database", e);
				return false;
			} 
			finally 
			{
				try { if(con != null) con.Close(); } 
				catch(Exception ex) 
				{
					log.Warn("Unable to close connection - exception occured.", ex);
				}
			}
			return true;
		} // end: method LoadWholeForest

		// Recursive helper-method to recursively load subcategories from database - outside a transaction
		private bool LoadChilds(CategoryNode parentNode, IDbConnection con) 
		{
			string conStr = ConfigurationSettings.AppSettings["DbConnectionString"];
			parentNode.ChildNodes = new ArrayList();
			try 
			{
				IDbCommand cmd = con.CreateCommand();
				cmd.CommandText = "SELECT Id, CategoryName FROM " + DbTableOwnerPrefix + "Categories WHERE Fk_ParentCategory = '" + parentNode.Id + "'";
				//cmd.Parameters.Add(new SqlParameter("@ParCat",SqlDbType.Int));
				//cmd.Parameters["@ParCat"] = parentNode.Id;
				IDataReader reader = cmd.ExecuteReader();
				while(reader.Read()) 
				{
					parentNode.Kind = CategoryNodeKind.NonLeaf;
					parentNode.ChildNodes.Add(new CategoryNode((int)reader["Id"],(string)reader["CategoryName"], parentNode));
				}
				reader.Close();
			} 
			catch(Exception e)
			{
				log.Error("Exception occured during load of CategoryNode from database", e);
				return false;
			} 
			foreach(CategoryNode tmpNode in parentNode.ChildNodes)
				if(!this.LoadChilds(tmpNode, con))
					return false;
			return true;
		} // end: method LoadChilds
		
		// Deleting a Node in the CatagoryNavigationForest

		internal enum DeleteCategoryResult {Ok, Failed, Error}
		
		/// <summary>
		/// This method deletes a Category with possible child-Categories from the Database and
		/// updates the NavigationForest accordingly. The update of the NavigationForest is not
		/// database driven but consistent with the database-access-responses.
		/// </summary>
		/// <param name="id">Id of the Category to delete.</param>
		/// <param name="errorMsg">Errormessage in case of a returnvalue other than Ok.</param>
		/// <returns>Ok, Error</returns>
		internal CategoryNavigationForest.DeleteCategoryResult DeleteCategory(int id, out string errorMsg) 
		{
			string conStr = ConfigurationSettings.AppSettings["DbConnectionString"];
			CategoryNode deleteRootNode = this.findCategoryNode(id);
			if(deleteRootNode == null) 
			{
				errorMsg = "Es wurde versucht eine nicht vorhandene Kategorie zu lschen.";
				return CategoryNavigationForest.DeleteCategoryResult.Error;
			}
			IDbConnection con = null;
			IDbTransaction trans = null;
			// load root Categories
			try 
			{
				con = new SqlConnection(conStr);
				con.Open();
				trans = con.BeginTransaction();
				if(!this.recursiveDeleteCategory(deleteRootNode, con, trans)) 
				{
					trans.Rollback();
					errorMsg = "Es ist ein Fehler beim Datenbankzugriff aufgetreten.";
					return CategoryNavigationForest.DeleteCategoryResult.Error;
				}
				else 
				{
					this.removeNodeFromForest(deleteRootNode);
					trans.Commit();
					errorMsg = "";
					return CategoryNavigationForest.DeleteCategoryResult.Ok;
				}
			} 
			catch(Exception e)
			{
				if(trans != null)
					trans.Rollback();
				log.Error("Exception occured during deleting a Category [Id: " + id + "]", e);
				errorMsg = "Es ist ein Fehler beim Datenbankzugriff aufgetreten.";
				return CategoryNavigationForest.DeleteCategoryResult.Error;
			} 
			finally 
			{
				try { if(con != null) con.Close(); } 
				catch(Exception ex) 
				{
					log.Warn("Unable to close connection - exception occured.", ex);
				}
			}
		} // end: method DeleteCategory

		private bool recursiveDeleteCategory(
			CategoryNode node,
			IDbConnection con,
			IDbTransaction trans) 
		{
			if(node == null)
				return false;
			// Delete Categories by postorder traversal, so it is guaranteed that each individual
			// delete is done for a leaf-node.
			if(node.ChildNodes != null) 
			{
				foreach(CategoryNode child in node.ChildNodes)
				{
					if(!this.recursiveDeleteCategory(child, con, trans))
						return false;
				}
			} 
			try 
			{
				IDbCommand cmd = con.CreateCommand();
				// Note: Rows of other tables referencing the category to delete are automatically
				//       deleted by the database due to cascading delete propagation.
				cmd.CommandText = "DELETE FROM " + DbTableOwnerPrefix + "Categories WHERE Id = '" + node.Id + "'";
				cmd.Transaction = trans;
				int numAffectedRows = cmd.ExecuteNonQuery();
				if(numAffectedRows != 1) 
				{
					log.Error("Deleting Category didnt't affect exactly one Row as expected [Id: " + node.Id + " #Affected: " + numAffectedRows + "]");
					return false;
				}
			} 
			catch(Exception e) 
			{
				log.Error("An exception occured during deleting Category [Id: " + node.Id + "]", e);
				return false;
			}
			return true;
		} // end: method deleteCategoryRecursively


		/******************************************************************************************************
		 ***** Html-Representatin specific stuff                                                         ******
		 ******************************************************************************************************/
		
		internal string GenerateHtmlRepresentation(
			string selectedFolderImg,
			string unselectedFolderImg,
			string plusImg,
			string minusImg,
			string neutralImg,
			string navigBaseUrl,
			string contentBaseUrl) 
		{
			this.selectedFolderImg = selectedFolderImg;
			this.unselectedFolderImg = unselectedFolderImg;
			this.plusImg = plusImg;
			this.minusImg = minusImg;
			this.neutralImg = neutralImg;
			this.navigBaseUrl = navigBaseUrl;
			this.contentBaseUrl = contentBaseUrl;
			StringBuilder sb = new StringBuilder();
			sb.Append("<nobr>");
			if(roots.Count > 0)
				sb.Append("<img src=\"" + this.minusImg + "\" border=\"0\" />");
			else
				sb.Append("<img src=\"" + this.neutralImg + "\" border=\"0\" />");
			if(this.selectedCategoryId == -1)
				sb.Append("<a href=\"" + this.contentBaseUrl + "?show=-1\" target=\"content\"><span><img src=\"" + this.selectedFolderImg + "\" border=\"0\" /><span style=\"POSITION: relative; BOTTOM: 2px; BACKGROUND-COLOR: #003f94; COLOR: white; TEXT-DECORATION: none;\">&nbsp;/&nbsp;</span></span></a>");
			else
				sb.Append("<a href=\"" + this.contentBaseUrl + "?show=-1\" target=\"content\"><span><img src=\"" + this.unselectedFolderImg + "\" border=\"0\" /><span style=\"POSITION: relative; BOTTOM: 2px; TEXT-DECORATION: none;\">&nbsp;/&nbsp;</span></span></a>");
			sb.Append("<br>");
			//sb.Append("<font size=\"-1\">");
			foreach(CategoryNode tmpNode in roots) 
			{
				this.recursiveGenerateHtmlRepresentation(1,sb,tmpNode);
			}
			//sb.Append("</font>");
			sb.Append("</nobr>");
			return sb.ToString();
		}

		private void recursiveGenerateHtmlRepresentation(
			int depth,
			StringBuilder sb,
			CategoryNode node) 
		{
			sb.Append(this.generateHtmlRepresentationOfSingleNode(depth, node));
			if(node.State == CategoryNodeState.Open)
				foreach(CategoryNode tmpNode in node.ChildNodes)
					this.recursiveGenerateHtmlRepresentation(depth + 1, sb, tmpNode);
		}

		private string generateHtmlRepresentationOfSingleNode(
			int depth,
			CategoryNode node) 
		{
			StringBuilder sb = new StringBuilder();
			for(int i = 0; i < depth; i++)
				sb.Append("&nbsp;&nbsp;&nbsp;");
			if(node.Kind == CategoryNodeKind.Leaf) 
			{
				sb.Append("<img src=\"" + this.neutralImg + "\" border=\"0\" />");
			} 
			else 
			{
				if(node.State == CategoryNodeState.Close)
					sb.Append("<a href=\"" + this.navigBaseUrl + "?open=" + node.Id + "\"><img src=\"" + this.plusImg + "\" border=\"0\" /></a>");
				else if(node.State == CategoryNodeState.Open)
					sb.Append("<a href=\"" + this.navigBaseUrl + "?close=" + node.Id + "\"><img src=\"" + this.minusImg + "\" border=\"0\" /></a>");
			}
			if(node.Id == this.selectedCategoryId)
				sb.Append("<a href=\"" + this.contentBaseUrl + "?show=" + node.Id + "\" target=\"content\"><span><img src=\"" + this.selectedFolderImg + "\" border=\"0\" /><span style=\"POSITION: relative; BOTTOM: 2px; BACKGROUND-COLOR: #003f94; COLOR: white; TEXT-DECORATION: none;\">&nbsp;" + node.CategoryName + "&nbsp;</span></span></a>");
			else
				sb.Append("<a href=\"" + this.contentBaseUrl + "?show=" + node.Id + "\" target=\"content\"><span><img src=\"" + this.unselectedFolderImg + "\" border=\"0\" /><span style=\"POSITION: relative; BOTTOM: 2px; TEXT-DECORATION: none;\">&nbsp;" + node.CategoryName + "&nbsp;</span></span></a>");
			sb.Append("<br>");
			return sb.ToString();
		}
		
	}
}