using System;
using System.Collections;
using System.Text;
using WebseitenMonitor.DomainObjects.Behaviour;
using log4net;

namespace WebseitenMonitor.Ui 
{
	internal class ObservedDocumentsTree 
	{
		private static readonly ILog log = LogManager.GetLogger(typeof(ObservedDocumentsTree));

		internal int Id;
		internal string Name;
		internal Uri Url;
		internal ChangeStateEnum ChangeState;
		internal DateTime Date;
		internal int Depth;
		
		internal string RootDocument 
		{
			get { return (this.rootDocument == null)?"null":this.rootDocument.title; }
		}

		private ObservedDocumentsTree.ObservedDocument rootDocument;
		private Hashtable NonLeafDocumentHashtable;

		internal static readonly string unchangedColor = "#000000";
		internal static readonly string unchangedBackColor = "#00ff00";
		internal static readonly string changedColor = "#ffff00";
		internal static readonly string changedBackColor = "#0000ff";
		internal static readonly string newColor = "ff0000";
		internal static readonly string newBackColor = "#00ffff";
		internal static readonly string invalidMimeTypeColor = "#000000";
		internal static readonly string invalidMimeTypeBackColor = "#909090";
		internal static readonly string unretrievableColor = "#ffffff";
		internal static readonly string unretrievableBackColor = "#ff0000";
		internal static readonly string notHashableColor = "#ffffff";
		internal static readonly string notHashableBackColor = "#ff8000";

		private static readonly string baseUrl = "./ChangeObservationDetail.aspx";
		private static readonly string plusImg = "../img/PlusSmall.bmp";
		private static readonly string minusImg = "../img/MinusSmall.bmp";
		private static readonly string plusRecursiveImg = "../img/PlusRecursive.bmp";
		private static readonly string minusRecursiveImg = "../img/MinusRecursive.bmp";
		private static readonly string followLinkImg = "../img/FollowLinkImage.bmp";

		private static readonly ObservedDocumentsTree.DescendentOrSelfChangeStateAggregation WholeFilterPattern = 
			DescendentOrSelfChangeStateAggregation.Unchanged | DescendentOrSelfChangeStateAggregation.Changed |
			DescendentOrSelfChangeStateAggregation.New | DescendentOrSelfChangeStateAggregation.InvalidMimeType |
			DescendentOrSelfChangeStateAggregation.Unretrievable | DescendentOrSelfChangeStateAggregation.NotHashable;

		private StringBuilder sb;
		private ObservedDocumentsTree.DescendentOrSelfChangeStateAggregation filterPattern;
		
		private int idCounter; // for id generation

		internal enum ViewEnum {Closed, Opened}
		internal enum DescendentOrSelfChangeStateAggregation 
		{
			Unchanged = 1,
			Changed = 2,
			New = 4,
			InvalidMimeType = 8,
			Unretrievable = 16,
			NotHashable = 32
		}

		private class ObservedDocument 
		{
			internal ObservedDocument(
				int id, 
				string title, 
				Uri url, 
				ChangeObservation.ObservedDocumentState changeState) 
			{
				this.id = id; // not equivalent with database id, because ObservedDocumentsTree need not be loaded from database
				this.title = title;
				this.url = url;
				this.changeState = changeState;
				this.viewState = ObservedDocumentsTree.ViewEnum.Closed;
				this.changeStateAggregation = (ObservedDocumentsTree.DescendentOrSelfChangeStateAggregation)0;
				ObservedDocumentsTree.aggregateChangeStateAggregation(changeState, ref this.changeStateAggregation);
				this.isLeaf = true;
			}
			internal int id;
			internal string title;
			internal Uri url;
			internal ChangeObservation.ObservedDocumentState changeState;
			internal ArrayList childDocuments;
			internal ObservedDocumentsTree.ViewEnum viewState;
			internal ObservedDocumentsTree.DescendentOrSelfChangeStateAggregation changeStateAggregation;
			internal bool isLeaf;
		}

		internal ObservedDocumentsTree(ChangeObservation obs) 
		{
			this.filterPattern = WholeFilterPattern;
			this.Id = obs.Id;
			this.Name = obs.Name;
			this.Url = obs.Url;
			this.ChangeState = obs.ChangeState;
			this.Date = obs.Date;
			this.Depth = obs.Depth;
			if(obs.RootDocument == null)
				return;
			log.Debug("Init FilterPattern: " + (int)this.filterPattern);
			this.buildObservedDocumentTree(obs.Depth, obs.RootDocument);
		}

		internal void SetFilterPattern(DescendentOrSelfChangeStateAggregation filterPattern) 
		{
			this.filterPattern = filterPattern;
		}

		internal bool ReadFilterPattern(DescendentOrSelfChangeStateAggregation filterToRead) 
		{
			if((int)(this.filterPattern & filterToRead) != 0)
				return true;
			else
				return false;
		}

		private int nextId() 
		{
			return this.idCounter++;
		}

		private void buildObservedDocumentTree(int depth, ChangeObservation.ObservedDocument rootDoc) 
		{
			this.rootDocument = new ObservedDocument(
				this.nextId(),
				rootDoc.title,
				rootDoc.url,
				rootDoc.state);
			this.NonLeafDocumentHashtable = new Hashtable();
			if(depth <= 0)
				return;
			else if(rootDoc.childDocuments != null) 
			{
				this.rootDocument.childDocuments = new ArrayList();
				foreach(ChangeObservation.ObservedDocument childDoc in rootDoc.childDocuments) 
				{
					ObservedDocumentsTree.ObservedDocument tmpObsDoc;
					if(childDoc.kind == ChangeObservation.ObservedDocumentKind.Delegate)
						tmpObsDoc = this.recursiveBuildObservedDocumentTree(depth -1, childDoc.delegateFor);
					else // Original
						tmpObsDoc = this.recursiveBuildObservedDocumentTree(depth - 1, childDoc);
					if(tmpObsDoc != null) 
					{
						this.rootDocument.childDocuments.Add(tmpObsDoc);
						aggregateChangeStateAggregation(tmpObsDoc.changeState, ref this.rootDocument.changeStateAggregation);
					}
				}
				this.NonLeafDocumentHashtable.Add(rootDocument.id, this.rootDocument);
				if(this.rootDocument.childDocuments.Count > 0)
					this.rootDocument.isLeaf = false;
			}
		}

		private ObservedDocumentsTree.ObservedDocument recursiveBuildObservedDocumentTree(
			int depth, 
			ChangeObservation.ObservedDocument doc) 
		{
			if(depth < 0)
				return null;
			ObservedDocumentsTree.ObservedDocument obsDoc = new ObservedDocumentsTree.ObservedDocument(
				this.nextId(),
				doc.title,
				doc.url,
				doc.state);
			if(depth == 0)
				return obsDoc;
			else if(doc.childDocuments != null) 
			{
				obsDoc.childDocuments = new ArrayList();
				foreach(ChangeObservation.ObservedDocument childDoc in doc.childDocuments) 
				{
					ObservedDocumentsTree.ObservedDocument tmpObsDoc;
					if(childDoc.kind == ChangeObservation.ObservedDocumentKind.Delegate)
						tmpObsDoc = this.recursiveBuildObservedDocumentTree(depth - 1, childDoc.delegateFor);
					else // Original
						tmpObsDoc = this.recursiveBuildObservedDocumentTree(depth -1, childDoc);
					if(tmpObsDoc != null) 
					{
						obsDoc.childDocuments.Add(tmpObsDoc);
						aggregateChangeStateAggregation(tmpObsDoc.changeState, ref obsDoc.changeStateAggregation);
					}
				}
				this.NonLeafDocumentHashtable.Add(obsDoc.id, obsDoc);
				if(obsDoc.childDocuments.Count > 0)
					obsDoc.isLeaf = false;
				return obsDoc;
			}
			return obsDoc;
		}

		internal static void aggregateChangeStateAggregation(
			ChangeObservation.ObservedDocumentState toAdd, 
			ref ObservedDocumentsTree.DescendentOrSelfChangeStateAggregation toUpdate) 
		{
			switch(toAdd) 
			{
				case ChangeObservation.ObservedDocumentState.Unchanged:
					toUpdate = toUpdate | ObservedDocumentsTree.DescendentOrSelfChangeStateAggregation.Unchanged;
					break;
				case ChangeObservation.ObservedDocumentState.Changed:
					toUpdate = toUpdate | ObservedDocumentsTree.DescendentOrSelfChangeStateAggregation.Changed;
					break;
				case ChangeObservation.ObservedDocumentState.New:
					toUpdate = toUpdate | ObservedDocumentsTree.DescendentOrSelfChangeStateAggregation.New;
					break;
				case ChangeObservation.ObservedDocumentState.InvalidMimeType:
					toUpdate = toUpdate | ObservedDocumentsTree.DescendentOrSelfChangeStateAggregation.InvalidMimeType;
					break;
				case ChangeObservation.ObservedDocumentState.Unretrievable:
					toUpdate = toUpdate | ObservedDocumentsTree.DescendentOrSelfChangeStateAggregation.Unretrievable;
					break;
				case ChangeObservation.ObservedDocumentState.NotHashable:
					toUpdate = toUpdate | ObservedDocumentsTree.DescendentOrSelfChangeStateAggregation.NotHashable;
					break;
			}
		}

		internal void OpenDocumentChilds(int id) 
		{
			ObservedDocumentsTree.ObservedDocument obsDoc = (ObservedDocumentsTree.ObservedDocument)this.NonLeafDocumentHashtable[id];
			if(obsDoc != null)
				obsDoc.viewState = ObservedDocumentsTree.ViewEnum.Opened;
		}

		internal void OpenDocumentDescendentsOrSelf(int id) 
		{
			ObservedDocumentsTree.ObservedDocument obsDoc = (ObservedDocumentsTree.ObservedDocument)this.NonLeafDocumentHashtable[id];
			if(obsDoc != null)
				this.recursiveOpenDocumentDescendentsOrSelf(obsDoc);
		}

		private void recursiveOpenDocumentDescendentsOrSelf(ObservedDocumentsTree.ObservedDocument doc) 
		{
			if(doc != null) 
			{
				doc.viewState = ViewEnum.Opened;
				if(doc.childDocuments != null)
					foreach(ObservedDocumentsTree.ObservedDocument tmpDoc in doc.childDocuments)
						this.recursiveOpenDocumentDescendentsOrSelf(tmpDoc);
			}
		}

		internal void CloseDocumentChilds(int id) 
		{
			ObservedDocumentsTree.ObservedDocument obsDoc = (ObservedDocumentsTree.ObservedDocument)this.NonLeafDocumentHashtable[id];
			if(obsDoc != null)
				obsDoc.viewState = ObservedDocumentsTree.ViewEnum.Closed;
		}

		internal void CloseDocumentDescendentsOrSelf(int id) 
		{
			ObservedDocumentsTree.ObservedDocument obsDoc = (ObservedDocumentsTree.ObservedDocument)this.NonLeafDocumentHashtable[id];
			if(obsDoc != null)
				this.recursiveCloseDocumentDescendentsOrSelf(obsDoc);
		}

		private void recursiveCloseDocumentDescendentsOrSelf(ObservedDocumentsTree.ObservedDocument doc) 
		{
			if(doc != null) 
			{
				doc.viewState = ViewEnum.Closed;
				if(doc.childDocuments != null)
					foreach(ObservedDocumentsTree.ObservedDocument tmpDoc in doc.childDocuments)
						this.recursiveCloseDocumentDescendentsOrSelf(tmpDoc);
			}
		}

		internal string GenerateHtmlRepresentation() 
		{
			if(this.rootDocument == null)
				if(this.ChangeState == ChangeStateEnum.Init)
					return "";
				else
					return "Keine Beobachtung gefunden.";
			this.sb = new StringBuilder();
			this.recursiveGenerateHtmlRepresentation(this.Depth, this.rootDocument);
			return sb.ToString();
		}

		private void recursiveGenerateHtmlRepresentation(
			int depth, 
			ObservedDocumentsTree.ObservedDocument doc) 
		{
			if(((int)(doc.changeStateAggregation & this.filterPattern)) != 0) 
			{
				this.generateHtmlRepresentationOfSingleObservedDocument(depth, doc);
				if(doc.childDocuments != null)
					if(doc.viewState == ViewEnum.Opened)
						foreach(ObservedDocumentsTree.ObservedDocument tmpDoc in doc.childDocuments)
							this.recursiveGenerateHtmlRepresentation(depth - 1, tmpDoc);
			}
		}

		private void generateHtmlRepresentationOfSingleObservedDocument(int depth, ObservedDocumentsTree.ObservedDocument doc) 
		{
			// depth indent
			int indent = (this.Depth - depth) * 40;
			sb.Append("<span style=\"POSITION: relative; LEFT: " + indent + "px;\">");
			// open/close img
			if(!doc.isLeaf) 
			{
				if(doc.viewState == ObservedDocumentsTree.ViewEnum.Closed) 
				{
					this.sb.Append("<a href=\"" + baseUrl + "?open=" + doc.id.ToString() + "\"><img src=\"" + plusImg + "\" border=\"0\" /></a>");
					this.sb.Append("<span style=\"POSITION: relative; LEFT: 5px;\"><a href=\"" + baseUrl + "?openrec=" + doc.id.ToString() + "\"><img src=\"" + plusRecursiveImg + "\" border=\"0\" /></a></span>");
				}
				else 
				{
					this.sb.Append("<a href=\"" + baseUrl + "?close=" + doc.id.ToString() + "\"><img src=\"" + minusImg + "\" border=\"0\" /></a>");
					this.sb.Append("<span style=\"POSITION: relative; LEFT: 5px;\"><a href=\"" + baseUrl + "?closerec=" + doc.id.ToString() + "\"><img src=\"" + minusRecursiveImg + "\" border=\"0\" /></a></span>");
				}
				this.sb.Append("<span style=\"POSITION: relative; LEFT: 10px;\"><a href=\"" + doc.url.ToString() + "\" target=\"_blank\"><img src=\"" + followLinkImg + "\" border=\"0\" /></a></span>");
				this.sb.Append("<span style=\"POSITION: relative; LEFT: 15px; ");
			}
			else 
			{
				this.sb.Append("<span style=\"POSITION: relative; LEFT: 10px;\"><a href=\"" + doc.url.ToString() + "\" target=\"_blank\"><img src=\"" + followLinkImg + "\" border=\"0\" /></a></span>");
				this.sb.Append("<span style=\"POSITION: relative; LEFT: 15px; ");
			}
			// set corresponding background-color/color
			if(doc.changeState == ChangeObservation.ObservedDocumentState.Unchanged)
				this.sb.Append("BACKGROUND-COLOR: " + unchangedBackColor + "; COLOR: " + unchangedColor + "\">");
			else if(doc.changeState == ChangeObservation.ObservedDocumentState.Changed)
				this.sb.Append("BACKGROUND-COLOR: " + changedBackColor + "; COLOR: " + changedColor + "\">");
			else if(doc.changeState == ChangeObservation.ObservedDocumentState.New)
				this.sb.Append("BACKGROUND-COLOR: " + newBackColor + "; COLOR: " + newColor + "\">");
			else if(doc.changeState == ChangeObservation.ObservedDocumentState.InvalidMimeType)
				this.sb.Append("BACKGROUND-COLOR: " + invalidMimeTypeBackColor + "; COLOR: " + invalidMimeTypeColor + "\">");
			else if(doc.changeState == ChangeObservation.ObservedDocumentState.Unretrievable)
				this.sb.Append("BACKGROUND-COLOR: " + unretrievableBackColor + "; COLOR: " + unretrievableColor + "\">");
			else if(doc.changeState == ChangeObservation.ObservedDocumentState.NotHashable)
				this.sb.Append("BACKGROUND-COLOR: " + notHashableBackColor + "; COLOR: " + notHashableColor + "\">");
			// title and Url
			this.sb.Append("&nbsp;Titel:&nbsp;" + doc.title + "&nbsp;&nbsp;Url:&nbsp;" + doc.url.ToString() + "&nbsp;");
			// close span
			this.sb.Append("</span></span><br>");
		}

	}
}