using System;
using System.Text;

namespace WebseitenMonitor.UriUtilities 
{
	public class UriConvertion 
	{
		/// <summary>
		/// This method returns the resulting absolute Uri out of the string uriToMakeAbsolute.
		/// The baseUri-object represents a baseUri which should be used as base if uriToMakeAbsolute
		/// is a relative one.
		/// Note: There are no assumptions about the base Uri.
		/// 
		/// Note: Excerpt from RFC1630 - URIs in WWW
		///		The rules for the use of a partial name relative to the URI of the context are:
		///
		///			If the scheme parts are different, the whole absolute URI must
		///			be given.  Otherwise, the scheme is omitted, and:
		///
		///				If the partial URI starts with a non-zero number of consecutive
		///				slashes, then everything from the context URI up to (but not
		///				including) the first occurrence of exactly the same number of
		///				consecutive slashes which has no greater number of consecutive
		///				slashes anywhere to the right of it is taken to be the same and
		///				so prepended to the partial URL to form the full URL. Otherwise:
		///			
		///				The last part of the path of the context URI (anything following
		///				the rightmost slash) is removed, and the given partial URI
		///				appended in its place, and then:
		///
		///				Within the result, all occurrences of "xxx/../" or "/." are
		///				recursively removed, where xxx, ".." and "." are complete path
		///				elements.
		///		
		///			Note: Trailing slashes
		///
		///				If a path of the context locator ends in slash, partial URIs are
		///				treated differently to the URI with the same path but without a
		///				trailing slash. The trailing slash indicates a void segment of the
		///				path.
		/// </summary>
		/// <param name="contextUri"></param>
		/// <param name="uriToMakeAbsolute">absolute or relative uri-string</param>
		/// <returns>null - it wasn't possible to obtain an absolute Uri; absolute Uri - otherwise</returns>
		internal static Uri GetAbsoluteUriCorrToContext(Uri contextUri, string uriToMakeAbsolute) 
		{
			if(uriToMakeAbsolute == null)
				return null;
			uriToMakeAbsolute = uriToMakeAbsolute.Trim();
			// check whether uriToMakeAbsolute is already an absolute wellformed uri
			Uri tmpUri = null;
			try 
			{
				tmpUri = new Uri(uriToMakeAbsolute);
			} 
			catch(UriFormatException){}
			if(tmpUri != null) 
			{
				// assertion: uriToMakeAbsolute is already a wellformed absolute URI
				return tmpUri;
			}
			// assertion: uriToMakeAbsolute is not an absolute one and therefore relative (or not wellformed)
			if(contextUri == null)
				return null;
			int slashStartCount = 0;
			for(int i = 0; i < uriToMakeAbsolute.Length; i++) 
				if(uriToMakeAbsolute[i] == '/') 
					slashStartCount ++;
				else
					break;
			string contextUriString = contextUri.ToString();
			int anchorIndex = -1;
			bool overreadSlashes = false;
			bool isRelUriWithAbsolutePath = slashStartCount > 0;
			if(!isRelUriWithAbsolutePath)
				slashStartCount = 1; // set anchor to root for later helping use
			int numSlashMatchesLeft = slashStartCount;
			// find anchor for relative URI with absolute path or root index
			for(int i = 0; i < contextUriString.Length; i++) 
			{
				if(contextUriString[i] == '/') 
				{
					if(!overreadSlashes) 
					{
						if(numSlashMatchesLeft == slashStartCount)
							anchorIndex = i;
						numSlashMatchesLeft--;
						if(numSlashMatchesLeft == 0)  
						{
							// check if following is further slash
							if(i == contextUriString.Length - 1)
								break; // anchor found
							else 
							{
								if(contextUriString[i + 1] == '/') 
								{
									overreadSlashes = true;
									anchorIndex = -1;
									numSlashMatchesLeft = slashStartCount;
								} 
								else
									break; // anchor found
							}
						}
					}
				} 
				else 
				{
					anchorIndex = -1;
					numSlashMatchesLeft = slashStartCount;
					overreadSlashes = false;
				}
			}
			if(isRelUriWithAbsolutePath) // anchor for relative URI with absolute path found (starts with '/')
			{
				try 
				{
					tmpUri = new Uri(contextUriString.Substring(0,anchorIndex) + uriToMakeAbsolute);
				} 
				catch(UriFormatException) {}
				return tmpUri; // null if UriFormatException occured, otherwise resulting absolute Uri
			} 
			else // !isRelUriWithAbsolutePath: relative Uri hasn't absolute path (doesn't start with '/') 
			{
				if(anchorIndex < 0) // should normally never happen
					anchorIndex = 0;
				int rightmostSlashIndex = -1;
				// search for rightmost slash
				for(int i = contextUriString.Length - 1; i >= 0; i--) 
				{
					if(contextUriString[i] == '/') 
					{
						rightmostSlashIndex = i;
						break;
					}
				}
				if(rightmostSlashIndex == -1) // no rightmost-slash found - malformed Uri
					return null;
				StringBuilder contextUriSb = new StringBuilder(contextUriString.Substring(0,rightmostSlashIndex + 1));
				contextUriSb.Append(uriToMakeAbsolute);
				int outerStartIndex = rightmostSlashIndex;
				int outerEndIndex = -1;
				string segString = null;
				while( (segString = readSegment(contextUriSb, outerStartIndex, out outerEndIndex)) != null ) 
				{
					// further segement found
					if(segString.Equals(".")) // this dir
					{
						if(outerEndIndex == -1) 
						{ 
							// last segement in Uri
							contextUriSb.Remove(outerStartIndex + 1, contextUriSb.Length - outerStartIndex - 1);
							break;
						} 
						else 
						{
							// assertion: outerEndIndex is outer index of current segment and indexes a '/' char
							// remove this segement from Uri
							contextUriSb.Remove(outerStartIndex + 1, outerEndIndex - outerStartIndex);
							// Note: outerStartIndex is correct for next loop pass through
						}
					} 
					else if(segString.Equals("..")) // father dir
					{
						// determine previous outerStartIndex
						int prevOuterStartIndex = getPreviousOuterStartIndex(contextUriSb, anchorIndex, outerStartIndex);
						if(prevOuterStartIndex == -1) // malformed Uri (too less parent directories)
							return null;
						if(outerEndIndex == -1) 
						{
							// last segment in Uri
							// remove this and it's parent
							contextUriSb.Remove(prevOuterStartIndex + 1, contextUriSb.Length - prevOuterStartIndex - 1);
							break;
						} 
						else 
						{
							// assertion: outerEndIndex is outer index of current segment and indexes a '/' char
							// remove this and it's parent
							contextUriSb.Remove(prevOuterStartIndex + 1, outerEndIndex - prevOuterStartIndex);
							outerStartIndex = prevOuterStartIndex;
							// Note: outerStartIndex got corrected for next loop pass through
						}
					} 
					else 
					{
						if(outerEndIndex == -1) // last segment in Uri
							break;
						outerStartIndex = outerEndIndex;
					}
				}
				// assertion: contextUriSb holds result Uri-string with all "." and ".." directories resolved
				try 
				{
					tmpUri = new Uri(contextUriSb.ToString());
				} 
				catch(UriFormatException) {}
				return tmpUri; // null if UriFormatException occured, otherwise resulting absolute Uri
				
			}
		}
		
		// returns -1 if no previous segement found
		private static int getPreviousOuterStartIndex(StringBuilder sb, int rootSlashIndex, int outerStartIndex) 
		{
			if(outerStartIndex <= rootSlashIndex)
				return -1;
			int startIndex = outerStartIndex - 1;
			for(int i = startIndex; i >= rootSlashIndex; i--) 
			{
				if(sb[i] == '/')
					return i;
			}
			return - 1;

		}
		
		// outerEndIndex gets -1 if end-reached
		private static string readSegment(StringBuilder sb, int outerStartIndex, out int outerEndIndex) 
		{
			if(outerStartIndex == sb.Length - 1) // already at the end
			{
				outerEndIndex = -1;
				return null;
			}
			for(int i = outerStartIndex + 1; i < sb.Length; i++) {
				if(sb[i] == '/') 
				{
					outerEndIndex = i;
					return sb.ToString(outerStartIndex + 1, i - outerStartIndex - 1);
				}
			}
			// assertion: read segement is last in sb
			outerEndIndex = -1;
			return sb.ToString(outerStartIndex + 1, sb.Length - outerStartIndex - 1);
		}

		/// <summary>
		/// e.g:
		///     http://localhost/base/index.htm?name=val:name1=val1 --> http://localhost/base/
		///     http://localhost/base/index.htm#fragment            --> http://localhost/base/
		///     http://localhost/base/index.htm                     --> http://localhost/base/
		///     http://localhost/base/index#fragment                --> http://localhost/base/
		///     http://localhost/base/                              --> http://localhost/base/
		///     http://localhost/base                               --> http://localhost/
		///     
		///     Note: That the method can expect sourceResponseUri as valid absolute Uri, because
		///     this is checked by the Uri-constructor.
		///     
		///     So the result BasePathUri always ends with a slash.
		/// </summary>
		/// <param name="sourceResponseUri">Uri of an arbritary Web-Response</param>
		/// <returns>The sourceResponseUri's Base-Path-Uri, or null if impossible to obtain - should normally never happen.</returns>
		internal static string getUrisBasePathUriString(Uri sourceResponseUri) 
		{
			if(sourceResponseUri == null)
				return null;
			string uriString = sourceResponseUri.ToString();
			int rightmostSlashIndex = -1;
			for(int i = uriString.Length - 1; i >= 0; i--) 
			{
				if(uriString[i] == '/') 
				{
					rightmostSlashIndex = i;
					break;
				}
			}
			if(rightmostSlashIndex == -1)
				return null;
			return uriString.Substring(0, rightmostSlashIndex + 1);
		} // end: method getUrisBasePathUriString

	} // end: class UriConvertion
}