using System;
using System.Collections;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Web;
using System.Web.SessionState;
using System.Web.UI;
using System.Web.UI.WebControls;
using System.Web.UI.HtmlControls;
using WebseitenMonitor.Ui;
using WebseitenMonitor.DataAccess;
using WebseitenMonitor.DomainObjects.Behaviour;
using log4net;
using WebseitenMonitor.Ui.CustomControls;
using WebseitenMonitor.User.Help;

namespace WebseitenMonitor.User
{
	/// <summary>
	/// Summary description for ChangeObservationEditProperties.
	/// </summary>
	public class ChangeObservationEditProperties : System.Web.UI.Page
	{
		private static readonly ILog log = LogManager.GetLogger(typeof(ChangeObservationEditProperties));

		protected TextBox NameTextBox;
		protected TextBox UrlTextBox;
		protected System.Web.UI.WebControls.RequiredFieldValidator NameRequiredFieldValidator;
		protected System.Web.UI.WebControls.CustomValidator NameCustomValidator;
		protected System.Web.UI.WebControls.RequiredFieldValidator UrlRequiredFieldValidator;
		protected System.Web.UI.WebControls.CustomValidator UrlCustomValidator;
		protected System.Web.UI.WebControls.RequiredFieldValidator DepthRequiredFieldValidator;
		protected System.Web.UI.WebControls.CustomValidator DepthCustomValidator;
		protected System.Web.UI.WebControls.Button OkButton;
		protected System.Web.UI.WebControls.Button CancelButton;
		protected TextBox DepthTextBox;
		protected Fold NameHelpFold;
		protected Fold UrlHelpFold;
		protected Fold DepthHelpFold;

		private void Page_Load(object sender, System.EventArgs e)
		{
			// Show current Properties as default values
			if(!IsPostBack) 
			{
				ObservedDocumentsTree obsTree = (ObservedDocumentsTree)Session["ObservedDocumentsTree"];
				if(obsTree == null) 
				{
					Session["ErrorMessage"] = "Es konnte kein aktueller Beobachtungskanal eruiert werden.";
					Response.Redirect("./ErrorPage.aspx");
					return;
				}
				this.NameTextBox.Text = obsTree.Name;
				this.UrlTextBox.Text = obsTree.Url.ToString();
				this.DepthTextBox.Text = obsTree.Depth.ToString();
				this.NameHelpFold.AlternateText = HelpTextConstants.NameHelpHtmlText;
				this.UrlHelpFold.AlternateText = HelpTextConstants.UrlHelpHtmlText;
				this.DepthHelpFold.AlternateText = HelpTextConstants.DepthHelpHtmlText;
			}
		}

		protected void ValidateName(object source, ServerValidateEventArgs args) 
		{
			// Name of ChangeObservation must at least be one sign long
			if(args.Value == null || args.Value.Length <= 0) 
			{
				((CustomValidator)source).ErrorMessage = "Es muss ein Name f&uuml;r den Beobachtungskanal eingegeben werden!";
				args.IsValid = false;
			}
			// Name must not have more than 50 sign's - DB constraint
			else if(args.Value.Length > 50) 
			{
				((CustomValidator)source).ErrorMessage = "Der Name des Beobachtungskanals darf maximal 50 Zeichen lang sein!";
				args.IsValid = false;
			}
			else 
			{
				// only Letters, Digits, - and _ are allowed
				foreach(char tmpChar in args.Value) 
				{
					if(!Char.IsLetterOrDigit(tmpChar) && !(tmpChar == '-') && !(tmpChar == '_')) 
					{
						((CustomValidator)source).ErrorMessage = "Der Name des Beobachtungskanals darf nur aus Buchstaben, Ziffern, &acute;-&acute; und &acute;_&acute; bestehen!";
						args.IsValid = false;
						break;
					}
				}
			}
		}

		protected void ValidateUrl(object source, ServerValidateEventArgs args) 
		{
			// Url of ChangeObservation must at least be one sign long
			if(args.Value == null || args.Value.Length <= 0) 
			{
				((CustomValidator)source).ErrorMessage = "Es muss eine der Beobachtung zugrunde liegende Url angegeben werden!";
				args.IsValid = false;
			} 
			else
			{
				try 
				{
					Uri tmpUri = new Uri(args.Value);
				} 
				catch 
				{
					((CustomValidator)source).ErrorMessage = "Die eingegebene Url ist syntaktisch ung&uuml;ltig!";
					args.IsValid = false;
				}
			}
		}

		protected void ValidateDepth(object source, ServerValidateEventArgs args) 
		{
			// Depth of ChangeObservation must at least be one sign long
			if(args.Value == null || args.Value.Length <= 0) 
			{
				((CustomValidator)source).ErrorMessage = "Es muss eine &Uuml;berpr&uuml;fungstiefe eingegeben werden!";
				args.IsValid = false;
			} 
			else
			{
				try 
				{
					int tmpInt = Convert.ToInt32(args.Value);
					if(0 > tmpInt || 3 < tmpInt) 
					{
						((CustomValidator)source).ErrorMessage = "Die &Uuml;berpr&uuml;fungstiefe muss im Bereich 0 bis 3 liegen!";
						args.IsValid = false;
					}
					
				} 
				catch 
				{
					((CustomValidator)source).ErrorMessage = "Die &Uuml;berpr&uuml;fungstiefe muss eine Zahl sein!";
					args.IsValid = false;
				}
			}
		}

		protected void OkButtonClicked(object sender, EventArgs args) 
		{
			if(Page.IsValid) 
			{
				ObservedDocumentsTree obsTree = (ObservedDocumentsTree)Session["ObservedDocumentsTree"];
				if(obsTree == null) 
				{
					Session["ErrorMessage"] = "Es konnte kein aktueller Beobachtungskanal eruiert werden.";
					Response.Redirect("./ErrorPage.aspx");
					return;
				}
				// Load ChangeObservation from database
				ChangeObservation changeObs;
				string errorMsg;
				ChangeObservationsMediator.ReadResult readRes = ChangeObservationsMediator.LoadChangeObservation(obsTree.Id, out changeObs, out errorMsg);
				if(readRes != ChangeObservationsMediator.ReadResult.Ok) 
				{
					log.Debug("Problem while loading ChangeObservation from Database: " + errorMsg);
					Session["ErrorMessage"] = "Beim Laden des Beobachtungskanals aus der Datenbank ist ein Problem aufgetreten!";
					Response.Redirect("./ErrorPage.aspx");
					return;
				}
				// Change Properties in object-representation
				ChangeObservation.ChangePropertiesEnum changeRes = changeObs.ChangeProperties(this.NameTextBox.Text, new Uri(this.UrlTextBox.Text), Convert.ToInt32(this.DepthTextBox.Text)); 
				if(changeRes != ChangeObservation.ChangePropertiesEnum.NoChange) 
				{
					if(changeRes == ChangeObservation.ChangePropertiesEnum.Inconsistence)
					{
						log.Debug("Problem while changing Properties of ChangeObservation in object-representation.");
						Session["ErrorMessage"] = "Beim Aktualisieren der Eigenschaften des Beobachtungskanals ist ein Problem aufgetreten!";
						Response.Redirect("./ErrorPage.aspx");
						return;
					} 
					else if(changeRes == ChangeObservation.ChangePropertiesEnum.Change) // something changed
					{
						// Update Changes to database
						ChangeObservationsMediator.UpdateResult updateRes = ChangeObservationsMediator.UpdateChangedPropertiesChangeObservation(changeObs, out errorMsg);
						if(updateRes != ChangeObservationsMediator.UpdateResult.Ok) 
						{
							log.Debug("Problem while changing Properties of ChangeObservation in database.");
							Session["ErrorMessage"] = "Beim Aktualisieren der Eigenschaften des Beobachtungskanals in der Datenbank ist ein Problem aufgetreten!";
							Response.Redirect("./ErrorPage.aspx");
							return;
						}
					}
				}
				Session["ObservedDocumentsTree"] = new ObservedDocumentsTree(changeObs);
				Response.Redirect("./ChangeObservationDetail.aspx");
			}
		}

		protected void CancelButtonClicked(object sender, EventArgs args) 
		{
			Response.Redirect("./ChangeObservationDetail.aspx");
		}

		#region Web Form Designer generated code
		override protected void OnInit(EventArgs e)
		{
			//
			// CODEGEN: This call is required by the ASP.NET Web Form Designer.
			//
			InitializeComponent();
			base.OnInit(e);
		}
		
		/// <summary>
		/// Required method for Designer support - do not modify
		/// the contents of this method with the code editor.
		/// </summary>
		private void InitializeComponent()
		{    
			this.Load += new System.EventHandler(this.Page_Load);

		}
		#endregion
	}
}
