using System;
using System.Text;
using System.Text.RegularExpressions;
using System.IO;
using System.Net;
using System.Configuration;
using System.Collections;
using log4net;
using WebseitenMonitor.HtmlUtilities;
using WebseitenMonitor.UriUtilities;

namespace WebseitenMonitor.UserSelection 
{
	public class UserSelectionHtmlPageEnhancer 
	{	
		private static readonly ILog log = LogManager.GetLogger(typeof(UserSelectionHtmlPageEnhancer));
		
		internal const int ElementNodeEnhanceCountAfterBodyElementNode = 2; // Form with encapsuled input tag

		public enum EnhanceHtmlPageResult {Ok, InvalidParams, Failed, IsFramesetHtml, MalformedHtml, ExceptionOccured}

		private HtmlTagScanner enhanceScanner = null;
		private StringBuilder enhanceStringBuilder = null;
		
		/// <summary>
		/// 
		/// </summary>
		/// <param name="sourceReader">TextReader for the HtmlPage-Text to enhance.</param>
		/// <param name="sourceResponseUri">Response Uri (Source Uri) of the HtmlPage to enhance.</param>
		/// <param name="myAbsoluteScriptUriString">Absolute Uri to JavaScript-Source to inject referentially.</param>
		/// <param name="myAbsoluteScriptFormTargetString">Absolute action-Uri of hidden form for post of script generated selection-string.</param>
		/// <param name="result">StringBuilder representing the enhanced HtmlPage-Text</param>
		/// <param name="errorMsg">Error Message in case of not Ok returned, or "" if Ok returned.</param>
		/// <returns></returns>
		public UserSelectionHtmlPageEnhancer.EnhanceHtmlPageResult EnhanceHtmlPage(
			TextReader sourceReader, 
			Uri sourceResponseUri,
			string myAbsoluteScriptUriString,
			string myAbsoluteScriptFormTargetUriString,
			out StringBuilder result,
			out string errorMsg) 
		{	
			// check input-parameters
			if(sourceReader == null) 
			{
				result = null;
				errorMsg = "Es wurde keine Quelle bergeben.";
				return UserSelectionHtmlPageEnhancer.EnhanceHtmlPageResult.InvalidParams;
			}
			if(sourceResponseUri == null)
			{
				result = null;
				errorMsg = "Es wurde keine Quell-Uri bergeben.";
				return UserSelectionHtmlPageEnhancer.EnhanceHtmlPageResult.InvalidParams;
			}
			if(myAbsoluteScriptUriString == null) 
			{
				result = null;
				errorMsg = "Es wurde kein Script-Uri bergeben.";
				return UserSelectionHtmlPageEnhancer.EnhanceHtmlPageResult.InvalidParams;
			} 
			else
			{
				// try if script uri is valid uri
				try {Uri tmpUri = new Uri(myAbsoluteScriptUriString);}
				catch 
				{
					result = null;
					errorMsg = "bergebener Script-Uri ist kein gltiger absoluter Uri.";
					return UserSelectionHtmlPageEnhancer.EnhanceHtmlPageResult.InvalidParams;
				}
			}
			if(myAbsoluteScriptFormTargetUriString == null) 
			{
				result = null;
				errorMsg = "Es wurde kein Script-Form-Target-Uri bergeben.";
				return UserSelectionHtmlPageEnhancer.EnhanceHtmlPageResult.InvalidParams;
			} 
			else 
			{
				try {Uri tmpUri = new Uri(myAbsoluteScriptFormTargetUriString);}
				catch
				{
					result = null;
					errorMsg = "bergebener Script-Form-Target-Uri ist kein gltiger absoluter Uri.";
					return UserSelectionHtmlPageEnhancer.EnhanceHtmlPageResult.InvalidParams;
				}
			}
			// Determine sourceBasePathUri from sourceResonseUri
			string sourceBasePathUriString = UriConvertion.getUrisBasePathUriString(sourceResponseUri);
			if(sourceBasePathUriString == null) 
			{
				result = null;
				errorMsg = "Es konnte kein gltiger Quell-Basis-Uri ermittelt werden.";
				return UserSelectionHtmlPageEnhancer.EnhanceHtmlPageResult.InvalidParams;

			}
			// search for injection-relevant html-tags in html-source
			this.enhanceStringBuilder = new StringBuilder(sourceReader.ReadToEnd());
			this.enhanceScanner = new HtmlTagScanner();
			this.enhanceScanner.Init(this.enhanceStringBuilder);
			HtmlTagToken currentTagToken = this.enhanceScanner.NextTag();
			HtmlTagToken openHtmlTag = null;
			HtmlTagToken openHeadTag = null;
			HtmlTagToken closeHeadTag = null;
			HtmlTagToken openBodyTag = null;
			HtmlTagToken openFramesetTag = null;
			HtmlTagToken nonCloseBaseTag = null;
			while(!(currentTagToken.kind == HtmlTagToken.eof)) 
			{
				if(currentTagToken.name.Equals("html") && currentTagToken.kind == HtmlTagToken.openingTag)
					openHtmlTag = currentTagToken;
				else if(currentTagToken.name.Equals("head") && currentTagToken.kind == HtmlTagToken.openingTag)
					openHeadTag = currentTagToken;
				else if(currentTagToken.name.Equals("head") && currentTagToken.kind == HtmlTagToken.closingTag)
					closeHeadTag = currentTagToken;
				else if(currentTagToken.name.Equals("body") && currentTagToken.kind == HtmlTagToken.openingTag) 
				{
					openBodyTag = currentTagToken;
					break;
				}
				else if(currentTagToken.name.Equals("frameset") && currentTagToken.kind == HtmlTagToken.openingTag) 
				{
					openFramesetTag = currentTagToken;
					break;
				} 
				else if(currentTagToken.name.Equals("base") && currentTagToken.kind != HtmlTagToken.closingTag)
					nonCloseBaseTag = currentTagToken;
				currentTagToken = this.enhanceScanner.NextTag();
			}
			// Check for coars-correctness of html-source
			if(openHtmlTag == null) 
			{
				result = null;
				errorMsg = "Es konnte kein Html-Beginn gefunden werden.";
				return UserSelectionHtmlPageEnhancer.EnhanceHtmlPageResult.MalformedHtml;
			}
			if(openFramesetTag != null) 
			{
				result = null;
				errorMsg = "Bei der bergebenen Html-Quelle handelt es sich um eine Frame-basierte.";
				return UserSelectionHtmlPageEnhancer.EnhanceHtmlPageResult.IsFramesetHtml;
			}
			// assertion: not a frameset-html
			if(openBodyTag == null) 
			{
				result = null;
				errorMsg = "Es konnte kein Html-Body-Beginn gefunden werden.";
				return UserSelectionHtmlPageEnhancer.EnhanceHtmlPageResult.MalformedHtml;
			}
			if(openHeadTag == null ^ closeHeadTag == null) 
			{
				result = null;
				errorMsg = "Der Html-Head-Bereich ist nur halb eingegrenzt.";
				return UserSelectionHtmlPageEnhancer.EnhanceHtmlPageResult.MalformedHtml;
			}
			if(nonCloseBaseTag != null && openHeadTag == null && closeHeadTag == null) 
			{
				result = null;
				errorMsg = "Es wurde ein Html-Base-Element gefunden, jedoch ist kein zugehriger Html-head definiert.";
				return UserSelectionHtmlPageEnhancer.EnhanceHtmlPageResult.MalformedHtml;
			}
			// assertion: non-frame html-source with opening body-element
			// do injections - !!!NOTE!!!: do insert in this.enhanceStringBuilder in reverse direction, so that HtmlTagToken-position keep right for respective next injection
			// inject hidden Html-form for selection-string submit
			this.injectMySelectedElementsSubmitFormAfterThis(openBodyTag, myAbsoluteScriptFormTargetUriString);
			// inject onPageLoad-eventhandler in body-tag
			this.injectMyOnPageLoadEventHandler(openBodyTag);
			// inject Html-base-element in Html-head if not present
			if(openHeadTag != null && closeHeadTag != null) // full head found
			{	
				if(nonCloseBaseTag == null)
					this.injectMyHtmlBaseTagAfterThis(openHeadTag, sourceBasePathUriString);
				else 
					this.injectMyHtmlBaseHrefAttrIfNotDefined(nonCloseBaseTag, sourceBasePathUriString);
				this.injectMyHtmlScriptTagAfterThis(openHeadTag, myAbsoluteScriptUriString);

			} 
			else // no head found
				this.injectMyHtmlHeadWithBaseTagAfterThis(openHtmlTag, myAbsoluteScriptUriString, sourceBasePathUriString);

			result = this.enhanceStringBuilder;
			errorMsg = "";
			return UserSelectionHtmlPageEnhancer.EnhanceHtmlPageResult.Ok;
		} // end: method EnhanceHtmlPage
		
		
		/// <summary>
		/// Note: Name of selectedElementsSubmitForm must be configured in AppSettings under the key "selectedElementsStringHiddenSubmitFormName".
		/// </summary>
		/// <param name="bodyTag">The valid Html-Body-Begin-Token, must not be null in case of call.</param>
		/// <param name="myAbsoluteScriptFormTargetString">Absolute action-Url for Form to inject.</param>
		private void injectMySelectedElementsSubmitFormAfterThis(HtmlTagToken bodyTag, string myAbsoluteScriptFormTargetString)
		{
			string submitFormText = @"
<FORM name=""" + ConfigurationSettings.AppSettings["selectedElementsStringHiddenSubmitFormName"] + @""" method=""post"" action=""" + myAbsoluteScriptFormTargetString + @""">
<INPUT name=""selectionPath"" type=""hidden"" value="""" />
</FORM>
";
			this.enhanceStringBuilder.Insert(bodyTag.endPos + 1, submitFormText);
		}
		
		/// <summary>
		/// Note: Name of selectedElementsSubmitForm must be configured in AppSettings under the key "selectElementsScriptOnLoadFunction".
		/// </summary>
		/// <param name="bodyTag"></param>
		private void injectMyOnPageLoadEventHandler(HtmlTagToken bodyTag) 
		{
			HtmlTagAttributeScanner s = new HtmlTagAttributeScanner();
			s.Init(this.enhanceStringBuilder.ToString(bodyTag.beginPos, bodyTag.endPos - bodyTag.beginPos + 1));
			HtmlTagAttributeToken currAttrToken = null;
			ArrayList relAttrList = new ArrayList();
			currAttrToken = s.NextAttribute();
			// search for all relevant Attributes i.e. onload and onunload
			while(currAttrToken.kind != HtmlTagAttributeToken.eof) 
			{
				if(currAttrToken.name.Equals("onload") || currAttrToken.name.Equals("onunload")) 
				{
					relAttrList.Add(currAttrToken);
				}
				currAttrToken = s.NextAttribute();
			}
			// Remove already present onload or onunload attributes - Note: Do this in reverse order to preserve token-position
			for(int i = relAttrList.Count - 1; i >= 0; i--) 
			{
				this.enhanceStringBuilder.Remove(
					((HtmlTagAttributeToken)relAttrList[i]).nameBeginPos + bodyTag.beginPos,
					((HtmlTagAttributeToken)relAttrList[i]).valOuterEndPos - ((HtmlTagAttributeToken)relAttrList[i]).nameBeginPos);
			}
			// Insert onload attribute for element-selection-script
			this.enhanceStringBuilder.Insert(
				bodyTag.beginPos + bodyTag.name.Length + 1,
				" onload=\"" + ConfigurationSettings.AppSettings["selectElementsScriptOnLoadFunction"] + "\" ");
		}

		private void injectMyHtmlBaseHrefAttrIfNotDefined(HtmlTagToken nonCloseBaseTag, string sourceBasePathUriString) 
		{
			HtmlTagAttributeScanner s = new HtmlTagAttributeScanner();
			s.Init(this.enhanceStringBuilder.ToString(nonCloseBaseTag.beginPos, nonCloseBaseTag.endPos - nonCloseBaseTag.beginPos + 1));
			HtmlTagAttributeToken hrefAttrToken = s.ScanForAttribute("href");
			if(hrefAttrToken == null) // no href-attribute in baseToken - inject one.
			{
				this.enhanceStringBuilder.Insert(
					nonCloseBaseTag.beginPos + nonCloseBaseTag.name.Length + 1,
					" href=\"" + sourceBasePathUriString + "\" ");
			}
		}
		
		private void injectMyHtmlBaseTagAfterThis(HtmlTagToken headTag, string sourceBasePathUriString) 
		{
			string baseTagText = @"
<base href=""" + sourceBasePathUriString + @""" >
";
			this.enhanceStringBuilder.Insert(headTag.endPos + 1, baseTagText);
		}

		private void injectMyHtmlScriptTagAfterThis(
			HtmlTagToken headTag, 
			string myAbsoluteScriptUriString) 
		{
			string scriptTagText = @"
<script src=""" + myAbsoluteScriptUriString + @""" type=""text/javascript""></script>
";
			this.enhanceStringBuilder.Insert(headTag.endPos + 1, scriptTagText);
		}
		
		private void injectMyHtmlHeadWithBaseTagAfterThis(
			HtmlTagToken htmlTag,
			string myAbsoluteScriptUriString,
			string sourceBasePathUriString) 
		{
			string headWithScriptTagText = @"
<head>
<title>WebseitenMonitor</title>
<script src=""" + myAbsoluteScriptUriString + @""" type=""text/javascript""></script>
<base href=""" + sourceBasePathUriString + @""" >
</head>
";
			this.enhanceStringBuilder.Insert(htmlTag.endPos + 1, headWithScriptTagText);
		}
		
	} // end: class UserSelectionHtmlPageEnhancer
	
} // end: namespace WebseitenMonitor.UserSelection